/*
 *	Diameter WebAuth Digest Authentication
 *	Copyright © Konni Hartmann <nnex.ai@googlemail.com>, 2008
 *	Copyright © Alexander Kiok <kiok@math.uni-goettingen.de>, 2008
 */

/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * mod_auth_digest: MD5 digest authentication
 *
 * Originally by Alexei Kosut <akosut@nueva.pvt.k12.ca.us>
 * Updated to RFC-2617 by Ronald Tschalär <ronald@innovation.ch>
 * based on mod_auth, by Rob McCool and Robert S. Thau
 *
 * This module an updated version of modules/standard/mod_digest.c
 * It is still fairly new and problems may turn up - submit problem
 * reports to the Apache bug-database, or send them directly to me
 * at ronald@innovation.ch.
 *
 * Requires either /dev/random (or equivalent) or the truerand library,
 * available for instance from
 * ftp://research.att.com/dist/mab/librand.shar
 *
 * Open Issues:
 *   - qop=auth-int (when streams and trailer support available)
 *   - nonce-format configurability
 *   - Proxy-Authorization-Info header is set by this module, but is
 *     currently ignored by mod_proxy (needs patch to mod_proxy)
 *   - generating the secret takes a while (~ 8 seconds) if using the
 *     truerand library
 *   - The source of the secret should be run-time directive (with server
 *     scope: RSRC_CONF). However, that could be tricky when trying to
 *     choose truerand vs. file...
 *   - shared-mem not completely tested yet. Seems to work ok for me,
 *     but... (definitely won't work on Windoze)
 *   - Sharing a realm among multiple servers has following problems:
 *     o Server name and port can't be included in nonce-hash
 *       (we need two nonce formats, which must be configured explicitly)
 *     o Nonce-count check can't be for equal, or then nonce-count checking
 *       must be disabled. What we could do is the following:
 *       (expected < received) ? set expected = received : issue error
 *       The only problem is that it allows replay attacks when somebody
 *       captures a packet sent to one server and sends it to another
 *       one. Should we add "AuthDigestNcCheck Strict"?
 *   - expired nonces give amaya fits.
 */

#include "apr_sha1.h"
#include "apr_base64.h"
#include "apr_lib.h"
#include "apr_time.h"
#include "apr_errno.h"
#include "apr_global_mutex.h"
#include "apr_strings.h"

#define APR_WANT_STRFUNC
#include "apr_want.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_log.h"
#include "http_protocol.h"
#include "apr_uri.h"
#include "util_md5.h"
#include "apr_shm.h"
#include "apr_rmm.h"
#include "ap_provider.h"

#include "mod_auth.h"

#include <local/circum.h>
#include <libcircum.h>
#include <libHX/list.h>

/* Disable shmem until pools/init gets sorted out
 * remove following two lines when fixed
 */
#undef APR_HAS_SHARED_MEMORY
#define APR_HAS_SHARED_MEMORY 0

/**
 * The auth_diameter configuration record.
 */
typedef struct digest_config_struct {
    const char  *dir_name;
    const char  *realm;
    apr_time_t    nonce_lifetime;
    const char  *ha1;
    const char *servername;
} digest_config_rec;

/********** Grundstrukturen für liste ***********/


//struct HXlist_head state_list;

struct CacheEntry {
    struct HXlist_head list;
    char *sessid;
    char *digest_response;
    authn_status authentication_status;
    /* Add needed variables here  */
};


/*

    HXlist_init(&state_list);


    struct stateentry *entry;
    entry = malloc(sizeof(struct stateentry));


    entry->sessid = xyz;

    HXlist_add_tail(&state_list, &entry->list);



    HXlist_for_each_entry(entry, &state_list, list) {
        if (strcmp(id, entry->sessid) == 0) {
            // Found entry with right id
        }
    }

 */


/********** Grundstrukturen für liste ***********/

#define DFLT_NONCE_LIFE apr_time_from_sec(300)
#define NEXTNONCE_DELTA apr_time_from_sec(30)


#define NONCE_TIME_LEN  (((sizeof(apr_time_t)+2)/3)*4)
#define NONCE_HASH_LEN  (2*APR_SHA1_DIGESTSIZE)
#define NONCE_LEN       (int )(NONCE_TIME_LEN + NONCE_HASH_LEN)

/* client list definitions */

/* struct to hold a parsed Authorization header */

enum hdr_sts {
    NO_HEADER, NOT_DIGEST, INVALID, VALID
};

typedef struct digest_header_struct {
    const char           *scheme;
    const char           *realm;
    const char           *username;
          char           *nonce;
    const char           *uri;
    const char           *method;
    const char           *digest;
    const char           *algorithm;
    const char           *cnonce;
    const char           *opaque;
    unsigned long         opaque_num;
    const char           *message_qop;
    const char           *nonce_count;
    /* the following fields are not (directly) from the header */
    apr_time_t            nonce_time;
    enum hdr_sts          auth_hdr_sts;
    const char           *raw_request_uri;
    apr_uri_t            *psd_request_uri;
    int                   needed_auth;
} digest_header_rec;

/* (mostly) nonce stuff */

typedef union time_union {
    apr_time_t    time;
    unsigned char arr[sizeof(apr_time_t)];
} time_rec;

extern module AP_MODULE_DECLARE_DATA auth_circum_module;

/*
 * initialization code
 */

static int initialize_module(apr_pool_t *p, apr_pool_t *plog,
        apr_pool_t *ptemp, server_rec *s) {
	/* Create the cache for storing the diameter session_ids */
	struct CacheEntry *cache_entry;
	cache_entry = apr_palloc(s->process->pool, sizeof(struct CacheEntry));
	HXlist_init(&cache_entry->list);
	apr_pool_userdata_set(cache_entry, "diameter_session_id_list", apr_pool_cleanup_null, s->process->pool);
        return OK;
    }

/**
 * Create default per server configuration.
 * @param pool: Apache resource pool
 * @param srv: Server configuration record
     */
static void *create_diameter_server_config(apr_pool_t *pool, server_rec *srv) {
    a3_initialize();

    //HXlist_init(&state_list);
    return NULL;
}

/**
 * Create default per directory configuration.
 * @param p: Apache resource pool
 * @param dir: The configured directory
 */
static void *create_digest_dir_config(apr_pool_t *p, char *dir) {
    digest_config_rec *conf;

    if (dir == NULL) {
        return NULL;
    }

    conf = (digest_config_rec *) apr_pcalloc(p, sizeof(digest_config_rec));
    if (conf) {
        conf->nonce_lifetime = DFLT_NONCE_LIFE;
        conf->dir_name       = apr_pstrdup(p, dir);
        conf->servername = "localhost";
    }

    return conf;
}

static const char *set_realm(cmd_parms *cmd, void *config, const char *realm) {
    digest_config_rec *conf = (digest_config_rec *) config;

    /* The core already handles the realm, but it's just too convenient to
     * grab it ourselves too and cache some setups. However, we need to
     * let the core get at it too, which is why we decline at the end -
     * this relies on the fact that http_core is last in the list.
     */
    conf->realm = realm;

    return DECLINE_CMD;
}

static const char *set_nonce_lifetime(cmd_parms *cmd, void *config,
        const char *t) {
    char *endptr;
    long  lifetime;

    lifetime = strtol(t, &endptr, 10);
    if (endptr < (t+strlen(t)) && !apr_isspace(*endptr)) {
        return apr_pstrcat(cmd->pool,
                           "Invalid time in AuthDigestNonceLifetime: ",
                           t, NULL);
    }

    ((digest_config_rec *) config)->nonce_lifetime = apr_time_from_sec(lifetime);
    return NULL;
}

/**
 * Set the servername of the diameter server from the Apache configuration.
 * @param cmd: Server information struct
 * @param cfg: The directory configuration record
 * @param arg: The argument to the directive
 */
static const char *conf_diameter_server_name(cmd_parms *cmd, void *cfg, const char *arg) {
    digest_config_rec *conf = (digest_config_rec*) cfg;
    conf->servername = arg;
    return NULL;
}

/**
 * The Apache configuration directives understood by mod_auth_diameter_digest.
 */
static const command_rec digest_cmds[] = {
    AP_INIT_TAKE1("AuthName", set_realm, NULL, OR_AUTHCFG,
     "The authentication realm (e.g. \"Members Only\")"),
    AP_INIT_TAKE1("AuthDigestNonceLifetime", set_nonce_lifetime, NULL, OR_AUTHCFG,
     "Maximum lifetime of the server nonce (seconds)"),
	AP_INIT_TAKE1("DiameterServerName", conf_diameter_server_name,
		NULL, OR_AUTHCFG, "Set servername"),
	{NULL},
};

/*
 * Authorization header parser code
 */

/**
 * Parse the Authorization header, if it exists
 */
static int get_digest_rec(request_rec *r, digest_header_rec *resp) {
    const char *auth_line;
    apr_size_t l;
    int vk = 0, vv = 0;
    char *key, *value;

    auth_line = apr_table_get(r->headers_in,
                             (PROXYREQ_PROXY == r->proxyreq)
                                 ? "Proxy-Authorization"
                                 : "Authorization");
    if (!auth_line) {
        resp->auth_hdr_sts = NO_HEADER;
        return !OK;
    }

    resp->scheme = ap_getword_white(r->pool, &auth_line);
    if (strcasecmp(resp->scheme, "Digest")) {
        resp->auth_hdr_sts = NOT_DIGEST;
        return !OK;
    }

    l = strlen(auth_line);

    key   = apr_palloc(r->pool, l+1);
    value = apr_palloc(r->pool, l+1);

    while (auth_line[0] != '\0') {

        /* find key */

        while (apr_isspace(auth_line[0])) {
            auth_line++;
        }
        vk = 0;
        while (auth_line[0] != '=' && auth_line[0] != ','
               && auth_line[0] != '\0' && !apr_isspace(auth_line[0])) {
            key[vk++] = *auth_line++;
        }
        key[vk] = '\0';
        while (apr_isspace(auth_line[0])) {
            auth_line++;
        }

        /* find value */

        if (auth_line[0] == '=') {
            auth_line++;
            while (apr_isspace(auth_line[0])) {
                auth_line++;
            }

            vv = 0;
            if (auth_line[0] == '\"') {         /* quoted string */
                auth_line++;
                while (auth_line[0] != '\"' && auth_line[0] != '\0') {
                    if (auth_line[0] == '\\' && auth_line[1] != '\0') {
                        auth_line++;            /* escaped char */
                    }
                    value[vv++] = *auth_line++;
                }
                if (auth_line[0] != '\0') {
                    auth_line++;
                }
            } else { /* token */
                while (auth_line[0] != ',' && auth_line[0] != '\0'
                       && !apr_isspace(auth_line[0])) {
                    value[vv++] = *auth_line++;
                }
            }
            value[vv] = '\0';
        }

        while (auth_line[0] != ',' && auth_line[0] != '\0') {
            auth_line++;
        }
        if (auth_line[0] != '\0') {
            auth_line++;
        }

        if (!strcasecmp(key, "username"))
            resp->username = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "realm"))
            resp->realm = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "nonce"))
            resp->nonce = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "uri"))
            resp->uri = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "response"))
            resp->digest = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "algorithm"))
            resp->algorithm = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "cnonce"))
            resp->cnonce = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "opaque"))
            resp->opaque = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "qop"))
            resp->message_qop = apr_pstrdup(r->pool, value);
        else if (!strcasecmp(key, "nc"))
            resp->nonce_count = apr_pstrdup(r->pool, value);
    }

    if (!resp->username || !resp->realm || !resp->nonce || !resp->uri
        || !resp->digest
        || (resp->message_qop && (!resp->cnonce || !resp->nonce_count))) {
        resp->auth_hdr_sts = INVALID;
        return !OK;
    }

    if (resp->opaque) {
        resp->opaque_num = (unsigned long) strtol(resp->opaque, NULL, 16);
    }

    resp->auth_hdr_sts = VALID;
    return OK;
}

/* Because the browser may preemptively send auth info, incrementing the
 * nonce-count when it does, and because the client does not get notified
 * if the URI didn't need authentication after all, we need to be sure to
 * update the nonce-count each time we receive an Authorization header no
 * matter what the final outcome of the request. Furthermore this is a
 * convenient place to get the request-uri (before any subrequests etc
 * are initiated) and to initialize the request_config.
 *
 * Note that this must be called after mod_proxy had its go so that
 * r->proxyreq is set correctly.
 */
static int parse_hdr_and_update_nc(request_rec *r) {
    digest_header_rec *resp;
    int res;

    if (!ap_is_initial_req(r)) {
        return DECLINED;
    }

    resp = apr_pcalloc(r->pool, sizeof(digest_header_rec));
    resp->raw_request_uri = r->unparsed_uri;
    resp->psd_request_uri = &r->parsed_uri;
    resp->needed_auth = 0;
    resp->method = r->method;
    ap_set_module_config(r->request_config, &auth_circum_module, resp);

    res = get_digest_rec(r, resp);
    return DECLINED;
}

/**	Structure to hold Diameter Digest challenge AVP
 */
struct d_str {
    char *d_realm;
    char *d_nonce;
    char *d_algorithm;
    char *d_qop;
    char *d_opaque;
    char *d_stale;
    char *d_domain;
};

static void fill_digest(struct a3_avp *top, struct d_str *dc)
{
	struct a3_avp *avp;

	if ((avp = a3_avp_find(top, "Digest-Realm")) != NULL)
		dc->d_realm = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Nonce")) != NULL)
		dc->d_nonce = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-QoP")) != NULL)
		dc->d_qop = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Algorithm")) != NULL)
		dc->d_algorithm = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Opaque")) != NULL)
		dc->d_opaque = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Stale")) != NULL)
		dc->d_stale = avp->blob;
	/* Digest-Domain? */
}

/**
 * Revised this whole function to ask diameter server for everything
 */
static void note_digest_auth_failure(request_rec *r,
        const digest_config_rec *conf,
        digest_header_rec *resp, int stale) {
    const char *qop, *opaque, *opaque_param, *domain, *nonce;

    /////// ASK DIAMETER !!!!!!!!!!

    struct a3_conn *ses;
    struct a3_packet *pkt;
    struct a3_avp *avp;
    struct d_str digestChallenge;

    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"Note Digest Auth failure");

    conf = (digest_config_rec *) ap_get_module_config(r->per_dir_config,
            &auth_circum_module);

    //TODO: set sessionid for the connection here!!!
	//TODO: Do we really need to send two packets to diameter?
	//		NO HEADER: note_digest_auth_failure talks to diameter server, as it should be.

    ses = a3_conn_new();
    if (ses == NULL) {
        perror("a3_conn_new");
        return;
    }
    ses->appl_id = a3_idmap_app("WebAuth");
    if (!a3_conn_open(ses, conf->servername)) {
        perror("connect");
        return;
    }
    ses->local_host = malloc(strlen(r->server->server_hostname)+1);
    strcpy(ses->local_host, r->server->server_hostname);
    pkt = a3_pkt_new();
	if (pkt == NULL)
		return;
	pkt->cmdflags = A3_CMDFLAG_REQUEST | A3_CMDFLAG_PROXYABLE;
	pkt->command = a3_idmap_cmd("AA");
	pkt->appl_id = ses->appl_id;

	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);
	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_DIGEST");

    a3_conn_send(ses, pkt);

    pkt = a3_conn_recv(ses);
	if (pkt == NULL)
		return;

    digestChallenge.d_realm = NULL;
    digestChallenge.d_nonce = NULL;
    digestChallenge.d_qop = NULL;
    digestChallenge.d_algorithm = NULL;
    digestChallenge.d_opaque = NULL;
    digestChallenge.d_domain = NULL;
    digestChallenge.d_stale = NULL;

    // Fill the digest Challenge struct
	if ((avp = a3_pkt_find(pkt, "HTTP-Digest-Challenge")) != NULL)
		fill_digest(avp, &digestChallenge);

	a3_pkt_destroy(pkt);

	a3_conn_close(ses);

	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"Received a challenge from our server");

    //TODO: parse digestChallenge.d_qop into a qop_list like structure
    /* Setup qop */
    qop = apr_pstrcat(r->pool, ", qop=\"", digestChallenge.d_qop, NULL);
        qop = apr_pstrcat(r->pool, qop, "\"", NULL);

    opaque = digestChallenge.d_opaque;

    //resp->client->nonce_count = 0;
    //fprintf(stderr, "nonce_count gesetzt\n");

    if (opaque) {
        opaque_param = apr_pstrcat(r->pool, ", opaque=\"", opaque, "\"", NULL);
    } else {
        opaque_param = NULL;
    }

    /* Setup nonce */

    nonce = digestChallenge.d_nonce;

    //FIXME: What is domain and where do we get it!!!
    if (r->proxyreq || !digestChallenge.d_domain) {
        domain = NULL;
    } else {
        domain = digestChallenge.d_domain;
    }

    apr_table_mergen(r->err_headers_out,
                     (PROXYREQ_PROXY == r->proxyreq)
                         ? "Proxy-Authenticate" : "WWW-Authenticate",
                     apr_psprintf(r->pool, "Digest realm=\"%s\", "
                                  "nonce=\"%s\", algorithm=%s%s%s%s%s",
            //FIXME: was digestChallenge.d_realm
            /*ap_auth_name(r)*/ ap_auth_name(r), nonce, digestChallenge.d_algorithm,
                                  opaque_param ? opaque_param : "",
                                  domain ? domain : "",
                                  stale ? ", stale=true" : "", qop));
}

static void copy_uri_components(apr_uri_t *dst,
                                apr_uri_t *src, request_rec *r) {
    if (src->scheme && src->scheme[0] != '\0') {
        dst->scheme = src->scheme;
    } else {
        dst->scheme = (char *) "http";
    }

    if (src->hostname && src->hostname[0] != '\0') {
        dst->hostname = apr_pstrdup(r->pool, src->hostname);
        ap_unescape_url(dst->hostname);
    } else {
        dst->hostname = (char *) ap_get_server_name(r);
    }

    if (src->port_str && src->port_str[0] != '\0') {
        dst->port = src->port;
    } else {
        dst->port = ap_get_server_port(r);
    }

    if (src->path && src->path[0] != '\0') {
        dst->path = apr_pstrdup(r->pool, src->path);
        ap_unescape_url(dst->path);
    } else {
        dst->path = src->path;
    }

    if (src->query && src->query[0] != '\0') {
        dst->query = apr_pstrdup(r->pool, src->query);
        ap_unescape_url(dst->query);
    } else {
        dst->query = src->query;
    }

    dst->hostinfo = src->hostinfo;
}

/* These functions return 0 if client is OK, and proper error status
 * if not... either HTTP_UNAUTHORIZED, if we made a check, and it failed, or
 * HTTP_INTERNAL_SERVER_ERROR, if things are so totally confused that we
 * couldn't figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if the attributes are correct, if it
 * really is that user, if the nonce is correct, etc.
 */

static int authenticate_digest_user(request_rec *r) {
    struct a3_conn *ses;
    struct a3_packet *pkt;
    struct a3_avp *avp;
    struct CacheEntry *cache_entry;
    char *sessid;
    bool found = false;
    digest_config_rec *conf;
    digest_header_rec *resp;
    request_rec       *mainreq;
    const char        *t;

    authn_status return_code = HTTP_UNAUTHORIZED;

    /* do we require Digest auth for this URI? */
    if (!(t = ap_auth_type(r)) || strcasecmp(t, "CircumDiameterDigest")) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                "Digest: Not our");
        return DECLINED;
    }

    if (!ap_auth_name(r)) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Digest: need AuthName: %s", r->uri);
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    /* get the client response and mark */

    mainreq = r;
    while (mainreq->main != NULL) {
        mainreq = mainreq->main;
    }
    while (mainreq->prev != NULL) {
        mainreq = mainreq->prev;
    }
    resp = (digest_header_rec *) ap_get_module_config(mainreq->request_config,
            &auth_circum_module);
    resp->needed_auth = 1;


    /* get our conf */
    conf = (digest_config_rec *) ap_get_module_config(r->per_dir_config,
            &auth_circum_module);

    /* check for existence and syntax of Auth header */
    if (resp->auth_hdr_sts != VALID) {
        if (resp->auth_hdr_sts == NOT_DIGEST) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "Digest: client used wrong authentication scheme "
                          "`%s': %s", resp->scheme, r->uri);
        } else if (resp->auth_hdr_sts == INVALID) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "Digest: missing user, realm, nonce, uri, digest, "
                          "cnonce, or nonce_count in authorization header: %s",
                          r->uri);
        }
        /* else (resp->auth_hdr_sts == NO_HEADER) */
        note_digest_auth_failure(r, conf, resp, 0);
        return HTTP_UNAUTHORIZED;
    }


    r->user         = (char *) resp->username;
    r->ap_auth_type = (char *) "Digest";

    /* check the auth attributes */

    if (strcmp(resp->uri, resp->raw_request_uri)) {
        /* Hmm, the simple match didn't work (probably a proxy modified the
         * request-uri), so lets do a more sophisticated match
         */
        apr_uri_t r_uri, d_uri;

        copy_uri_components(&r_uri, resp->psd_request_uri, r);
        if (apr_uri_parse(r->pool, resp->uri, &d_uri) != APR_SUCCESS) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "Digest: invalid uri <%s> in Authorization header",
                          resp->uri);
            return HTTP_BAD_REQUEST;
        }

        if (d_uri.hostname) {
            ap_unescape_url(d_uri.hostname);
        }
        if (d_uri.path) {
            ap_unescape_url(d_uri.path);
        }

        if (d_uri.query) {
            ap_unescape_url(d_uri.query);
        } else if (r_uri.query) {
            /* MSIE compatibility hack.  MSIE has some RFC issues - doesn't
             * include the query string in the uri Authorization component
             * or when computing the response component.  the second part
             * works out ok, since we can hash the header and get the same
             * result.  however, the uri from the request line won't match
             * the uri Authorization component since the header lacks the
             * query string, leaving us incompatable with a (broken) MSIE.
             *
             * the workaround is to fake a query string match if in the proper
             * environment - BrowserMatch MSIE, for example.  the cool thing
             * is that if MSIE ever fixes itself the simple match ought to
             * work and this code won't be reached anyway, even if the
             * environment is set.
             */

            if (apr_table_get(r->subprocess_env,
                              "AuthDigestEnableQueryStringHack")) {

                ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "Digest: "
                              "applying AuthDigestEnableQueryStringHack "
                              "to uri <%s>", resp->raw_request_uri);

               d_uri.query = r_uri.query;
            }
        }

        if (r->method_number == M_CONNECT) {
            if (!r_uri.hostinfo || strcmp(resp->uri, r_uri.hostinfo)) {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                              "Digest: uri mismatch - <%s> does not match "
                              "request-uri <%s>", resp->uri, r_uri.hostinfo);
                return HTTP_BAD_REQUEST;
            }
        } else if (
            /* check hostname matches, if present */
            (d_uri.hostname && d_uri.hostname[0] != '\0'
              && strcasecmp(d_uri.hostname, r_uri.hostname))
            /* check port matches, if present */
            || (d_uri.port_str && d_uri.port != r_uri.port)
            /* check that server-port is default port if no port present */
            || (d_uri.hostname && d_uri.hostname[0] != '\0'
                && !d_uri.port_str && r_uri.port != ap_default_port(r))
            /* check that path matches */
            || (d_uri.path != r_uri.path
                /* either exact match */
                && (!d_uri.path || !r_uri.path
                    || strcmp(d_uri.path, r_uri.path))
                /* or '*' matches empty path in scheme://host */
                && !(d_uri.path && !r_uri.path && resp->psd_request_uri->hostname
                    && d_uri.path[0] == '*' && d_uri.path[1] == '\0'))
            /* check that query matches */
            || (d_uri.query != r_uri.query
                && (!d_uri.query || !r_uri.query
                    || strcmp(d_uri.query, r_uri.query)))
            ) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "Digest: uri mismatch - <%s> does not match "
                          "request-uri <%s>", resp->uri, resp->raw_request_uri);
            return HTTP_BAD_REQUEST;
        }
    }

    //    return_code = get_hash(r, r->user, conf);

    sessid = malloc(strlen(resp->opaque)+1);
    strcpy(sessid, resp->opaque);

    /* Look in cache for this session and authenticate directly*/
    void *data;
    apr_pool_userdata_get(&data, "diameter_session_id_list", r->server->process->pool);
    HXlist_for_each_entry(cache_entry, &((struct CacheEntry *)data)->list, list) {
        if (!strcmp(cache_entry->sessid, resp->opaque) && !strcmp (cache_entry->digest_response, resp->digest)) {
        	found = TRUE;
        	return cache_entry->authentication_status;
        }
    }

    if (!found)

    {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "Sending AA digest responce with all the needed info =):\n");
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"Opening connection to diameter server:\n");

        ses = a3_conn_new();
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"New Session");
        if (ses == NULL) {
            perror("a3_conn_new");
            return AUTH_GENERAL_ERROR;
        }
        ses->appl_id = a3_idmap_app("WebAuth");
        if (sessid != NULL) {
            ses->session_id = sessid;
    }
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"Session open follows");
        if (!a3_conn_open(ses, conf->servername)) {
            perror("connect");
            return AUTH_GENERAL_ERROR;
        }
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"New Session with session id: %s", ses->session_id);

        ses->local_host = malloc(strlen(r->server->server_hostname)+1);
    	strcpy(ses->local_host, r->server->server_hostname);

        pkt = a3_pkt_new();
	pkt->cmdflags = A3_CMDFLAG_REQUEST | A3_CMDFLAG_PROXYABLE;
	pkt->command = a3_idmap_cmd("AA");
	pkt->appl_id = ses->appl_id;

	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);

        //FIXME: ERROR Here
	//a3_pkt_add_text(pkt, "Destination-Realm", "");

	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_DIGEST");

        avp = a3_pkt_add_avp(pkt, "HTTP-Digest-Response", NULL, 0);
        if (resp->username)
            a3_avp_add_text(avp, "Digest-Username", resp->username);
        if (resp->realm)
            a3_avp_add_text(avp, "Digest-Realm", resp->realm);
        if (resp->nonce)
            a3_avp_add_text(avp, "Digest-Nonce", resp->nonce);
        if (resp->uri)
            a3_avp_add_text(avp, "Digest-URI", resp->uri);
        if (resp->digest)
            a3_avp_add_text(avp, "Digest-Response", resp->digest);

        //optional
        if (resp->algorithm)
            a3_avp_add_text(avp, "Digest-Algorithm", resp->algorithm);
        if (resp->cnonce)
            a3_avp_add_text(avp, "Digest-CNonce", resp->cnonce);

        if (resp->opaque)
            a3_avp_add_text(avp, "Digest-Opaque", resp->opaque);
        if (resp->message_qop)
            a3_avp_add_text(avp, "Digest-QoP", resp->message_qop);
        if (resp->nonce_count)
            a3_avp_add_text(avp, "Digest-Nonce-Count", resp->nonce_count);
        if (resp->method)
            a3_avp_add_text(avp, "Digest-Method", resp->method); // apache modul should forward this
        //a3_avp_add_text(avp, "Digest-Entity-Body-Hash", "0a4c3235xaa"); // dunno =)

            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"Sending Packet");
        a3_conn_send(ses, pkt);
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"Sent Packet");
        pkt = a3_conn_recv(ses);
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"Received Packet");

	if ((avp = a3_pkt_find(pkt, "Result-Code")) != NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"WebAuth said: %u,%u", avp->code, avp->d.v_uint32);
		if (avp->d.v_uint32 == a3_idmap_enum("Result-Code", "Multi-Round-Auth")) {
                    note_digest_auth_failure(r, conf, resp, 1);
                    return_code = HTTP_UNAUTHORIZED;
		} else if (avp->d.v_uint32 == a3_idmap_enum("Result-Code", "Success")) {
                    return_code = OK;
		} else {
            note_digest_auth_failure(r, conf, resp, 0);
                    return_code = HTTP_UNAUTHORIZED;
		}
        }
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "Will now destroy packet");
        a3_pkt_destroy(pkt);
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "Just destroyed packet");

        /* Save session_id, digest_response and authentication status */
        cache_entry = apr_palloc(r->server->process->pool, sizeof(struct CacheEntry));
        cache_entry->sessid = apr_palloc(r->server->process->pool, strlen(sessid) + 1);
        cache_entry->digest_response = apr_palloc(r->server->process->pool, strlen(resp->digest) + 1);
        strcpy(cache_entry->sessid, sessid);
        strcpy(cache_entry->digest_response, resp->digest);
        cache_entry->authentication_status = return_code;
        HXlist_add_tail(&((struct CacheEntry *)data)->list, &cache_entry->list);
        HXlist_for_each_entry(cache_entry, &((struct CacheEntry *)data)->list, list)
        	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        	"cache_entry->sessid %s", cache_entry->sessid);

        a3_conn_close(ses);
    }

    return return_code;

}

/**
 * Authorization-Info header code
 * @param r: The request record
 */
static int add_auth_info(request_rec *r) {
    const digest_config_rec *conf =
                (digest_config_rec *) ap_get_module_config(r->per_dir_config,
            &auth_circum_module);
    digest_header_rec *resp =
                (digest_header_rec *) ap_get_module_config(r->request_config,
            &auth_circum_module);
    const char *ai = NULL, *nextnonce = "";

    if (resp == NULL || !resp->needed_auth || conf == NULL) {
        return OK;
    }


    {
        const char *resp_dig, *ha1, *a2, *ha2;

            ha1 = conf->ha1;
            a2 = apr_pstrcat(r->pool, ":", resp->uri, NULL);
        ha2 = ap_md5(r->pool, (const unsigned char *)a2);

        resp_dig = ap_md5(r->pool,
                          (unsigned char *)apr_pstrcat(r->pool, ha1, ":",
                                                       resp->nonce, ":",
                                                       resp->nonce_count, ":",
                                                       resp->cnonce, ":",
                                                       resp->message_qop ?
                                                         resp->message_qop : "",
                                                       ":", ha2, NULL));

        /* assemble Authentication-Info header
         */
        ai = apr_pstrcat(r->pool,
                         "rspauth=\"", resp_dig, "\"",
                         nextnonce,
                         resp->cnonce ? ", cnonce=\"" : "",
                         resp->cnonce
                           ? ap_escape_quotes(r->pool, resp->cnonce)
                           : "",
                         resp->cnonce ? "\"" : "",
                         resp->nonce_count ? ", nc=" : "",
                         resp->nonce_count ? resp->nonce_count : "",
                         resp->message_qop ? ", qop=" : "",
                         resp->message_qop ? resp->message_qop : "",
                         NULL);
    }

    if (ai && ai[0]) {
        apr_table_mergen(r->headers_out,
                         (PROXYREQ_PROXY == r->proxyreq)
                             ? "Proxy-Authentication-Info"
                             : "Authentication-Info",
                         ai);
    }

    return OK;
}

/**
 * Register hooks.
 * @param p: Apache resource pool
 */
static void register_hooks(apr_pool_t *p) {
    static const char * const cfgPost[] = {"http_core.c", NULL
    };
    static const char * const parsePre[] = {"mod_proxy.c", NULL
    };

    ap_hook_post_config(initialize_module, NULL, cfgPost, APR_HOOK_MIDDLE);
    ap_hook_post_read_request(parse_hdr_and_update_nc, parsePre, NULL, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(authenticate_digest_user, NULL, NULL, APR_HOOK_MIDDLE);

    ap_hook_fixups(add_auth_info, NULL, NULL, APR_HOOK_MIDDLE);
}

EXPORT_SYMBOL module AP_MODULE_DECLARE_DATA auth_circum_module = {
    STANDARD20_MODULE_STUFF,
    create_digest_dir_config,   /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    create_diameter_server_config, /* server config */
    NULL,                       /* merge server config */
    digest_cmds,                /* command table */
    register_hooks              /* register hooks */
};
