/*
 *	Diameter WebAuth Basic Authentication
 *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 2008
 *	Copyright © Konni Hartmann <nnex.ai@googlemail.com>, 2008
 *	Copyright © Alexander Kiok <kiok@math.uni-goettingen.de>, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "mod_auth.h"
#include "ap_config.h"
#include "ap_provider.h"
#include <local/circum.h>
#include <libcircum.h>

struct circum_config_record {
	char *host;
};

extern module AP_MODULE_DECLARE_DATA authn_circum_module;

/**
 * circum_check_password - check if login/password is valid
 * @req:
 * @user:
 * @password:
 *
 * Returns one of %AUTH_DENIED, %AUTH_GRANTED, %AUTH_GENERAL_ERROR.
 */
static authn_status circum_check_password(request_rec *req, const char *user,
    const char *password)
{
	struct circum_config_record *config;
	struct a3_conn *conn;
	struct a3_packet *pkt;
	struct a3_avp *avp;
	authn_status result;

	if (req->per_dir_config == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
			"%s: per_dir_config == NULL", __func__);
		return AUTH_GENERAL_ERROR;
	}

	config = ap_get_module_config(req->per_dir_config, &authn_circum_module);

	if (config == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
			"%s: unexpected config == NULL", __func__);
		return AUTH_GENERAL_ERROR;
	}

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, req,
		"Opening connection to diameter server:");
	conn = a3_conn_new();
	if (conn == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
			"malloc: %s", strerror(errno));
		return AUTH_GENERAL_ERROR;
	}
	conn->appl_id = a3_idmap_app("WebAuth");
	if (!a3_conn_open(conn, config->host)) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, req,
			"connect %s: %s", config->host, strerror(errno));
		free(conn);
		return AUTH_GENERAL_ERROR;
	}

	pkt           = a3_pkt_new();
	pkt->cmdflags = A3_CMDFLAG_REQUEST | A3_CMDFLAG_PROXYABLE;
	pkt->command  = a3_idmap_cmd("AA");
	pkt->appl_id  = conn->appl_id;
	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);
	a3_pkt_add_text(pkt, "Destination-Realm", "");
	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_BASIC");
	a3_pkt_add_text(pkt, "User-Password", password);
	a3_pkt_add_text(pkt, "User-Name", user);

	a3_conn_send(conn, pkt); /* autodestroy */

	result = AUTH_GENERAL_ERROR;
	pkt = a3_conn_recv(conn);
	if (pkt == NULL)
		goto out;

	result = AUTH_DENIED;
	if ((avp = a3_pkt_find(pkt, "Result-Code")) != NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, req,
			"WebAuth reply: %u: %u",
			avp->code, avp->d.v_uint32);
		if (avp->d.v_uint32 ==
		    a3_idmap_enum("Result-Code", "Success"))
			result = AUTH_GRANTED;
	}

	a3_pkt_destroy(pkt);
 out:
	a3_conn_close(conn);
	return result;
}

/**
 * circum_create_server_config - create per-server configuration
 */
static void *circum_create_server_config(apr_pool_t *pool, server_rec *srv)
{
	a3_initialize();
	return NULL;
}

static void *circum_create_dir_config(apr_pool_t *pool, char *dir)
{
	struct circum_config_record *conf = apr_pcalloc(pool, sizeof(*conf));
	conf->host = "localhost";
	return conf;
}

static const authn_provider authn_circum_provider = {
	.check_password = circum_check_password,
};

static void circum_register_hooks(apr_pool_t *pool)
{
	ap_register_provider(pool, AUTHN_PROVIDER_GROUP, "circum", "0",
	                     &authn_circum_provider);
}

static const char *circum_conf_server_name(cmd_parms *cmd, void *vconfig,
    const char *arg)
{
	struct circum_config_record *config = vconfig;
	config->host = arg;
	return NULL;
}

static struct command_struct circum_cmds[] = {
	AP_INIT_TAKE1("DiameterServerName", circum_conf_server_name,
		NULL, OR_AUTHCFG, "Set server name"),
	{NULL},
};

EXPORT_SYMBOL module AP_MODULE_DECLARE_DATA authn_circum_module = {
	STANDARD20_MODULE_STUFF,
	.create_server_config = circum_create_server_config,
	.create_dir_config    = circum_create_dir_config,
	.cmds                 = circum_cmds,
	.register_hooks       = circum_register_hooks,
};
