/*
 *	Circumference Daemon
 *	Configuration and option parsing
 *
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <arpa/inet.h>
#include <sys/socket.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <libHX/option.h>
#include <libHX/string.h>
#include <libHX/libxml_helper.h>
#include <libxml/parser.h>

#include <local/circum.h>
#include "conf.h"

struct callbackmap {
	const char *name;
	const char *(*func)(xmlNode *, struct circumd_config *);
};

static const char *cfparse_socket(xmlNode *ptr, struct circumd_config *config)
{
	char *s_type  = xml_getprop(ptr, "type");
	char *s_proto = xml_getprop(ptr, "proto");
	char *s_port  = xml_getprop(ptr, "port");
	char *s_mand  = xml_getprop(ptr, "mandatory");
	char *s_tls   = xml_getprop(ptr, "tls");
	struct circumd_server_socket sock, *sockptr;
	const char *ret = NULL;

	memset(&sock, 0, sizeof(sock));

	if (s_type == NULL)
		ret = "\"type\" attribute missing for <socket>";
	else if (strcasecmp(s_type, "ipv4") == 0)
		sock.l3proto = PF_INET;
	else if (strcasecmp(s_type, "ipv6") == 0)
		sock.l3proto = PF_INET6;
	else
		ret = "Unknown layer-3 protocol specified in <socket>";

	if (s_proto == NULL && ret == NULL)
		ret = "\"proto\" attribute missing for <socket>";
	else if (strcasecmp(s_proto, "sctp") == 0)
		sock.l4proto = IPPROTO_SCTP;
	else if (strcasecmp(s_proto, "tcp") == 0)
		sock.l4proto = IPPROTO_TCP;
	else
		ret = "Unknown layer-4 protocol specified in <socket>";

	if (s_port == NULL)
		sock.port = 3868;
	else
		sock.port = strtoul(s_port, NULL, 0);

	if (s_mand == NULL)
		sock.mandatory = true;
	else
		sock.mandatory = strtoul(s_mand, NULL, 0);

	sock.tls = s_tls != NULL && (strcmp(s_tls, "1") == 0 ||
	           strcmp(s_tls, "on") == 0 || strcmp(s_tls, "true") == 0);

	if (ret != NULL)
		goto out;

	sockptr = HX_memdup(&sock, sizeof(sock));
	if (sockptr == NULL) {
		perror("malloc");
		ret = "Memory allocation failure";
	} else {
		HXlist_init(&sockptr->list);
		HXclist_push(&config->server_sockets, &sockptr->list);
	}

 out:
	free(s_type);
	free(s_proto);
	free(s_port);
	free(s_mand);
	free(s_tls);
	return ret;
}

static struct callbackmap circumd_conf_tags[] = {
	{"socket", cfparse_socket},
	{NULL},
};

static bool circumd_parse_config(const char *file, struct circumd_config *config)
{
	const struct callbackmap *cmd;
	const char *err;
	xmlNode *ptr;
	xmlDoc *doc;

	doc = xmlParseFile(file);
	if (doc == NULL)
		return false;
	ptr = xmlDocGetRootElement(doc);
	if (ptr == NULL || xml_strcmp(ptr->name, "circumd") != 0) {
		xmlFreeDoc(doc);
		return false;
	}

	for (ptr = ptr->children; ptr != NULL; ptr = ptr->next) {
		if (ptr->type != XML_ELEMENT_NODE)
			continue;
		for (cmd = circumd_conf_tags; cmd->name != NULL; ++cmd)
			if (xml_strcmp(ptr->name, cmd->name) == 0) {
				err = (*cmd->func)(ptr, config);
				if (err != NULL)
					fprintf(stderr, "%s\n", err);
				break;
			}
	}

	xmlFreeDoc(doc);
	return true;
}

static void circumd_flag_D(const struct HXoptcb *cbi)
{
	struct circumd_config *config = cbi->current->uptr;
	config->libcircum_debug = true;
	config->libcircum_debug_flags = cbi->data_long;
}

static void circumd_switch_user(const char *user)
{
	struct passwd *ent;
	if ((ent = getpwnam(user)) != NULL)
		setuid(ent->pw_uid);
	else
		setuid(65534);
}

/**
 * circumd_get_options - parse options
 * @config:	configuration that will be filled in
 */
bool circumd_get_options(int *argc, const char ***argv,
    struct circumd_config *config)
{
	char *config_file = NULL, *user = "nobody";
	struct HXoption options_table[] = {
		{.sh = 'D', .type = HXTYPE_UINT, .cb = circumd_flag_D,
		 .uptr = config, .help = "libcircum debug flags to set"},
		{.sh = 'U', .type = HXTYPE_STRING, .ptr = &user,
		 .help = "Switch to user", .htyp = "USER"},
		{.sh = 'X', .type = HXTYPE_NONE, .ptr = &config->single_threaded,
		 .help = "Run in single-threaded mode (debug)"},
		{.sh = 'c', .type = HXTYPE_STRING, .ptr = &config_file,
		 .help = "Configuration file", .htyp = "FILE"},
		HXOPT_AUTOHELP,
		HXOPT_TABLEEND,
	};
	bool ret;

	if (HX_getopt(options_table, argc, argv, HXOPT_USAGEONERR) < 0)
		return false;

	circumd_switch_user(user);
	if (config_file == NULL) {
		ret = circumd_parse_config(CONFIG_PKGSYSCONFDIR
		      "/circumd.conf.xml", config);
		if (!ret) {
			fprintf(stderr, "Configuration file not found at "
			        "%s/circumd.conf.xml, trying "
			        "./circumd.conf.xml instead.\n",
			        CONFIG_PKGSYSCONFDIR);
			ret = circumd_parse_config("circumd.conf.xml", config);
		}
	} else {
		ret = circumd_parse_config(config_file, config);
		free(config_file);
	}

	return ret;
}
