/*
 *	Digest calculation per RFC 2617
 *	Copyright © Konni Hartmann <nnex.ai@googlemail.com>, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <string.h>
#include <openssl/md5.h>
#include "digcalc.h"

void md5_to_text(char *out, const unsigned char *md)
{
	static const char xlat[] = "0123456789abcdef";
	unsigned int i;

	for (i = 0; i < MD5_DIGEST_LENGTH; ++i) {
		*out++ = xlat[md[i] >> 4];
		*out++ = xlat[md[i] & 0xF];
	}
	*out = '\0';
}

/**
 * @algo:	Algorithm to use ("md5" or "md5-sess")
 * @hexout:	Buffer to put the resulting hex string in.
 * 		size should be at least MD5_DIGEST_LENGTH*2+1.
 *
 * Calculate H(A1).
 */
void digest_calc_ha1(const char *algo, const char *username, const char *realm,
    const char *password, const char *nonce, const char *cnonce,
    char *hexout)
{
	unsigned char md[MD5_DIGEST_LENGTH];
	MD5_CTX ctx;

	MD5_Init(&ctx);
	MD5_Update(&ctx, username, strlen(username));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, realm, strlen(realm));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, password, strlen(password));
	MD5_Final(md, &ctx);

	if (strcasecmp(algo, "md5-sess") == 0) {
		MD5_Init(&ctx);
		MD5_Update(&ctx, md, sizeof(md));
		MD5_Update(&ctx, ":", 1);
		MD5_Update(&ctx, nonce, strlen(nonce));
		MD5_Update(&ctx, ":", 1);
		MD5_Update(&ctx, cnonce, strlen(cnonce));
		MD5_Final(md, &ctx);
	}

	md5_to_text(hexout, md);
}

/**
 * @h_a1:		H(A1), in hex form
 * @nonce:		nonce from the server
 * @nonce_count:	8 hex digits
 * @cnonce:		Client nonce
 * @qop:		QOP-value ("", "auth" or "auth-int")
 * @method:		Method from the request
 * @digest_uri:		Requested URL
 * @entity_hash:	Hash over the entity body, if qop=="auth-int",
 * 			in hex form.
 * @hexout:		Buffer to put the resulting hex string in.
 * 			Size should be at least MD5_DIGEST_LENGTH*2+1.
 *
 * Calculate request-digest / response-digest as per HTTP Digest spec
 */
void digest_calc_response(const char *h_a1, const char *nonce,
    const char *nonce_count, const char *cnonce, const char *qop,
    const char *method, const char *digest_uri, const char *entity_hash,
    char hexout[33])
{
	unsigned char md[MD5_DIGEST_LENGTH];
	MD5_CTX ctx;

	/* Calculate H(A2) */
	MD5_Init(&ctx);
	MD5_Update(&ctx, method, strlen(method));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, digest_uri, strlen(digest_uri));

	if (strcasecmp(qop, "auth-int") == 0) {
		MD5_Update(&ctx, ":", 1);
		MD5_Update(&ctx, entity_hash, MD5_DIGEST_LENGTH * 2);
	}

	MD5_Final(md, &ctx);
	md5_to_text(hexout, md);

	/* Calculate response */
	MD5_Init(&ctx);
	MD5_Update(&ctx, h_a1, MD5_DIGEST_LENGTH * 2);
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, nonce, strlen(nonce));
	MD5_Update(&ctx, ":", 1);

	if (*qop != '\0') {
		MD5_Update(&ctx, nonce_count, strlen(nonce_count));
		MD5_Update(&ctx, ":", 1);
		MD5_Update(&ctx, cnonce, strlen(cnonce));
		MD5_Update(&ctx, ":", 1);
		MD5_Update(&ctx, qop, strlen(qop));
		MD5_Update(&ctx, ":", 1);
	}

	MD5_Update(&ctx, hexout, MD5_DIGEST_LENGTH * 2);
	MD5_Final(md, &ctx);
	md5_to_text(hexout, md);
}
