/*
 *	Circumference Daemon
 *
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <libHX/list.h>
#include <libHX/clist.h>
#include <libHX/misc.h>
#include <libxml/parser.h>

#include <libcircum.h>
#include <local/circum.h>
#include <circumd/module.h>
#include "conf.h"

bool circumd_running = true;
struct HXlist_head circumd_modules;

static void circumd_load_modules(void)
{
	const char *dentry;
	void *dh, *ph;
	struct circumd_module *mod;
	char path[256];
	int ret;

	HXlist_init(&circumd_modules);

	dh = HXdir_open(CONFIG_PKGLIBDIR);
	if (dh == NULL)
		return;

	while ((dentry = HXdir_read(dh)) != NULL) {
		if (*dentry == '.')
			continue;
		if (strlen(dentry) < 3)
			continue;
		if (strcmp(&dentry[strlen(dentry)-3], ".so") != 0)
			continue;

		snprintf(path, sizeof(path), CONFIG_PKGLIBDIR "/%s", dentry);
		ph = HX_dlopen(path);
		if (ph == NULL) {
			fprintf(stderr, "Skipping %s: %s\n",
			        path, HX_dlerror());
			continue;
		}
		if ((mod = HX_dlsym(ph, "circum_module_data")) == NULL) {
			HX_dlclose(ph);
			continue;
		}
		HXlist_init(&mod->list);
		if (mod->init != NULL && (ret = mod->init()) <= 0) {
			fprintf(stderr, "%s (%s) returned an error: %s\n",
			        dentry, mod->app_name, strerror(-ret));
			continue;
		}
		HXlist_add_tail(&circumd_modules, &mod->list);
		fprintf(stderr, "Loaded %s (%s)\n", dentry, mod->app_name);
	}
	HXdir_close(dh);
}

static void circumd_unload_modules(void)
{
	const struct circumd_module *mod;

	HXlist_for_each_entry(mod, &circumd_modules, list) {
		fprintf(stderr, "Unloading module (%s)\n", mod->app_name);
		if (mod->exit != NULL)
			mod->exit();
	}
}

static void circumd_sighandler(int s)
{
	circumd_running = false;
	fprintf(stderr, "Received signal to shut down...\n");
}

static bool circumd_init_signals(void)
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = circumd_sighandler,
	sigemptyset(&sa.sa_mask);
	if (sigaction(SIGINT, &sa, NULL) < 0) {
		perror("sigaction SIGINT");
		return false;
	}

	return true;
}

int main(int argc, const char **argv)
{
	struct circumd_config config;
	int ret = EXIT_FAILURE;

	setvbuf(stdout, NULL, _IOLBF, 0);
	memset(&config, 0, sizeof(config));
	HXclist_init(&config.server_sockets);

	if (!circumd_get_options(&argc, &argv, &config))
		return EXIT_FAILURE;

	a3_initialize();
	if (config.libcircum_debug)
		circum_debug(circum_debug(0) ^ config.libcircum_debug_flags);
	circumd_load_modules();

	if (config.server_sockets.items == 0) {
		fprintf(stderr, "No sockets defined\n");
		return EXIT_FAILURE;
	}
	if (!circumd_init_signals())
		return EXIT_FAILURE;
	if (!circumd_create_sockets(&config))
		goto out;
	if (circumd_server_loop(&config))
		ret = EXIT_SUCCESS;

 out:
	circumd_unload_modules();
	circumd_destroy_sockets(&config);
	xmlCleanupParser();
	a3_deinitialize();
	return ret;
}
