/*
 *	Circumference Daemon
 *	Socket handling
 *
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <sys/types.h> /* must come before socket.h for OpenBSD */
#include <sys/socket.h>
#include <errno.h>
#include <poll.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <libHX/defs.h>
#include <libHX/list.h>
#include <libHX/string.h>

#include <local/circum.h>
#include <libcircum.h>
#include <circumd/module.h>
#include "conf.h"

union sockaddr_big {
	struct sockaddr_in6 inet6;
	struct sockaddr_in inet4;
	struct sockaddr inet0;
};

/**
 * client_capexch_intersect -
 * @in_pkt:	Capability Exchange Request
 * @out_pkt:	Capability Exchange Answer
 *
 * Compute intersection between supported applications.
 */
static unsigned int client_capexch_intersect(const struct a3_packet *in_pkt,
    struct a3_packet *out_pkt)
{
	const struct circumd_module *mod;
	const struct a3_avp *avp;
	unsigned int i = 0;

	/*
	 * Must not use a3_pkt_find here because there may be multiple
	 * AVPs with the same name.
	 */
	HXlist_for_each_entry(avp, &in_pkt->avp_list, list)
		if (avp->code == a3_idmap_avp("Auth-Application-Id"))
			HXlist_for_each_entry(mod, &circumd_modules, list) {
				if (mod->app_id != avp->d.v_uint32)
					continue;
				a3_pkt_add_int(out_pkt, "Auth-Application-Id",
					mod->app_id);
				++i;
			}
		else if (avp->code == a3_idmap_avp("Acct-Application-Id"))
			HXlist_for_each_entry(mod, &circumd_modules, list) {
				if (mod->app_id != avp->d.v_uint32)
					continue;
				a3_pkt_add_int(out_pkt, "Acct-Application-Id",
					mod->app_id);
				++i;
			}

	return i;
}

static bool client_capexch_process(struct circumd_conn *cdc,
    const struct a3_packet *in_pkt)
{
	struct a3_packet *out_pkt;
	const struct a3_avp *avp;

	if ((avp = a3_pkt_find(in_pkt, "Session-Id")) != NULL) {
		free(cdc->l_conn->session_id);
		cdc->l_conn->session_id = HX_strdup(avp->blob);
		if (cdc->l_conn->session_id == NULL)
			return false;
	}

	out_pkt = a3_pkt_new();
	out_pkt->command = in_pkt->command;
	if (client_capexch_intersect(in_pkt, out_pkt) == 0) {
		a3_pkt_add_enum(out_pkt, "Result-Code", "No-Common-Application");
		a3_conn_send(cdc->l_conn, out_pkt);
		return false;
	} else {
		a3_pkt_add_enum(out_pkt, "Result-Code", "Success");
		a3_conn_send(cdc->l_conn, out_pkt);
		return true;
	}
}

/**
 * circumd_send_cmderror - reply for unknown commands
 * @cdc:	circumd connection info
 * @pkt:	original incoming packet
 */
static void circumd_send_cmderror(struct circumd_conn *cdc,
    const struct a3_packet *pkt)
{
	struct a3_packet *out_pkt;

	out_pkt = a3_pkt_new();
	out_pkt->command  = pkt->command;
	out_pkt->cmdflags = (pkt->cmdflags & A3_CMDFLAG_PROXYABLE) |
	                    A3_CMDFLAG_ERROR;
	a3_pkt_add_enum(out_pkt, "Result-Code", "Command-Unsupported");
	a3_conn_send(cdc->l_conn, out_pkt);
}

static void circumd_terminate_session(struct circumd_conn *cdc,
    const struct a3_packet *pkt)
{
	struct a3_packet *out_pkt;

	out_pkt = a3_pkt_new();
	out_pkt->command  = pkt->command;
	out_pkt->cmdflags = A3_CMDFLAG_PROXYABLE;
	a3_pkt_add_enum(out_pkt, "Result-Code", "Success");
	/* Session-Id typically included during send */
	a3_conn_send(cdc->l_conn, out_pkt);
}

static void circumd_handle_packet(struct circumd_conn *cdc,
    struct a3_packet *pkt)
{
	const struct circumd_module *mod;

	if (pkt->command == a3_idmap_cmd("Session-Termination") ||
	    pkt->command == a3_idmap_cmd("Abort-Session"))
		circumd_terminate_session(cdc, pkt);
	else if (pkt->command == a3_idmap_cmd("Capabilities-Exchange"))
		client_capexch_process(cdc, pkt);
	else {
		bool handled = false;

		HXlist_for_each_entry(mod, &circumd_modules, list)
			if ((mod->app_id == 0 || mod->app_id == pkt->appl_id) &&
			    (*mod->packet)(cdc->l_conn, pkt) == CMOD_STOP) {
				handled = true;
				break;
			}

		if (!handled)
			circumd_send_cmderror(cdc, pkt);
	}

	a3_pkt_destroy(pkt);
}

static void *circumd_client_thread(void *arg)
{
	struct circumd_conn *cdc = arg;
	struct a3_conn *conn = cdc->l_conn;
	struct a3_packet *pkt;

	fprintf(stderr, "[%p,%d] Accepted client\n", cdc, conn->fd);

	if (cdc->sv_socket->tls)
		if (circum_conn_starttls(cdc->l_conn, false) <= 0)
			goto out;

	/* Cap-Exchange must happen */
	pkt = a3_conn_recv(cdc->l_conn);
	if (pkt == NULL)
		goto out;
	if (pkt->command != a3_idmap_cmd("Capabilities-Exchange"))
		goto out;
	if (!client_capexch_process(cdc, pkt))
		goto out;
	a3_pkt_destroy(pkt);

	/* OPEN state */
	while ((pkt = a3_conn_recv(conn)) != NULL)
		circumd_handle_packet(cdc, pkt);
	if (errno != 0)
		perror("circum_recv_pkt");

 out:
	fprintf(stderr, "[%p,%d] connection closed\n", cdc, conn->fd);
	circum_conn_leavetls(conn);
	circum_conn_free(conn);
	free(cdc);
	return NULL;
}

/**
 * circumd_new_conn - handle incoming connection
 * @sock:	server-side socket
 * @sg:		single-threaded mode (for debug), block until connection done
 *
 * Accept a new connection and build the a3_session structure with the
 * information about the connected peers so that later a3_conn_send()
 * will construct the AVPs correctly.
 */
static void circumd_new_conn(const struct circumd_server_socket *sock, bool sg)
{
	union sockaddr_big loaddr, rmaddr;
	struct circumd_conn *cdc;
	struct a3_conn *lconn = NULL;
	char node_name[80];
	socklen_t sksize;
	pthread_t ptid;

	cdc = calloc(1, sizeof(struct circumd_conn));
	if (cdc == NULL) {
		perror("calloc");
		goto out;
	}
	cdc->l_conn = calloc(1, sizeof(struct a3_conn));
	if (cdc->l_conn == NULL) {
		perror("calloc");
		free(cdc);
		goto out;
	}

	cdc->sv_socket = sock;
	lconn          = cdc->l_conn;
	lconn->l3proto = sock->l3proto;

	sksize = (sock->l3proto == PF_INET6) ?
	         sizeof(rmaddr.inet6) : sizeof(rmaddr.inet4);
	lconn->fd = accept(sock->fd, &rmaddr.inet0, &sksize);
	if (lconn->fd < 0) {
		perror("accept");
		goto out;
	}

	if (lconn->l3proto == PF_INET6) {
		if (inet_ntop(lconn->l3proto, &rmaddr.inet6.sin6_addr,
		    node_name, sizeof(node_name)) < 0) {
			perror("inet_ntop");
			goto out;
		}
	} else if (lconn->l3proto == PF_INET) {
		if (inet_ntop(lconn->l3proto, &rmaddr.inet4.sin_addr,
		    node_name, sizeof(node_name)) < 0) {
			perror("inet_ntop");
			goto out;
		}
	}
	if ((lconn->remote_host = HX_strdup(node_name)) == NULL) {
		perror("malloc");
		goto out;
	}

	sksize = (lconn->l3proto == PF_INET6) ?
	         sizeof(loaddr.inet6) : sizeof(loaddr.inet4);
	if (getsockname(lconn->fd, &loaddr.inet0, &sksize) < 0) {
		perror("getsockname");
		goto out;
	}
	if (lconn->l3proto == PF_INET6) {
		if (inet_ntop(lconn->l3proto, &loaddr.inet6.sin6_addr,
		    node_name, sizeof(node_name)) < 0) {
			perror("inet_ntop");
			goto out;
		}
	} else if (lconn->l3proto == PF_INET) {
		if (inet_ntop(lconn->l3proto, &loaddr.inet4.sin_addr,
		    node_name, sizeof(node_name)) < 0) {
			perror("inet_ntop");
			goto out;
		}
	}
	lconn->local_addr = lconn->local_host = HX_strdup(node_name);
	if (lconn->local_addr == NULL) {
		perror("strdup");
		goto out;
	}

	/* Launch it */
	if (sg) {
		circumd_client_thread(cdc);
		return;
	}

	pthread_create(&ptid, NULL, circumd_client_thread, cdc);
	pthread_detach(ptid);
	return;

 out:
	perror(__func__);
	if (lconn != NULL)
		circum_conn_free(lconn);
	if (cdc != NULL)
		free(cdc);
}

bool circumd_server_loop(struct circumd_config *config)
{
	const struct circumd_server_socket *sock;
	struct pollfd *pollinfo;
	unsigned int i = 0;

	pollinfo = malloc(config->server_sockets.items * sizeof(struct pollfd));
	if (pollinfo == NULL) {
		perror("malloc");
		return false;
	}

	HXlist_for_each_entry(sock, &config->server_sockets, list) {
		pollinfo[i].fd = sock->fd;
		pollinfo[i].events = POLLIN;
		pollinfo[i++].revents = 0;
	}

	while (circumd_running) {
		if (poll(pollinfo, config->server_sockets.items, -1) < 0)
			/* nothing */;

		i = 0;
		HXlist_for_each_entry(sock, &config->server_sockets, list) {
			if (!(pollinfo[i++].revents & POLLIN))
				continue;
			circumd_new_conn(sock, config->single_threaded);
		}
	}

	free(pollinfo);
	return true;
}

bool circumd_create_sockets(struct circumd_config *config)
{
	struct circumd_server_socket *sock, *safe;
	const int y = 1;
	int ret;

	HXlist_for_each_entry_safe(sock, safe, &config->server_sockets, list) {
		sock->fd = socket(sock->l3proto, SOCK_STREAM, sock->l4proto);
		if (sock->fd < 0) {
			fprintf(stderr, "socket(%u, %u): %s\n",
			        sock->l3proto, sock->l4proto, strerror(errno));
			if (sock->mandatory)
				return false;
			goto nuke;
		}

		fprintf(stderr, "fd=%u l3proto=%u l4proto=%u\n", sock->fd, sock->l3proto, sock->l4proto);

		ret = setsockopt(sock->fd, SOL_SOCKET, SO_REUSEADDR,
		      &y, sizeof(y));
		if (ret < 0) {
			perror("setsockopt");
			if (sock->mandatory)
				return false;
		}

		if (sock->l3proto == PF_INET) {
			memset(&sock->inet4, 0, sizeof(sock->inet4));
			sock->inet4.sin_family = PF_INET;
			sock->inet4.sin_port   = htons(sock->port);
			ret = bind(sock->fd, static_cast(const void *, &sock->inet4), sizeof(sock->inet4));
		} else if (sock->l3proto == PF_INET6) {
#ifdef IPV6_V6ONLY
			if (setsockopt(sock->fd, IPPROTO_IPV6, IPV6_V6ONLY,
			           &y, sizeof(y)) < 0)
				perror("setsockopt IPV6_V6ONLY");
			if (setsockopt(sock->fd, IPPROTO_IPV6, IPV6_V6ONLY,
			           &y, sizeof(y)) < 0)
				perror("setsockopt IPV6_V6ONLY");
#endif
			memset(&sock->inet6, 0, sizeof(sock->inet6));
			sock->inet6.sin6_family = PF_INET6;
			sock->inet6.sin6_port   = htons(sock->port);
			ret = bind(sock->fd, static_cast(const void *, &sock->inet6), sizeof(sock->inet6));
		}

		if (ret < 0) {
			fprintf(stderr, "bind %u: %s\n", sock->fd, strerror(errno));
			if (sock->mandatory)
				return false;
			goto nuke;
		}

		if (listen(sock->fd, SOMAXCONN) < 0) {
			perror("listen");
			if (sock->mandatory)
				return false;
			goto nuke;
		}

		continue;

 nuke:
		close(sock->fd);
		HXclist_del(&config->server_sockets, &sock->list);
		free(sock);
	}

	return true;
}

void circumd_destroy_sockets(const struct circumd_config *config)
{
	struct circumd_server_socket *sock, *next;

	HXlist_for_each_entry_safe(sock, next, &config->server_sockets, list) {
		close(sock->fd);
		free(sock);
	}
}
