/*
 *	WebAuth module for circumd
 *	Copyright © Konni Hartmann <nnex.ai@googlemail.com>, 2008
 *	Copyright © Alexander Kiok <kiok@math.uni-goettingen.de>, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <sys/stat.h>
#include <errno.h>
#include <pthread.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/arbtree.h>
#include <libHX/ctype_helper.h>
#include <libHX/deque.h>
#include <libHX/defs.h>
#include <libHX/misc.h>
#include <libHX/string.h>
#include <openssl/md5.h>
#include <openssl/rand.h>
#include <pcre.h>

#include <local/circum.h>
#include <local/config.h>
#ifdef HAVE_SECURITY_PAM_MODULES_H
#	include <security/pam_appl.h>
#	define WITH_PAM 1
#elif defined(HAVE_PAM_PAM_MODULES_H)
#	include <pam/pam_modules.h>
#	define WITH_PAM 1
#endif
#include <libcircum.h>
#include <circumd/module.h>
#include "digcalc.h"

struct wa_authz_request {
	const char *uri;
	const char *service;
	const char *remote_addr, *remote_user;
	const char *username;
};

struct wa_digest_response {
	const char *username;
	const char *digest_realm;
	const char *nonce;
	const char *uri;
	const char *response;

	const char *algorithm;
	const char *cnonce;
	const char *opaque;
	const char *qop;
	const char *nonce_count;
	const char *method;
	const char *entity_body_hash;
	const char *auth_param;
};

/**
 * wa_authmap - one complete authmap
 * @list:	list of struct wa_authmap_perm
 */
struct wa_authmap {
	struct HXdeque *list;
	const char *file;
	struct stat timestamp;
	pthread_mutex_t lock;
};

/**
 * wa_authmap_perm - a single permission entry
 * @users:	zvec of users for who this entry applies
 * @stack:	zvec of the operand/operator stack
 * @stack_size:	size of the zvec
 * @pattern:	original pattern (for debugging purposes)
 */
struct wa_authmap_perm {
	char **users;
	char **stack;
	unsigned int stack_size;
	char *pattern;
};

struct wa_usermap {
	const char *file;
	struct HXbtree *map;
	struct stat timestamp;
	pthread_mutex_t lock;
};

struct wa_userdata {
	char *realm, *hash;
};

enum {
	WR_SUCCESS,
	WR_AUTHN_REJ,
	WR_AUTHZ_REJ,
	WR_NONCE,
	WR_STALE,
	WR_UNABLE,
	WR_UNSUPPORTED,
};

static struct wa_authmap wa_authmap = {
	.file = CONFIG_PKGSYSCONFDIR "/webauth_authz.txt",
	.lock = PTHREAD_MUTEX_INITIALIZER,
};

static struct wa_usermap wa_usermap[] = {
	{
		.file   = CONFIG_PKGSYSCONFDIR "/webauth_users_basic.txt",
		.lock   = PTHREAD_MUTEX_INITIALIZER,
	},
	{
		.file   = CONFIG_PKGSYSCONFDIR "/webauth_users_digest.txt",
		.lock   = PTHREAD_MUTEX_INITIALIZER,
	},
};

/**
 * Length of the generated secret in bytes
 */
static bool secret_length = 32;
static unsigned char *webauth_secret;

static struct HXbtree *webauth_nonce_db;

/**
 * If set to true, the module will allow any nonce received, provided that the
 * response hash correlates with the nonce. Can be useful to reduce amount of
 * traffic since clients may generate their own nonces, saving one round trip.
 * In the current implementation, this option, when enabled, disables the
 * ability to check the nonce's lifetime, subsequently allowing replay attacks.
 */
static bool accept_foreign_nonces = false;

/**
 * If set to true, the module will send the userhash HA1 together with the
 * nonce. This way an authenticating instance (like Apache) may provide
 * authentication and authorization without further asking the Diameter
 * server. Not recommended since the HA1 value is nearly as useful as the
 * plaintext password. Especially sending the HA1 value of a requested user to
 * any requesting client would be a ridiculous security hole.
 */
static bool send_ha1_with_nonce = false;

/**
 * Lifetime of a nonce in seconds. If a nonce older than %NONCE_TIMEOUT is
 * recieved, an answer with the STALE value set to true is sent.
 */
static unsigned int nonce_timeout = 60;


static inline bool parse_bool(const char *s)
{
	return strcasecmp(s, "yes") == 0 || strcasecmp(s, "on") == 0 ||
	       strcasecmp(s, "true") == 0 || strcmp(s, "1") == 0;
}

static void webauth_read_config(const char *file)
{
	hxmc_t *line = NULL;
	char *key, *value;
	FILE *fp;

	if ((fp = fopen(file, "r")) == NULL) {
		fprintf(stderr, "Cannot open %s: %s\n", file, strerror(errno));
		return;
	}

	while (HX_getl(&line, fp) != NULL) {
		HX_chomp(line);
		key = line;
		while (HX_isspace(*key))
			++key;
		if (*key == '\0' || *key == '#')
			continue;

		value = key;
		while (!HX_isspace(*value) && *value != '\0')
			++value;
		*value++ = '\0';
		while (HX_isspace(*value))
			++value;

		if (strcasecmp(key, "Timeout") == 0)
			nonce_timeout = strtoul(value, NULL, 0);
		else if (strcasecmp(key, "AcceptAllNonces") == 0)
			accept_foreign_nonces = parse_bool(value);
		else if (strcasecmp(key, "SendHA1") == 0)
			send_ha1_with_nonce = parse_bool(value);
		else if (strcasecmp(key, "SecretLength") == 0)
			secret_length = strtoul(value, NULL, 0);
		else if (strcasecmp(key, "BasicUserFile") == 0)
			wa_usermap[0].file = HX_strdup(value);
		else if (strcasecmp(key, "DigestUserFile") == 0)
			wa_usermap[1].file = HX_strdup(value);
		else if (strcasecmp(key, "AuthorizationFile") == 0)
			wa_authmap.file = HX_strdup(value);
		else
			fprintf(stderr, "Unknown tag: %s\n", key);
	}

	fclose(fp);
	HXmc_free(line);
}

static void webauth_clean_space(char *o, const char *i)
{
	while (*i != '\0')
		if (HX_isspace(*i))
			++i;
		else
			*o++ = *i++;
	*o++ = '\0';
}

static void webauth_filltree(struct HXdeque *stack, const char *str)
{
	size_t len = strlen(str);

	if (str[0] == '!') {
		webauth_filltree(stack, &str[1]);
		HXdeque_push(stack, HX_strdup("!"));
	} else if (str[0] == '(' && str[len-1] == ')') {
		unsigned int idx = 1, par = 0;
		char *op, *tmp;

		while (true) {
			switch (str[idx++]) {
			case '(':
				++par;
				continue;
			case ')':
				--par;
				continue;
			case '=':
				if (par != 0)
					continue;
				op = HX_strdup("=");
				goto finished;
			case '|':
				if (par != 0)
					continue;
				op = HX_strdup("|");
				goto finished;
			case '&':
				if (par != 0)
					continue;
				op = HX_strdup("&");
				goto finished;
			} /* switch */
		} /* while */
 finished:
		tmp = HXmc_meminit(&str[1], idx - 2);
		webauth_filltree(stack, tmp);
		HXmc_free(tmp);
		tmp = HXmc_meminit(&str[idx], len - idx - 1);
		webauth_filltree(stack, tmp);
		HXmc_free(tmp);
		HXdeque_push(stack, op);
	} else {
		HXdeque_push(stack, HX_strdup(str));
	}
}

/**
 * webauth_authmap_kill - free entire authz map
 */
static void webauth_authmap_kill(struct wa_authmap *ctl)
{
	const struct HXdeque_node *node;

	for (node = ctl->list->first; node != NULL; node = node->next) {
		struct wa_authmap_perm *perm = node->ptr;

		HX_zvecfree(perm->stack);
		HX_zvecfree(perm->users);
		free(perm->pattern);
	}

	HXdeque_genocide(ctl->list);
}

/**
 * webauth_authmap_refresh - reread auth maps if necessary
 * @ctl:	auth map descriptor
 *
 * Checks the file's timestamps and if they mismatch, reread the auth map from
 * disk. Otherwise nothing happens.
 */
static void webauth_authmap_refresh(struct wa_authmap *ctl)
{
	struct wa_authmap_perm *perm = NULL;
	hxmc_t *line = NULL;
	bool flipstate = 0;
	struct stat sb;
	FILE *fp;

	pthread_mutex_lock(&ctl->lock);

	if (stat(ctl->file, &sb) < 0) {
		fprintf(stderr, "%s: stat %s: %s\n", __func__,
		        ctl->file, strerror(errno));
		goto out;
	}

	/* only second granularity - for now */
	if (sb.st_mtime == ctl->timestamp.st_mtime)
		/* Smart operation: only refresh when really changed */
		goto out;

	if (ctl->list != NULL)
		webauth_authmap_kill(ctl);

	if ((ctl->list = HXdeque_init()) == NULL) {
		perror("HXdeque_init");
		abort();
	}

	fp = fopen(ctl->file, "r");
	if (fp == NULL) {
		fprintf(stderr, "%s: fopen %s: %s\n", __func__,
		        ctl->file, strerror(errno));
		goto out;
	}

	while (HX_getl(&line, fp) != NULL) {
		char *p = line;

		HX_chomp(p);
		while (HX_isspace(*p))
			++p;
		if (*p == '\0' || *p == '#')
			continue;

		if (!flipstate) {
			perm = malloc(sizeof(*perm));
			if (perm == NULL) {
				perror("malloc");
				abort();
			}
			perm->users = HX_split(line, ",", NULL, 0);
		} else {
			struct HXdeque *stack = HXdeque_init();
			if (stack == NULL) {
				perror("HXdeque_init");
				abort();
			}
			perm->pattern = HX_strdup(line);
			webauth_clean_space(line, line);
			webauth_filltree(stack, line);
			perm->stack = reinterpret_cast(char **,
				HXdeque_to_vec(stack, &perm->stack_size));
			HXdeque_push(ctl->list, perm);
			HXdeque_free(stack);
		}
		flipstate = !flipstate;
	}

	memcpy(&ctl->timestamp, &sb, sizeof(sb));
	HXmc_free(line);
	fclose(fp);
	fprintf(stderr, "%s: Updated authmap[%p] from %s with %u entries\n",
	        __func__, ctl, ctl->file, ctl->list->items);

 out:
	pthread_mutex_unlock(&ctl->lock);
}

static void webauth_usermap_kill(struct wa_usermap *ctl)
{
	const struct HXbtree_node *node;
	void *trav = HXbtrav_init(ctl->map);

	while ((node = HXbtraverse(trav)) != NULL) {
		struct wa_userdata *u = node->data;
		free(u->realm);
		free(u->hash);
		free(u);
	}
	HXbtrav_free(trav);
	HXbtree_free(ctl->map);
}

/**
 * webauth_usermap_refresh - reread user maps if necessary
 * @ctl:	user map descriptor
 *
 * Checks the file's timestamps and if they mismatch, reread the user map from
 * disk. Otherwise nothing happens.
 */
static void webauth_usermap_refresh(struct wa_usermap *ctl)
{
	hxmc_t *line = NULL;
	struct stat sb;
	FILE *fp;

	pthread_mutex_lock(&ctl->lock);

	if (stat(ctl->file, &sb) < 0) {
		fprintf(stderr, "%s: stat %s: %s\n", __func__,
		        ctl->file, strerror(errno));
		goto out;
	}

	/* only second granularity - for now */
	if (sb.st_mtime == ctl->timestamp.st_mtime)
		/* Smart operation: only refresh when really changed */
		goto out;

	if (ctl->map != NULL)
		webauth_usermap_kill(ctl);

	ctl->map = HXbtree_init(HXBT_MAP | HXBT_CKEY | HXBT_SCMP);
	if (ctl->map == NULL) {
		perror("HXbtree_init");
		goto out;
	}

	fp = fopen(ctl->file, "r");
	if (fp == NULL) {
		fprintf(stderr, "%s: fopen %s: %s\n", __func__,
		        ctl->file, strerror(errno));
		goto out;
	}

	while (HX_getl(&line, fp) != NULL) {
		struct wa_userdata *user;
		char *stk[3] = {NULL};

		HX_chomp(line);
		if (*line == '\0' || *line == '#')
			continue;
		if (HX_split5(line, ":", ARRAY_SIZE(stk), stk) !=
		    ARRAY_SIZE(stk))
			continue;
		if ((user = malloc(sizeof(struct wa_userdata))) == NULL) {
			perror("malloc");
			break;
		}

		if (*stk[1] == '\0')
			user->realm = NULL;
		else
			if ((user->realm = HX_strdup(stk[1])) == NULL) {
				perror("malloc");
				free(user);
				break;
			}

		if ((user->hash = HX_strdup(stk[2])) == NULL) {
			perror("malloc");
			free(user->realm);
			free(user);
			break;
		}

		if (HXbtree_add(ctl->map, stk[0], user) == NULL) {
			perror("HXbtree_add");
			free(user->realm);
			free(user->hash);
			free(user);
			break;
		}
	}

	memcpy(&ctl->timestamp, &sb, sizeof(sb));
	HXmc_free(line);
	fclose(fp);
	fprintf(stderr, "%s: Updated usermap[%p] from %s with %u user(s)\n",
	        __func__, ctl, ctl->file, ctl->map->items);

 out:
	pthread_mutex_unlock(&ctl->lock);
}

#if WITH_PAM
/**
 * webauth_simple_conv - supplementary func to provide password back to PAM
 */
static int webauth_simple_conv(int num_msg, const struct pam_message **msg_ap,
    struct pam_response **res_ap, void *ptr)
{
	int j;

	for (j = 0; j < num_msg; ++j) {
		switch (msg_ap[j]->msg_style) {
			case PAM_PROMPT_ECHO_ON:
				/* username given to pam already */
				return PAM_CONV_ERR;
			case PAM_PROMPT_ECHO_OFF:
				res_ap[j] = malloc(sizeof(struct pam_response));
				if (res_ap[j] == NULL)
					goto free_up;
				res_ap[j]->resp_retcode = PAM_SUCCESS;
				res_ap[j]->resp = HX_strdup(ptr);
				break;
			case PAM_TEXT_INFO:
				break;
			default:
				return PAM_CONV_ERR;
		}
	}
	return PAM_SUCCESS;

 free_up:
	while (j >= 0) {
		struct pam_response *r = res_ap[j];
		free(r->resp);
		free(r);
		--j;
	}
	errno = ENOMEM;
	return PAM_CONV_ERR;
}

static unsigned int webauth_pam_basic(const char *user, const char *password)
{
	const struct pam_conv conv = {
		.conv        = webauth_simple_conv,
		.appdata_ptr = static_cast(void *, const_cast(char *, password)),
	};
	pam_handle_t *ph;
	int ret;

	if ((ret = pam_start("circumd_webauth", user, &conv, &ph)) ==
	    PAM_SUCCESS) {
		ret = pam_authenticate(ph, 0);
		pam_end(ph, ret);

		switch (ret) {
		case PAM_SUCCESS:
			return WR_SUCCESS;
		case PAM_AUTH_ERR:
		case PAM_ACCT_EXPIRED:
			return WR_AUTHN_REJ;
		case PAM_USER_UNKNOWN:
			return -1;
		}
	}

	return WR_UNABLE;
}
#endif

/**
 * webauth_text_basic - authenticate against users from a text file
 */
static unsigned int webauth_text_basic(const char *user, const char *password)
{
	const struct wa_userdata *data;

	webauth_usermap_refresh(&wa_usermap[0]);

	if ((data = HXbtree_get(wa_usermap[0].map, user)) == NULL)
		return WR_AUTHN_REJ;

	return (strcmp(data->hash, password) == 0) ? WR_SUCCESS : WR_AUTHN_REJ;
}

/**
 * webauth_basic - try all basic auth methods
 *
 */
static unsigned int webauth_basic(const char *user, const char *password)
{
#if WITH_PAM
	unsigned int ret;

	if ((ret = webauth_pam_basic(user, password)) != -1 &&
	    ret != a3_idmap_enum("Result-Code", "Unable-To-Comply"))
		return ret;
#endif

	return webauth_text_basic(user, password);
}

static hxmc_t *webauth_gen_nonce(const char *now, const char *etag,
    const char *sess_id)
{
	unsigned char md[MD5_DIGEST_LENGTH];
	char mdhex[MD5_DIGEST_LENGTH*2+1];
	MD5_CTX ctx;
	hxmc_t *ret;

	if (now != NULL) {
		ret = HXmc_strinit(now);
	} else {
		char tmp[sizeof("18446744073709551616")];
		snprintf(tmp, sizeof(tmp), "%012ld",
		         static_cast(long, time(NULL)));
		ret = HXmc_strinit(tmp);
	}

	MD5_Init(&ctx);
	MD5_Update(&ctx, ret, HXmc_length(ret));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, etag, strlen(etag));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, sess_id, strlen(sess_id));
	MD5_Update(&ctx, ":", 1);
	MD5_Update(&ctx, webauth_secret, secret_length);
	MD5_Final(md, &ctx);

#define q "%02x%02x%02x%02x"
	snprintf(mdhex, sizeof(mdhex), q q q q,
	         md[0], md[1], md[2], md[3], md[4], md[5], md[6], md[7],
	         md[8], md[9], md[10], md[11], md[12], md[13], md[14], md[15]);
#undef q
	HXmc_strcat(&ret, ":");
	HXmc_strcat(&ret, mdhex);
	return ret;
}

/**
 * @hash:	HA1
 */
static unsigned int webauth_send_nonce(struct a3_conn *conn,
    const char *sess_id, const char *hash, bool stale)
{
	struct a3_packet *pkt = a3_pkt_new();
	struct a3_avp *avp;
	hxmc_t *nonce = webauth_gen_nonce(NULL, "", sess_id);

	fprintf(stderr, "Sending nonce\n");

	pkt->cmdflags = A3_CMDFLAG_PROXYABLE;
	pkt->command  = a3_idmap_cmd("AA");
	pkt->appl_id  = a3_idmap_app("WebAuth");

	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);
	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");
	a3_pkt_add_enum(pkt, "Result-Code", "Multi-Round-Auth");
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_DIGEST");
	avp = a3_pkt_add_avp(pkt, "HTTP-Digest-Challenge", NULL, 0);
	a3_avp_add_text(avp, "Digest-Nonce", nonce);
	if (sess_id != NULL)
		/* Info that should be returned unchanged */
		a3_avp_add_text(avp, "Digest-Opaque", sess_id);
	if (stale)
		a3_avp_add_text(avp, "Digest-Stale", "true");
	/* We only support MD5 */
	a3_avp_add_text(avp, "Digest-Algorithm", "MD5");
	/* Only support auth */
	a3_avp_add_text(avp, "Digest-QoP", "auth");

	if (send_ha1_with_nonce && hash != NULL)
		a3_avp_add_text(avp, "Digest-HA1", hash);

	a3_conn_send(conn, pkt);
	HXmc_free(nonce);
	return CMOD_STOP;
}

static unsigned int webauth_send_answer(struct a3_conn *conn,
    const char *sess_id, unsigned int result)
{
	struct a3_packet *pkt = a3_pkt_new();

	pkt->cmdflags = A3_CMDFLAG_PROXYABLE;
	pkt->command  = a3_idmap_cmd("AA");
	pkt->appl_id  = a3_idmap_app("WebAuth");
	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);
	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");
	a3_pkt_add_int(pkt, "Result-Code", result);
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_BASIC");

	a3_conn_send(conn, pkt);
	return CMOD_STOP;
}

/**
 * webauth_copy_dresp - link digest response to AVP data
 * @top:	HTTP-Digest-Response group AVP
 */
static void webauth_copy_dresp(struct wa_digest_response *r,
    const struct a3_avp *top)
{
	const struct a3_avp *avp;

	if ((avp = a3_avp_find(top, "Digest-Response")) != NULL)
		r->response = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Realm")) != NULL)
		r->digest_realm = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Nonce")) != NULL)
		r->nonce = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Method")) != NULL)
		r->method = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-URI")) != NULL)
		r->uri = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-QoP")) != NULL)
		r->qop = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Algorithm")) != NULL)
		r->algorithm = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Entity-Body-Hash")) != NULL)
		r->entity_body_hash = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-CNonce")) != NULL)
		r->cnonce = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Nonce-Count")) != NULL)
		r->nonce_count = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Username")) != NULL)
		r->username = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Opaque")) != NULL)
		r->opaque = avp->blob;
	if ((avp = a3_avp_find(top, "Digest-Auth-Param")) != NULL)
		r->auth_param = avp->blob;
}

/**
 * Authenticate a user/hash combination against the content of shadow.txt
 * This function directly answers the request from the client.
 */
static unsigned int webauth_digest(const struct wa_digest_response *r,
    const char *sess_id)
{
	const struct wa_userdata *uentry;
	char response[MD5_DIGEST_LENGTH*2+1];
	time_t nonce_time;

	if (r->opaque != NULL && strcmp(r->opaque, sess_id) != 0) {
		fprintf(stderr, "Session-Id does not correspond to "
		        "Multi-Round request (client may be stateless)\n");
		return WR_AUTHN_REJ;
	}

	webauth_usermap_refresh(&wa_usermap[1]);
	uentry = HXbtree_get(wa_usermap[1].map, r->username);
	if (uentry == NULL) {
		fprintf(stderr, "Unknown user: %s\n", r->username);
		return WR_AUTHN_REJ;
	}

	if (accept_foreign_nonces) {
		char nonce_time_ex[13];
		hxmc_t *check_nonce;

		if (strlen(r->nonce) != 45 || r->nonce[12] != ':') {
			fprintf(stderr, "WebAuth: Malformaed nonce (%s) of "
			        "length %Zu received. Possible quickmode. "
			        "Sending nonce.\n",
			        r->nonce, strlen(r->nonce));
			return WR_AUTHN_REJ;
		}

		strncpy(nonce_time_ex, r->nonce, 12);
		nonce_time_ex[12] = '\0';
		nonce_time  = strtoul(nonce_time_ex, NULL, 0);
		check_nonce = webauth_gen_nonce(nonce_time_ex, "", sess_id);
		if (check_nonce == NULL || strcmp(r->nonce, check_nonce) != 0) {
			fprintf(stderr, "WebAuth: Wrong nonce provided.\n"
			        "Was:\t%s\n" "Should:\t%s\n"
			        "Possible quickmode. Sending nonce.\n",
			        r->nonce, check_nonce);
			HXmc_free(check_nonce);
			return WR_AUTHN_REJ;
		}
		HXmc_free(check_nonce);
	} else {
		const struct HXbtree_node *nonce_entry =
			HXbtree_find(webauth_nonce_db, r->nonce);

		if (nonce_entry == NULL)
			HXbtree_add(webauth_nonce_db, r->nonce,
				reinterpret_cast(void *,
				static_cast(long, nonce_time = time(NULL))));
		else
			nonce_time = reinterpret_cast(long, nonce_entry->data);
	}

	if (uentry->hash == NULL) {
		fprintf(stderr, "WebAuth: User: %s has no HA1\n", r->username);
		return WR_AUTHN_REJ;
	}

	if (r->qop == NULL) {
		fprintf(stderr, "WebAuth: Digest Legacy Mode used "
		        "(see RFC2617 for updated information)\n");
		digest_calc_response(uentry->hash, r->nonce, NULL, NULL,
			"", r->method, r->uri, r->entity_body_hash, response);
	} else {
		digest_calc_response(uentry->hash, r->nonce, r->nonce_count,
			r->cnonce, r->qop, r->method, r->uri,
			r->entity_body_hash, response);
	}

	if (strcmp(r->response, response) == 0) {
		if (time(NULL) - nonce_time >= nonce_timeout) {
			fprintf(stderr, "WebAuth: Old nonce received\n");
			return WR_STALE;
		} else {
			return WR_SUCCESS;
		}
	} else {
		fprintf(stderr, "WebAuth: User %s provided wrong hash, "
		        "should have been %s\n", r->username, response);
		return WR_AUTHN_REJ;
	}
}

/**
 * webauth_copy_authz - link authz request to AVP data
 * @top:	WebAuth-Authorization-Request group AVP
 */
static void webauth_copy_authz(struct wa_authz_request *r,
    const struct a3_avp *top)
{
	const struct a3_avp *avp;

	if ((avp = a3_avp_find(top, "WebAuth-URI")) != NULL)
		r->uri = avp->blob;
	if ((avp = a3_avp_find(top, "WebAuth-Service")) != NULL)
		r->service = avp->blob;
	if ((avp = a3_avp_find(top, "Remote-User")) != NULL)
		r->remote_user = avp->blob;
	if ((avp = a3_avp_find(top, "User-Name")) != NULL)
		r->username = avp->blob;
	if ((avp = a3_avp_find(top, "Remote-Address")) != NULL)
		r->remote_addr = avp->blob;
}

/**
 * Function to return the value of a variable referenced by its name as a
 * string. Works only on the data rad from an Authorization-Request AVP.
 */
static const char *webauth_getvarvalue(const struct wa_authz_request *z,
    const char *name)
{
	if (strcasecmp(name, "WebAuth-URI") == 0)
		return z->uri;
	else if (strcasecmp(name, "WebAuth-Service") == 0)
		return z->service;
	else if (strcasecmp(name, "Remote-User") == 0)
		return z->remote_user;
	else if (strcasecmp(name, "User-Name") == 0)
		return z->username;
	else if (strcasecmp(name, "Remote-Address") == 0)
		return z->remote_addr;
	return NULL;
}

static bool webauth_evaltree(const char *const *stack,
    const struct wa_authz_request *z, unsigned int *idx)
{
	const char *lh, *rh, *op;
	bool ret = false;

	op = stack[(*idx)--];
	switch (*op) {
	case '&':
		ret = webauth_evaltree(stack, z, idx) &&
		      webauth_evaltree(stack, z, idx);
		break;
	case '|':
		ret = webauth_evaltree(stack, z, idx) ||
		      webauth_evaltree(stack, z, idx);
		break;
	case '!':
		ret = !webauth_evaltree(stack, z, idx);
		break;
	case '=': {
		const char *err;
		int eoff;
		pcre *re;

		rh = stack[(*idx)--];
		lh = webauth_getvarvalue(z, stack[(*idx)--]);

		re = pcre_compile(rh, PCRE_CASELESS | PCRE_UTF8,
		     &err, &eoff, NULL);
		if (re == NULL) {
			fprintf(stderr, "pcre_compile: %s\n", err);
			return false;
		}

		/*
		 * pcre_exec returns the number of captures (which is 0,
		 * since we do not do any), and returns -1 when no match
		 * was found.
		 */
		eoff = pcre_exec(re, NULL, lh, strlen(lh), 0, 0, NULL, 0) >= 0;
		ret = eoff;
		pcre_free(re);
		break;
	    }
	}

	return ret;
}

/**
 * webauth_authz_user - authorize user
 *
 * Returns a Result-Code.
 */
static unsigned int webauth_authz_user(struct wa_authz_request *z)
{
	const struct wa_authmap_perm *re = NULL, *re_anon = NULL;
	const struct HXdeque_node *node;
	unsigned int idx;

	if (z->username == NULL || *z->username == '\0')
		z->username = "<unknown>";

	webauth_authmap_refresh(&wa_authmap);
	for (node = wa_authmap.list->first; node != NULL; node = node->next) {
		const struct wa_authmap_perm *rl = NULL;
		char *const *user_ptr;
		bool found = false;

		rl = node->ptr;
		for (user_ptr = rl->users; *user_ptr != NULL; ++user_ptr) {
			if (strcasecmp(*user_ptr, "<unknown>") == 0)
				re_anon = rl;
			if (strcasecmp(*user_ptr, z->username) == 0) {
				re = rl;
				found = true;
				break;
			}
		}
		if (found)
			break;
	}

	if (re == NULL) {
		fprintf(stderr, "User %s not found - using "
		        "default rule <unknown>\n", z->username);
		re = re_anon;
	} else if (re->stack == NULL || re->stack_size == 0) {
		fprintf(stderr, "Incomplete authz file!\n");
		return WR_AUTHZ_REJ;
	}

	idx = re->stack_size - 1;
	if (!webauth_evaltree(const_cast(const char *const *, re->stack),
	    z, &idx)) {
		fprintf(stderr, "Pattern %s did not match the xmit data\n",
		        re->pattern);
		return WR_AUTHZ_REJ;
	}

	return WR_SUCCESS;
}

/**
 * @ses:	Opened circumd session. Used to send packets to the client.
 * @pkt:	Received packet containing a Diameter AA command.
 *
 * This function processes a received WebAuth AA command. The contained
 * information is saved for further processing and a function corresponding to
 * the type of the command (especially Authentication-Type) is called.
 *
 * Returns %CMD_STOP if the module could handle the request, or %CMOD_CONTINUE
 * if another module should receive the AA-Request message.
 */
static unsigned int webauth_packet(struct a3_conn *conn,
    struct a3_packet *pkt)
{
	unsigned int round = 1, rq_type = 0;
	unsigned int auth_type = -1, result = WR_UNSUPPORTED;
	const char *user = "", *password = "";
	const char *sess_id = NULL, *hash = NULL;
	const struct a3_avp *avp;
	struct wa_digest_response dresp;
	struct wa_authz_request zrq;
	bool authz_available = false;

	if (pkt->command != a3_idmap_cmd("AA"))
		/* we dunno - let it go */
		return CMOD_CONTINUE;

	memset(&dresp, 0, sizeof(dresp));
	memset(&zrq, 0, sizeof(zrq));
	zrq.username    = "<unknown>";
	zrq.remote_user = "*";

	if ((avp = a3_pkt_find(pkt, "User-Name")) != NULL)
		user = avp->blob;
	if ((avp = a3_pkt_find(pkt, "User-Password")) != NULL)
		password = avp->blob;
	if ((avp = a3_pkt_find(pkt, "Auth-Request-Type")) != NULL)
		rq_type = avp->d.v_uint32;
	if ((avp = a3_pkt_find(pkt, "Session-Id")) != NULL)
		sess_id = avp->blob;
	if ((avp = a3_pkt_find(pkt, "WebAuth-Authentication-Type")) != NULL)
		auth_type = avp->d.v_uint32;
	if ((avp = a3_pkt_find(pkt, "HTTP-Digest-Response")) != NULL) {
		round = 2;
		webauth_copy_dresp(&dresp, avp);
	}
	if ((avp = a3_pkt_find(pkt, "WebAuth-Authorization-Request")) != NULL) {
		authz_available = true;
		webauth_copy_authz(&zrq, avp);
	}

	fprintf(stderr, "Let's see, got reqtype %u in round %u\n",
	        rq_type, round);

	/* used for server-side quick mode */
	if (send_ha1_with_nonce && user != NULL) {
		const struct wa_userdata *u =
			HXbtree_get(wa_usermap[1].map, user);
		if (u != NULL)
			hash = u->hash;
	}

	if (auth_type == a3_idmap_enum("WebAuth-Authentication-Type", "HTTP_BASIC")) {
		result = WR_SUCCESS;

		if (rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHENTICATE_ONLY") ||
		    rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_AUTHENTICATE"))
			result = webauth_basic(user, password);

		if (result == WR_SUCCESS &&
		    (rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_ONLY") ||
		    rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_AUTHENTICATE")))
		{
			if (!authz_available)
				result = WR_AUTHZ_REJ;
			else
				result = webauth_authz_user(&zrq);
		}
	} else if (auth_type == a3_idmap_enum("WebAuth-Authentication-Type", "HTTP_DIGEST")) {
		result = WR_SUCCESS;

		if (rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHENTICATE_ONLY") ||
		    rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_AUTHENTICATE"))
		{
			if (round == 1)
				result = WR_NONCE;
			else if (round == 2)
				result = webauth_digest(&dresp, sess_id);
			else
				result = WR_AUTHN_REJ;
		}

		if (result == WR_SUCCESS &&
		    (rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_ONLY") ||
		    rq_type == a3_idmap_enum("Auth-Request-Type", "AUTHORIZE_AUTHENTICATE")))
		{
			if (!authz_available)
				result = WR_AUTHZ_REJ;
			else
				result = webauth_authz_user(&zrq);
		}
	}

	switch (result) {
	case WR_SUCCESS:
		return webauth_send_answer(conn, sess_id,
		       a3_idmap_enum("Result-Code", "Success"));
	case WR_AUTHN_REJ:
		return webauth_send_answer(conn, sess_id,
		       a3_idmap_enum("Result-Code", "Authentication-Rejected"));
	case WR_AUTHZ_REJ:
		return webauth_send_answer(conn, sess_id,
		       a3_idmap_enum("Result-Code", "Authorization-Rejected"));
	case WR_NONCE:
		return webauth_send_nonce(conn, sess_id, hash, false);
	case WR_STALE:
		return webauth_send_nonce(conn, sess_id, hash, true);
	case WR_UNABLE:
		return webauth_send_answer(conn, sess_id,
		       a3_idmap_enum("Result-Code", "Unable-To-Comply"));
	default:
		return webauth_send_answer(conn, sess_id,
		       a3_idmap_enum("Result-Code", "Unsupported-Version"));
	}

	return CMOD_STOP;
}

static void webauth_exit(void)
{
	webauth_usermap_kill(&wa_usermap[0]);
	webauth_usermap_kill(&wa_usermap[1]);
	webauth_authmap_kill(&wa_authmap);
	HXbtree_free(webauth_nonce_db);
	memset(webauth_secret, 0, secret_length); /* be secure */
	free(webauth_secret);
}

static int webauth_init(void);

EXPORT_SYMBOL struct circumd_module circum_module_data = {
	.app_name = "WebAuth",
	.app_id   = 11000,

	.init     = webauth_init,
	.packet   = webauth_packet,
	.exit     = webauth_exit,
};

static int webauth_init(void)
{
	circum_module_data.app_id = a3_idmap_app(circum_module_data.app_name);
	if (circum_module_data.app_id == 0) {
		fprintf(stderr, "%s not found as an application in the id maps\n",
		        circum_module_data.app_name);
		return 0;
	}

	webauth_nonce_db = HXbtree_init(HXBT_CKEY | HXBT_MAP | HXBT_ICMP);
	if (webauth_nonce_db == NULL) {
		perror("HXbtree_init");
		return -errno;
	}

	/*
	 * Could remove these calls to refresh and automatically get deferred
	 * initialization of the user lists.
	 */
	webauth_read_config(CONFIG_PKGSYSCONFDIR "/webauth_config.txt");
	webauth_usermap_refresh(&wa_usermap[0]);
	webauth_usermap_refresh(&wa_usermap[1]);
	webauth_authmap_refresh(&wa_authmap);

	/* Generate WebAuth's secret */
	if ((webauth_secret = malloc(secret_length)) == NULL) {
		perror("malloc");
		abort();
	}
	RAND_bytes(webauth_secret, secret_length);
	return 1;
}
