/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <sys/stat.h>
#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>
#include <libHX/defs.h>
#include <libHX/list.h>
#include <libHX/misc.h>

#include <local/circum.h>
#include "debug.h"
#include "idmap.h"
#include "internal.h"
#include "packet.h"

static const char *const avp_type_name[] = {
	[A3_AVPTYPE_GROUP]    = "GROUP",
	[A3_AVPTYPE_INT32]    = "INT32",
	[A3_AVPTYPE_UINT32]   = "UINT32",
	[A3_AVPTYPE_INT64]    = "INT64",
	[A3_AVPTYPE_UINT64]   = "UINT64",
	[A3_AVPTYPE_FLOAT32]  = "FLOAT32",
	[A3_AVPTYPE_FLOAT64]  = "FLOAT64",
};

static const char *const avp_dyntype_name[] = {
	[A3_AVPTYPE_STRING - __A3_AVPTYPE_DYN] = "STRING",
	[A3_AVPTYPE_BLOB   - __A3_AVPTYPE_DYN] = "BLOB",
};

unsigned int circum_debug_flags;

/**
 * circum_debug - tune debug flags
 * @flags:	new flags to XOR into
 *
 * The given flags are XORed into the debug_flags variable which is returned.
 * This allows for querying the flags without changing them while also
 * providing way to (re)set or clear them.
 */
EXPORT_SYMBOL unsigned int circum_debug(unsigned int flags)
{
	return circum_debug_flags ^= flags;
}

EXPORT_SYMBOL void circum_treedump_avp(const struct a3_avp *avp,
    unsigned int indent)
{
	const struct a3_idmap_avp *id = circum_idmap_avp_code(avp->code);
	const struct a3_avp *subavp;

	printf("%-*s`- ", indent - 3 + 1, "");
	if (circum_debug_flags & CIRCUM_VERBOSE_TREE) {
		printf("avp(%p) code=", avp);
		if (id != NULL)
			printf("%s(%u),", id->name, avp->code);
		else
			printf("%u,", avp->code);

		if (avp->type < __A3_AVPTYPE_DYN)
			printf("type=%s,val", avp_type_name[avp->type]);
		else
			printf("type=%s,val", avp_dyntype_name[avp->type - __A3_AVPTYPE_DYN]);
	} else {
		if (id != NULL)
			printf("%s", id->name);
		else
			printf("%u", avp->code);
	}

	switch (avp->type) {
		case A3_AVPTYPE_INT32:
			printf("=%d", static_cast(int, avp->d.v_int32));
			break;
		case A3_AVPTYPE_UINT32:
			printf("=%u", static_cast(unsigned int, avp->d.v_uint32));
			break;
		case A3_AVPTYPE_INT64:
			printf("=%lld", static_cast(long long, avp->d.v_int64));
			break;
		case A3_AVPTYPE_UINT64:
			printf("=%llu", static_cast(unsigned long long, avp->d.v_uint64));
			break;
		case A3_AVPTYPE_FLOAT32:
			printf("=%f", avp->d.v_float32);
			break;
		case A3_AVPTYPE_FLOAT64:
			printf("=%f", avp->d.v_float64);
			break;
		case A3_AVPTYPE_STRING:
			printf("=\"%s\"", avp->blob);
			break;
		case A3_AVPTYPE_BLOB:
			printf("=B\"%s\"", avp->blob);
			break;
		default:
			break;
	}

	if (circum_debug_flags & CIRCUM_VERBOSE_TREE)
		printf(",bl=%u", avp->blob_length);

	printf("\n");
	if (avp->type == A3_AVPTYPE_GROUP)
		HXlist_for_each_entry(subavp, &avp->d.avp_list, list)
			circum_treedump_avp(subavp, indent + 3);
}

EXPORT_SYMBOL void circum_treedump_pkt(const struct a3_packet *pkt,
    unsigned int indent)
{
	const struct a3_idmap_cmd *id = circum_idmap_cmd_code(pkt->command);
	const struct a3_avp *avp;

	if (circum_debug_flags & CIRCUM_VERBOSE_TREE) {
		printf("pkt(%p) cmd=", pkt);
		if (id != NULL)
			printf("%s(%u)", id->name, pkt->command);
		else
			printf("%u", pkt->command);
	} else {
		if (id != NULL)
			printf("pkt:%s", id->name);
		else
			printf("pkt:%u", pkt->command);
	}

	printf(",flags=0x%x\n", pkt->cmdflags);
	HXlist_for_each_entry(avp, &pkt->avp_list, list)
		circum_treedump_avp(avp, indent + 3);
}

EXPORT_SYMBOL void circum_hexdump(const void *vptr, unsigned int len)
{
	HX_hexdump(stdout, vptr, len);
}
