/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <sys/stat.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/arbtree.h>
#include <libHX/clist.h>
#include <libHX/defs.h>
#include <libHX/misc.h>
#include <libHX/libxml_helper.h>
#include <libxml/parser.h>

#include <local/circum.h>
#include "debug.h"
#include "internal.h"
#include "idmap.h"
#include "packet.h"

struct a3_idmap_enum {
	char *name;
	unsigned int value;
};

enum {
	P_PRIMARY   = 1 << 0,
	P_SECONDARY = 1 << 1,
};

/* For now, do not bother with efficient lookup */
static struct HXbtree *circum_idmap_app_tree;
static struct HXbtree *circum_idmap_cmd_tree;
static struct HXbtree *circum_idmap_cmdrev_tree;
static struct HXbtree *circum_idmap_avp_tree;
static struct HXbtree *circum_idmap_avprev_tree;

EXPORT_SYMBOL unsigned int a3_idmap_app(const char *name)
{
	const struct a3_idmap_app *app =
		HXbtree_get(circum_idmap_app_tree, name);

	if (app != NULL)
		return app->id;

	fprintf(stderr, "%s: unknown app: \"%s\"\n", __func__, name);
	return 0;
}

EXPORT_SYMBOL unsigned int a3_idmap_cmd(const char *name)
{
	const struct a3_idmap_cmd *cmd =
		HXbtree_get(circum_idmap_cmd_tree, name);

	if (cmd != NULL)
		return cmd->code;

	fprintf(stderr, "%s: unknown cmd: \"%s\"\n", __func__, name);
	return 0;
}

/**
 * circum_idmap_cmd_code - return command idmap structure
 * @code:	command code
 */
struct a3_idmap_cmd *circum_idmap_cmd_code(unsigned int code)
{
	return HXbtree_get(circum_idmap_cmdrev_tree,
	       reinterpret_cast(const void *, static_cast(long, code)));
}

/**
 * circum_idmap_avp_code - return AVP idmap structure based on name
 * @name:	AVP name to resolve
 *
 * Only exported for libcircumplus.
 */
EXPORT_SYMBOL struct a3_idmap_avp *circum_idmap_avp_name(const char *name)
{
	return HXbtree_get(circum_idmap_avp_tree, name);
}

/**
 * circum_idmap_avp_code - return AVP idmap structure based on code
 * @code:	AVP code to resolve
 */
struct a3_idmap_avp *circum_idmap_avp_code(unsigned int code)
{
	return HXbtree_get(circum_idmap_avprev_tree,
	       reinterpret_cast(const void *, static_cast(long, code)));
}

/**
 * a3_idmap_avp -
 * @name:	AVP name
 *
 * Returns the AVP code for a given AVP name, or 0 when not found.
 */
EXPORT_SYMBOL unsigned int a3_idmap_avp(const char *name)
{
	const struct a3_idmap_avp *avp =
		HXbtree_get(circum_idmap_avp_tree, name);

	if (avp != NULL)
		return avp->code;
	fprintf(stderr, "%s: unknown avp: \"%s\"\n", __func__, name);
	return 0;
}

EXPORT_SYMBOL unsigned int a3_idmap_enum(const char *avp_name,
    const char *enum_name)
{
	const struct a3_idmap_enum *enum_entry;
	const struct a3_idmap_avp *avp_entry;

	avp_entry = HXbtree_get(circum_idmap_avp_tree, avp_name);
	if (avp_entry == NULL) {
		fprintf(stderr, "%s: unknown avp: \"%s\"\n",
		        __func__, avp_name);
		return -1;
	}

	enum_entry = HXbtree_get(avp_entry->enum_tree, enum_name);
	if (enum_entry != NULL)
		return enum_entry->value;

	fprintf(stderr, "%s: unknown enum \"%s\" for avp \"%s\"\n",
	        __func__, enum_name, avp_name);
	return -1;
}

/**
 * circum_dict_recursive - process one level of child nodes
 * @map:	array of accepted tags
 * @ptr:	node whose children shall be examined
 * @private:	caller-supplied pointer
 */
const char *circum_dict_recursive(const struct circum_xmltag_map *map,
    xmlNode *ptr, void *private)
{
	const struct circum_xmltag_map *entry;
	const char *err;

	for (ptr = ptr->children; ptr != NULL; ptr = ptr->next) {
		if (ptr->type != XML_ELEMENT_NODE)
			continue;
		for (entry = map; entry->name != NULL || entry->func != NULL;
		     ++entry)
		{
			if (entry->name != NULL &&
			    xml_strcmp(ptr->name, entry->name) != 0)
				continue;
			err = (*entry->func)(ptr, private);
			if (err != NULL)
				return err;
			break;
		}
	}

	return NULL;
}

static const char *circum_dict_tag_enum(xmlNode *ptr, void *private)
{
	struct a3_idmap_avp *avp = private;
	struct a3_idmap_enum *enum_entry;
	char *name, *code;

	enum_entry = malloc(sizeof(struct a3_idmap_enum));
	name = xml_getprop(ptr, "name");
	code = xml_getprop(ptr, "code");
	if (enum_entry == NULL || name == NULL || code == NULL)
		goto out;

	enum_entry->name  = name;
	enum_entry->value = strtoul(code, NULL, 0);
	HXbtree_add(avp->enum_tree, name, enum_entry);
	/* do not free(name) -- enum_entry->name still has it */
	free(code);
	return NULL;

 out:
	free(enum_entry);
	free(name);
	free(code);
	return "Malformed <enum> tag or out of memory";
}

static const char *circum_dict_tag_grouped(xmlNode *ptr, void *private)
{
	struct a3_idmap_avp *avp = private;

	/* AVPs start with an enum tree */
	if (avp->type != A3_AVPTYPE_GROUP)
		HXbtree_free(avp->enum_tree);
	/* Presence of <grouped> is enough */
	avp->type = A3_AVPTYPE_GROUP;
	HXclist_init(&avp->rules_fixed);
	HXclist_init(&avp->rules_var);
	return NULL;
}

static const char *circum_dict_tag_type(xmlNode *ptr, void *private)
{
	struct a3_idmap_avp *avp = private;
	char *s;

	s = xml_getprop(ptr, "type-name");
	if (s == NULL)
		return "Malformed <type> tag or out of memory";

	if (strcasecmp(s, "Enumerated") == 0) {
		avp->type = A3_AVPTYPE_ENUM;
	}
	else if (strcasecmp(s, "Unsigned32") == 0)
		avp->type = A3_AVPTYPE_UINT32;
	else if (strcasecmp(s, "Unsigned64") == 0)
		avp->type = A3_AVPTYPE_UINT64;
	else if (strcasecmp(s, "Integer32") == 0)
		avp->type = A3_AVPTYPE_INT32;
	else if (strcasecmp(s, "Integer64") == 0)
		avp->type = A3_AVPTYPE_INT64;
	else if (strcasecmp(s, "UTF8String") == 0)
		avp->type = A3_AVPTYPE_STRING;
	else if (strcasecmp(s, "OctetString") == 0)
		avp->type = A3_AVPTYPE_STRING7B;
	else if (strcasecmp(s, "Time") == 0)
		avp->type = A3_AVPTYPE_TIME;
	else if (strcasecmp(s, "DiameterIdentity") == 0)
		avp->type = A3_AVPTYPE_DIAMIDENT;
	else if (strcasecmp(s, "DiameterURI") == 0)
		avp->type = A3_AVPTYPE_DIAMURI;
	else if (strcasecmp(s, "Float64") == 0)
		avp->type = A3_AVPTYPE_FLOAT64;
	else if (strcasecmp(s, "Float32") == 0)
		avp->type = A3_AVPTYPE_FLOAT32;

	free(s);
	return NULL;
}

static const struct circum_xmltag_map circum_dict_avplevel_tags[] = {
	{"enum",    circum_dict_tag_enum},
	{"grouped", circum_dict_tag_grouped},
	{"type",    circum_dict_tag_type},
	{NULL},
};

/**
 * circum_dict_select_pred - select predicate and raise flag
 * @avp:	avp definition to modify
 * @s:		string from XML, must be of {must,may,shouldnot,mustnot}
 * @flag:	flag to raise if string matches
 */
static void circum_dict_select_pred(struct a3_idmap_avp *avp, const char *s,
    unsigned int flag)
{
	if (strcasecmp(s, "must") == 0)
		avp->must |= flag;
	else if (strcasecmp(s, "mustnot") == 0)
		avp->must_not |= flag;
}

/**
 * circum_dict_tag_avp - handle <avp> tag
 * @ptr:	xml node
 *
 * Allocate a struct a3_idmap_avp, fill in the name, code and constraints, and
 * add it to the global list of known commands. Returns %NULL on success, or
 * an error message on failure.
 */
static const char *circum_dict_tag_avp(xmlNode *ptr, void *private)
{
	struct a3_idmap_avp *avp;
	unsigned int code;
	char *name, *s;

	s = xml_getprop(ptr, "code");
	if (s == NULL)
		return NULL;
	code = strtoul(s, NULL, 0);
	free(s);

	name = xml_getprop(ptr, "name");
	if (name == NULL)
		goto out;

	avp = circum_idmap_avp_code(code);
	if (avp != NULL && strcasecmp(name, avp->name) != 0)
		goto conflict;

	avp = circum_idmap_avp_name(name);
	if (avp != NULL && avp->code != code)
		goto conflict;

	if (avp == NULL) {
		avp = calloc(1, sizeof(struct a3_idmap_avp));
		if (avp == NULL)
			goto out;

		avp->code = code;
		avp->name = name;

		s = xml_getprop(ptr, "mandatory");
		if (s != NULL) {
			circum_dict_select_pred(avp, s, A3_AVPFLAG_MANDATORY);
			free(s);
		}
		s = xml_getprop(ptr, "protected");
		if (s != NULL) {
			circum_dict_select_pred(avp, s, A3_AVPFLAG_PROTECT);
			free(s);
		}

		avp->type = A3_AVPTYPE_BLOB; /* default */
		/*
		 * We always initialize an enum_tree because some definitions
		 * use A3_AVPTYPE_UINT32 and then still register enums.
		 */
		avp->enum_tree = HXbtree_init(HXBT_MAP | HXBT_CID |
		                 HXBT_CMPFN, strcasecmp);
		if (avp->enum_tree == NULL) {
			perror("HXbtree_init");
			abort();
		}
		HXbtree_add(circum_idmap_avp_tree, avp->name, avp);
		HXbtree_add(circum_idmap_avprev_tree, reinterpret_cast(const void *,
			static_cast(long, avp->code)), avp);
	} else {
		free(name);
	}

	return circum_dict_recursive(circum_dict_avplevel_tags, ptr, avp);

 out:
	free(name);
	return "Malformed <avp> tag or out of memory";

 conflict:
	fprintf(stderr, "%s: Trying to register AVP %s (%u) conflicts "
	        "with already-existing AVP %s (%u)\n",
	        __func__, name, code, avp->name, avp->code);
	free(name);
	return "AVP conflict";
}

static const struct circum_xmltag_map circum_dict_toplevel_tags[];

/**
 * circum_dict_tag_avp - handle <avp> tag
 * @ptr:	xml node
 *
 * Allocate a struct a3_idmap_avp, fill in the name, code and constraints, and
 * add it to the global list of known commands. Returns %NULL on success, or
 * an error message on failure.
 */
static const char *circum_dict_tag_base(xmlNode *ptr, void *private)
{
	return circum_dict_recursive(circum_dict_toplevel_tags, ptr, private);
}

/**
 * circum_dict_tag_command - handle <command> tag
 * @ptr:	xml node
 *
 * Allocate a struct a3_idmap_cmd, fill in the name and code, and add it to
 * the global list of known commands. Returns %NULL on success, or an error
 * message on failure.
 */
static const char *circum_dict_tag_command(xmlNode *ptr, void *nothing)
{
	struct a3_idmap_cmd *cmd;
	char *s = NULL;

	cmd = malloc(sizeof(struct a3_idmap_cmd));
	if (cmd == NULL)
		goto out;

	s = xml_getprop(ptr, "code");
	if (s == NULL)
		goto out;
	cmd->code = strtoul(s, NULL, 0);
	free(s);

	if (cmd->code == 0)
		goto ignore;

	cmd->name = xml_getprop(ptr, "name");
	if (cmd->name == NULL)
		goto out;

	if (HXbtree_find(circum_idmap_cmd_tree, cmd->name) != NULL) {
		fprintf(stderr, "%s: warning: Command \"%s\" already exists!\n",
		        __func__, cmd->name);
		goto ignore2;
	}

	HXclist_init(&cmd->reqrules_fixed);
	HXclist_init(&cmd->reqrules_var);
	HXclist_init(&cmd->rsprules_fixed);
	HXclist_init(&cmd->rsprules_var);
	HXbtree_add(circum_idmap_cmd_tree, cmd->name, cmd);
	HXbtree_add(circum_idmap_cmdrev_tree, reinterpret_cast(const void *,
		static_cast(long, cmd->code)), cmd);
	return NULL;

 ignore2:
	free(cmd->name);
 ignore:
	free(cmd);
	return NULL;
 out:
	if (cmd != NULL)
		free(cmd->name);
	free(cmd);
	free(s);
	return "Malformed <command> tag or out of memory";
}

static const char *circum_dict_tag_application(xmlNode *ptr, void *private)
{
	struct a3_idmap_app *app;
	char *s;

	app = malloc(sizeof(struct a3_idmap_app));
	if (app == NULL)
		goto out;

	s = xml_getprop(ptr, "id");
	if (s != NULL) {
		app->id = strtoul(s, NULL, 0);
		free(s);
	} else {
		app->id = 0;
	}

	app->name = xml_getprop(ptr, "name");
	if (app->name == NULL)
		goto out;

	HXbtree_add(circum_idmap_app_tree, app->name, app);
	return circum_dict_recursive(circum_dict_toplevel_tags, ptr, NULL);

 out:
	if (app != NULL)
		free(app->name);
	free(app);
	return "Malformed <application> tag or out of memory";
}

static const struct circum_xmltag_map circum_dict_toplevel_tags[] = {
	{"avp",         circum_dict_tag_avp},
	{"base",        circum_dict_tag_base},
	{"command",     circum_dict_tag_command},
	/* <typedefn> is ignored because it needs coded logic anyway */
	{"application", circum_dict_tag_application},
	{NULL},
};

/**
 * circum_dict_register_file - read a "dictionary" file
 * @path:	pathname to the file
 * @flags:	primary/secondary pass selector
 *
 * Processing of dictionaries is a two-step process. In the first pass, the
 * command, AVP and application definitions are read, and in the second pass,
 * the rules are read. Splitting it this way makes it easier to program, since
 * the definitions are already in memory and can be looked up quickly.
 */
static bool circum_dict_register_file(const char *path, unsigned int flags)
{
	const char *err = NULL;
	xmlNode *ptr;
	xmlDoc *doc;

	doc = xmlParseFile(path);
	if (doc == NULL) {
		fprintf(stderr, "%s: error reading %s\n", __func__, path);
		return false;
	}

	ptr = xmlDocGetRootElement(doc);
	if (ptr == NULL || xml_strcmp(ptr->name, "dictionary") != 0) {
		xmlFreeDoc(doc);
		return false;
	}

	if (flags & P_PRIMARY)
		err = circum_dict_recursive(circum_dict_toplevel_tags, ptr, NULL);
	if (err == NULL && (flags & P_SECONDARY))
		err = circum_dict_recursive(circum_dict_vfytop_tags, ptr, NULL);
	xmlFreeDoc(doc);

	if (err != NULL) {
		fprintf(stderr, "%s: %s\n", __func__, err);
		return false;
	}

	return true;
}

/**
 * circum_dict_register_directory -
 * @dir:	pathspec
 * @flags:	primary/secondary pass selector
 *
 * Slurp in all .xml files of a directory.
 */
static bool circum_dict_register_directory(const char *dir, unsigned int flags)
{
	const char *dentry;
	char path[256];
	void *dh;

	if ((dh = HXdir_open(dir)) == NULL) {
		fprintf(stderr, "Could not open %s: %s\n", CONFIG_DATADIR,
		        strerror(errno));
		return false;
	}

	while ((dentry = HXdir_read(dh)) != NULL) {
		if (*dentry == '.')
			continue;
		if (strlen(dentry) < 4)
			continue;
		if (strcmp(&dentry[strlen(dentry)-4], ".xml") != 0)
			continue;

		snprintf(path, sizeof(path), "%s/%s", dir, dentry);
		if (!circum_dict_register_file(path, flags)) {
			fprintf(stderr, "libcircum: problem reading %s\n", path);
			return false;
		}
	}

	HXdir_close(dh);
	return true;
}

/**
 * a3_dict_register - register dictionaries
 * @path:	pathspec to either directory or a file
 *
 * If @path points to an XML file, it is read. If it points to a directory,
 * all .xml files in that directory are registered.
 */
EXPORT_SYMBOL bool a3_dict_register(const char *path)
{
	struct stat sb;

	if (stat(path, &sb) < 0) {
		fprintf(stderr, "stat: %s: %s\n", path, strerror(errno));
		return false;
	}

	/*
	 * When one registers single files, attention must be paid to the order
	 * with respect to rules referencing AVPs.
	 */
	if (!S_ISDIR(sb.st_mode))
		return circum_dict_register_file(path, P_PRIMARY | P_SECONDARY);

	/*
	 * It is possible that rules from, say, @path/a.xml reference AVPs from
	 * @path/b.xml. In this case, the entire directory needs to have a
	 * first pass before it can get the second one.
	 */
	if (!circum_dict_register_directory(path, P_PRIMARY))
		return false;
	return circum_dict_register_directory(path, P_SECONDARY);
}

void circum_idmap_init(void)
{
	circum_idmap_cmd_tree = HXbtree_init(HXBT_MAP | HXBT_CID |
	                    HXBT_CMPFN, strcasecmp);
	circum_idmap_avp_tree = HXbtree_init(HXBT_MAP | HXBT_CID |
	                    HXBT_CMPFN, strcasecmp);
	circum_idmap_app_tree = HXbtree_init(HXBT_MAP | HXBT_CID |
	                    HXBT_CMPFN, strcasecmp);
	circum_idmap_avprev_tree = HXbtree_init(HXBT_MAP | HXBT_ICMP);
	circum_idmap_cmdrev_tree = HXbtree_init(HXBT_MAP | HXBT_ICMP);
	if (circum_idmap_cmd_tree == NULL || circum_idmap_cmdrev_tree == NULL ||
	    circum_idmap_avp_tree == NULL || circum_idmap_avprev_tree == NULL ||
	    circum_idmap_app_tree == NULL) {
		fprintf(stderr, "HXbtree_init: %s\n", strerror(ENOMEM));
		abort();
	}

	if (!a3_dict_register(CONFIG_DATADIR))
		abort();

	fprintf(stderr, "libcircum: Registered %u application IDs, "
	        "%u command IDs, %u AVP IDs\n",
	        circum_idmap_app_tree->items, circum_idmap_cmd_tree->items,
	        circum_idmap_avp_tree->items);
}

void circum_idmap_exit(void)
{
	const struct HXbtree_node *node;
	void *trav;

	trav = HXbtrav_init(circum_idmap_avp_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		struct a3_idmap_avp *avp = node->data;

		if (avp->type == A3_AVPTYPE_GROUP) {
			struct circum_avprule *rule, *next;

			HXlist_for_each_entry_safe(rule, next,
			    &avp->rules_fixed, list)
				free(rule);
			HXlist_for_each_entry_safe(rule, next,
			    &avp->rules_var, list)
				free(rule);
		} else if (avp->type == A3_AVPTYPE_ENUM) {
			const struct HXbtree_node *node2;
			void *trav2;

			trav2 = HXbtrav_init(avp->enum_tree);
			while ((node2 = HXbtraverse(trav2)) != NULL) {
				struct a3_idmap_enum *enump = node2->data;

				free(enump->name);
				free(enump);
			}
			HXbtrav_free(trav2);
			HXbtree_free(avp->enum_tree);
		} else {
			HXbtree_free(avp->enum_tree);
		}

		free(avp->name);
		free(avp);
	}
	HXbtrav_free(trav);
	HXbtree_free(circum_idmap_avp_tree);
	HXbtree_free(circum_idmap_avprev_tree);

	trav = HXbtrav_init(circum_idmap_cmd_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		struct a3_idmap_cmd *cmd = node->data;
		struct circum_avprule *rule, *next;

		HXlist_for_each_entry_safe(rule, next,
		    &cmd->reqrules_fixed, list)
			free(rule);
		HXlist_for_each_entry_safe(rule, next,
		    &cmd->reqrules_var, list)
			free(rule);
		HXlist_for_each_entry_safe(rule, next,
		    &cmd->rsprules_fixed, list)
			free(rule);
		HXlist_for_each_entry_safe(rule, next,
		    &cmd->rsprules_var, list)
			free(rule);

		free(cmd->name);
		free(cmd);
	}
	HXbtrav_free(trav);
	HXbtree_free(circum_idmap_cmd_tree);
	HXbtree_free(circum_idmap_cmdrev_tree);

	trav = HXbtrav_init(circum_idmap_app_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		struct a3_idmap_app *app = node->data;

		free(app->name);
		free(app);
	}
	HXbtrav_free(trav);
	HXbtree_free(circum_idmap_app_tree);
}

/**
 * a3_idmap_dump - used for debugging memory trashing
 */
EXPORT_SYMBOL void circum_idmap_dump(void)
{
	const struct HXbtree_node *node;
	void *trav;

	trav = HXbtrav_init(circum_idmap_app_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		const struct a3_idmap_app *app = node->data;

		printf("app %p %s -> %u\n", app, app->name, app->id);
	}
	HXbtrav_free(trav);

	trav = HXbtrav_init(circum_idmap_cmd_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		const struct a3_idmap_cmd *cmd = node->data;

		printf("cmd %p %s -> %u\n", cmd, cmd->name, cmd->code);
	}
	HXbtrav_free(trav);

	trav = HXbtrav_init(circum_idmap_cmdrev_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		const struct a3_idmap_cmd *cmd = node->data;

		printf("cmd %p %u -> %s\n", cmd, cmd->code, cmd->name);
	}
	HXbtrav_free(trav);

	trav = HXbtrav_init(circum_idmap_avp_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		const struct a3_idmap_avp *avp = node->data;
		const struct HXbtree_node *node2;
		void *trav2;

		printf("avp %p %s -> %u\n", avp, avp->name, avp->code);
		if (avp->type != A3_AVPTYPE_ENUM)
			continue;

		printf("\ttree %p root %p\n",
		       avp->enum_tree, avp->enum_tree->root);
		trav2 = HXbtrav_init(avp->enum_tree);
		while ((node2 = HXbtraverse(trav2)) != NULL) {
			const struct a3_idmap_enum *enum_id = node2->data;

			printf("\tenum %p %s -> %u\n",
			       enum_id, enum_id->name, enum_id->value);
		}
		HXbtrav_free(trav2);
	}
	HXbtrav_free(trav);

	trav = HXbtrav_init(circum_idmap_avprev_tree);
	while ((node = HXbtraverse(trav)) != NULL) {
		const struct a3_idmap_avp *avp = node->data;

		printf("avp %p %u -> %s\n", avp, avp->code, avp->name);
	}
	HXbtrav_free(trav);
}
