/*
 *	Minimal testsuite for libcircum
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <arpa/inet.h>
#include <sys/time.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX/defs.h>
#include <libHX/option.h>
#include <libcircum.h>

static unsigned int conn_iterations   = 1;
static unsigned int packet_iterations = 1;
static struct timeval timestamp;

static void dump_paths(void)
{
	printf(
		"Circumference was configured with:\n"
		"\t" "prefix=" PC_PREFIX "\n"
		"\t" "bindir=" PC_BINDIR "\n"
		"\t" "sbindir=" PC_SBINDIR "\n"
		"\t" "sysconfdir=" PC_SYSCONFDIR "\n"
		"\t" "datadir=" PC_DATADIR "\n"
		"\t" "mandir=" PC_MANDIR "\n"
		"\t" "pkgsysconfdir=" PC_PKGSYSCONFDIR "\n"
		"\t" "pkgdatadir=" PC_PKGDATADIR "\n"
		"\t" "pkglibdir=" PC_PKGLIBDIR "\n"
		"\t" "pkgconfigdir=" PC_PKGCONFIGDIR "\n"
	);
}

static void time_start(void)
{
	gettimeofday(&timestamp, NULL);
	printf("Clock started at %ld.%06ld\n",
	       static_cast(long, timestamp.tv_sec),
	       static_cast(long, timestamp.tv_usec));
}

static void time_stop(long long *delta)
{
	struct timeval tv;
	long delta_sec, delta_usec;

	gettimeofday(&tv, NULL);

	delta_sec  = tv.tv_sec - timestamp.tv_sec;
	delta_usec = tv.tv_usec - timestamp.tv_usec;
	if (delta_usec < 0) {
		--delta_sec;
		delta_usec += 1000000;
	}
	if (delta != NULL)
		*delta = delta_sec * 1000000 + delta_usec;

	printf("Clock stopped at %ld.%06ld. Delta: %ld.%06ld.\n",
	       static_cast(long, tv.tv_sec),
	       static_cast(long, tv.tv_usec),
	       delta_sec, delta_usec);
}

static struct a3_packet *capex(void)
{
	struct a3_packet *pkt;
	struct a3_avp *avp;

	pkt = a3_pkt_new();
	pkt->command  = a3_idmap_cmd("Capabilities-Exchange");
	pkt->cmdflags = A3_CMDFLAG_REQUEST;

	a3_pkt_add_text(pkt, "Origin-Host", "localhost");
	a3_pkt_add_text(pkt, "Origin-Realm", "localhost");
	a3_pkt_add_text(pkt, "Host-IP-Address", "127.0.0.1");
	a3_pkt_add_int(pkt, "Vendor-Id", 1337);
	a3_pkt_add_text(pkt, "Product-Name", "Circumference Internal, libcircum");

	avp = a3_pkt_add_avp(pkt, "Vendor-Specific-Application-Id", NULL, 0);
	a3_avp_add_int(avp, "Auth-Application-Id", 0x4500);
	avp = a3_avp_add_avp(avp, "Vendor-Specific-Application-Id", NULL, 0);
	a3_avp_add_int(avp, "Auth-Application-Id", 0x4500);
	avp = a3_avp_add_avp(avp, "Vendor-Specific-Application-Id", NULL, 0);
	avp = a3_avp_add_avp(avp, "Vendor-Specific-Application-Id", NULL, 0);

/*
	avp = a3_pkt_add_avp(pkt, "UNREFERENCED GROUP", NULL, 0);
	a3_avp_add_int(avp, "Auth-Application-Id", 0xffe);
	avp = a3_pkt_add_avp(pkt, "UNREFERENCED GROUP", NULL, 0);
	a3_avp_add_int(avp, "Auth-Application-Id", 0xffd);
*/

	a3_pkt_add_int(pkt, "Auth-Application-Id", a3_idmap_app("WebAuth"));
	a3_pkt_add_enum(pkt, "Auth-Request-Type", "AUTHENTICATE_ONLY");

	return pkt;
}

static void pkt_lookup_speed(const struct a3_packet *pkt)
{
	static const unsigned int iters = 100000;
	const struct a3_avp *avp;
	volatile unsigned int x;
	unsigned int i;
	long long t1, t2;

	printf("Timing a3_pkt_find:\n");
	time_start();
	for (i = 0; i < iters; ++i) {
		avp = a3_pkt_find(pkt, "Auth-Request-Type");
		x = avp->code;
		avp = a3_pkt_find(pkt, "Auth-Application-Id");
		x = avp->code;
	}
	time_stop(&t1);

	printf("Timing iterator:\n");
	time_start();
	for (i = 0; i < iters; ++i) {
		HXlist_for_each_entry(avp, &pkt->avp_list, list) {
			if (avp->code == a3_idmap_avp("Auth-Request-Type"))
				x = avp->code;
			else if (avp->code == a3_idmap_avp("Auth-Application-Id"))
				x = avp->code;
		}
	}
	time_stop(&t2);

	printf("Ratio iterator/pkt_find = %.2f\n",
	       static_cast(double, t2) / t1);
}

static void pkt_construct_test(void)
{
	struct a3_packet *pkt;
	unsigned int size;
	void *stream;

	pkt = capex();
	circum_treedump_pkt(pkt, 0);

	stream = a3_pkt_serialize(pkt, &size);
	if (stream == NULL)
		perror("royal failure in a3pkt_serialize");
	else
		circum_hexdump(stream, size);

	a3_pkt_destroy(pkt);

	pkt = a3_pkt_unserialize(stream, size);
	if (pkt == NULL)
		perror("royal failure in a3pkt_unserialize");
	else
		circum_treedump_pkt(pkt, 0);

	pkt_lookup_speed(pkt);
	a3_pkt_destroy(pkt);
	free(stream);

	printf("Done testing packet construction.\n\n\n");
}

static void pkt_verify_test(void)
{
	struct a3_packet *pkt;
	int ret;

	pkt = a3_pkt_new();
	if (pkt == NULL)
		return;

	pkt->command = a3_idmap_cmd("AA"); /* just something */
	a3_pkt_add_text(pkt, "User-Name", "A@cobalt");
	a3_pkt_add_text(pkt, "User-Name", "A@zinc");
	a3_pkt_add_text(pkt, "Origin-Host", "K@tellur");
	a3_pkt_add_text(pkt, "Origin-Realm", "W@nickel");
	a3_pkt_add_text(pkt, "User-Name", "A@iod");
	a3_pkt_add_text(pkt, "Redirect-Host", "B@kana");
	a3_pkt_add_text(pkt, "Origin-Host", "K@boron");
	a3_pkt_add_text(pkt, "Origin-Realm", "W@carbon");
	a3_pkt_add_text(pkt, "User-Name", "A@niob");

	printf("Clustering AVPs, verifying packet\n");
	ret = a3_pkt_verify(pkt);
	if (ret < 0)
		printf("Unable to verify: definitions not available\n");
	else if (ret > 0)
		printf("Passed verification\n");
	else
		printf("Failed verification\n");

	circum_treedump_pkt(pkt, 0);
	printf("Done testing packet verification.\n\n\n");
	a3_pkt_destroy(pkt);
}

static void do_auth(const char *host, const char *user, const char *password,
    const char *uri, const char *type)
{
	struct a3_conn *conn;
	struct a3_packet *pkt;
	struct a3_avp *avp;

	conn = a3_conn_new();
	if (conn == NULL) {
		perror("a3_conn_new");
		return;
	}
	conn->appl_id = a3_idmap_app("WebAuth");
	if (!a3_conn_open(conn, host)) {
		fprintf(stderr, "Failed to connect to %s: %s\n",
		        host, strerror(errno));
		return;
	}

	printf("Sending %s...\n", type);
	pkt = a3_pkt_new();
	pkt->cmdflags = A3_CMDFLAG_REQUEST | A3_CMDFLAG_PROXYABLE;
	pkt->command  = a3_idmap_cmd("AA");
	pkt->appl_id  = conn->appl_id;
	a3_pkt_add_int(pkt, "Auth-Application-Id", pkt->appl_id);
	a3_pkt_add_enum(pkt, "Auth-Request-Type", type);
	a3_pkt_add_enum(pkt, "WebAuth-Authentication-Type", "HTTP_BASIC");

	/* authn */
	a3_pkt_add_text(pkt, "User-Name", user);
	a3_pkt_add_text(pkt, "User-Password", password);

	/* authz */
	avp = a3_pkt_add_avp(pkt, "WebAuth-Authorization-Request", NULL, 0);
	a3_avp_add_text(avp, "WebAuth-URI", uri);
	a3_avp_add_text(avp, "WebAuth-Service", "http");
	a3_avp_add_text(avp, "User-Name", user);
	a3_avp_add_text(avp, "Remote-User", "?");
	a3_avp_add_text(avp, "Remote-Address", "localhost");

	a3_conn_send(conn, pkt);
	pkt = a3_conn_recv(conn);
	if ((avp = a3_pkt_find(pkt, "Result-Code")) != NULL)
		printf("WebAuth/authz said: %u\n", avp->d.v_uint32);
	a3_pkt_destroy(pkt);
	a3_conn_close(conn);
}

static void connection_test(const char *host)
{
	const struct a3_avp *avp;
	struct a3_packet *pkt;
	struct a3_conn *conn;
	unsigned int i;

	/*
	 * Testing session ID generation and host resolution.
	 * I cannot set a reverse lookup for ::1 in BIND9 (just won't resolve).
	 * But everything else works as usual...
	 */
	conn = a3_conn_new();
	if (conn == NULL) {
		perror("a3_conn_new");
		return;
	}
	conn->appl_id = a3_idmap_app("WebAuth");

	if (host == NULL)
		host = "2001:db8::34";
	if (!a3_conn_open(conn, host)) {
		printf("Failed to connect to %s.\n", host);
		return;
	}

	printf("New unit. Connected to %s.\n", host);

	/* Try unknown command code */
	for (i = 0; i < packet_iterations; ++i) {
		pkt = a3_pkt_new();
		pkt->command = 0xC0FFEE;
		a3_conn_send(conn, pkt);

		pkt = a3_conn_recv(conn);
		if (pkt->command != 0xC0FFEE)
			printf("Did not receive same command code\n");
		if ((avp = a3_pkt_find(pkt, "Result-Code")) == NULL) {
			fprintf(stderr, "Result-Code missing?\n");
		} else if (avp->d.v_uint32 !=
		    a3_idmap_enum("Result-Code", "Command-Unsupported")) {
			fprintf(stderr, "Unexpected Result-Code\n");
		}
		a3_pkt_destroy(pkt);
	}

	/* Send more capabilities-exchanges */
	pkt = capex();
	a3_conn_send(conn, pkt);
	pkt = a3_conn_recv(conn);
	if (pkt == NULL ||
	    pkt->command != a3_idmap_cmd("Capabilities-Exchange"))
		printf("Extra CAPEX failure\n");
	a3_pkt_destroy(pkt);
	a3_conn_close(conn);
}

int main(int argc, const char **argv)
{
	char *host = "localhost", *user = "", *password = "", *uri = "/";
	unsigned int i;
	struct HXoption options_table[] = {
		{.sh = 'N', .type = HXTYPE_UINT, .ptr = &conn_iterations,
		 .help = "Connection iterations"},
		{.sh = 'h', .type = HXTYPE_STRING, .ptr = &host,
		 .help = "Host to connect to"},
		{.sh = 'n', .type = HXTYPE_UINT, .ptr = &packet_iterations,
		 .help = "Packet iterations"},
		{.ln = "pw", .type = HXTYPE_STRING, .ptr = &password,
		 .help = "Password to input for authentication"},
		{.ln = "uri", .type = HXTYPE_STRING, .ptr = &uri,
		 .help = "URI to pass in for authorization"},
		{.ln = "user", .type = HXTYPE_STRING, .ptr = &user,
		 .help = "User to sign in for authenticate/authorization"},
		HXOPT_AUTOHELP,
		HXOPT_TABLEEND,
	};

	if (HX_getopt(options_table, &argc, &argv, HXOPT_USAGEONERR) <= 0)
		return EXIT_FAILURE;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	dump_paths();
	a3_initialize();
	pkt_construct_test();
	pkt_verify_test();

	do_auth(host, user, password, uri, "AUTHENTICATE_ONLY");
	do_auth(host, user, password, uri, "AUTHORIZE_ONLY");
	do_auth(host, user, password, uri, "AUTHORIZE_AUTHENTICATE");
	for (i = 0; i < conn_iterations; ++i)
		connection_test(host);

	a3_deinitialize();
	return 0;
}
