/*
 *	Copyright © Jan Engelhardt, 2008 - 2009
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX/defs.h>
#include <libHX/list.h>
#include <libHX/misc.h>
#include <libHX/string.h>

#include <local/circum.h>
#include <libcircum/wirepacket.h>
#include "debug.h"
#include "idmap.h"
#include "internal.h"
#include "net.h"
#include "packet.h"

#if defined(__linux__)
#	define OS_SUFFIX " (Linux)"
#elif defined(__sun__)
#	define OS_SUFFIX " (Solaris)"
#else
#	define OS_SUFFIX
#endif

union sockaddr_big {
	struct sockaddr_in6 inet6;
	struct sockaddr_in inet4;
	struct sockaddr inet0;
};

static uint16_t fakeident[6];
static uint64_t session_mctr;
static pthread_mutex_t session_mctr_lock;

/**
 * circum_net_recv - read packet in blocking mode
 * @fd:		file descriptor
 * @buf:	buffer
 * @count:	size of data
 *
 * read()ing from a socket may return less bytes than requested, for example
 * due to fragmentation into TCP windows. Loop as long as data is wanted.
 * Retains all semantics of read().
 */
EXPORT_SYMBOL ssize_t circum_net_recv(int fd, void *buf, size_t count)
{
	size_t processed = 0;

	while (processed < count) {
		ssize_t ret = read(fd, buf + processed, count);
		if (ret < 0) {
			if (processed > 0)
				break;
			else
				return ret;
		}
		if (ret == 0)
			break;
		processed += ret;
	}
	return processed;
}

/**
 * circum_conn_recv_bpkt - receive a complete frame from the network
 * @conn:	connection
 * @length:	store point for length
 *
 * Reads a complete packet from the network and returns it in an allocated
 * block, which must be freed by the caller when not needed anymore. The
 * length of the buffer is stored in @length so as to provide a length for
 * a potential call to circum_hexdump().
 * On error, %NULL will be returned and errno is set.
 */
EXPORT_SYMBOL void *circum_conn_recv_bpkt(struct a3_conn *conn,
    unsigned int *length)
{
	struct a3_wire_packet wire_pkt;
	void *buffer;
	ssize_t ret;

	ret = conn->tls_active ?
	      circum_tls_recv(conn, &wire_pkt, sizeof(wire_pkt)) :
	      circum_net_recv(conn->fd, &wire_pkt, sizeof(wire_pkt));
	if (ret != sizeof(wire_pkt)) {
		errno = 0;
		return NULL;
	}

	*length = circum_ntoh24(wire_pkt.length);
	buffer = malloc(*length);
	if (buffer == NULL)
		return NULL;

	memcpy(buffer, &wire_pkt, sizeof(wire_pkt));
	if (conn->tls_active)
		ret = circum_tls_recv(conn, buffer + sizeof(wire_pkt),
		      *length - sizeof(wire_pkt));
	else
		ret = circum_net_recv(conn->fd, buffer + sizeof(wire_pkt),
		      *length - sizeof(wire_pkt));

	if (ret != *length - sizeof(wire_pkt)) {
		free(buffer);
		return NULL;
	}

	return buffer;
}

/**
 * a3_conn_recv - receive frame from network
 * @conn:	connection metadata structure and descriptor
 *
 * Receive a complete frame, unserialize it and return as a3_packet.
 * On error, %NULL is returned and errno may be set to non-zero.
 */
EXPORT_SYMBOL struct a3_packet *a3_conn_recv(struct a3_conn *conn)
{
	unsigned int length;
	void *buffer = circum_conn_recv_bpkt(conn, &length);
	struct a3_packet *pkt;
	int saved_errno;

	if (buffer == NULL)
		return NULL;
	if (circum_debug_flags & (CIRCUM_DUMP_TREE | CIRCUM_DUMP_WPKT))
		printf("<< RECV\n");
	if (circum_debug_flags & CIRCUM_DUMP_WPKT)
		circum_hexdump(buffer, length);
	pkt = a3_pkt_unserialize(buffer, length);
	saved_errno = errno;
	free(buffer);
	if ((circum_debug_flags & CIRCUM_DUMP_TREE) && pkt != NULL)
		circum_treedump_pkt(pkt, 0);
	errno = saved_errno;
	return pkt;
}

/**
 * circum_net_send - send packet in blocking mode
 * @fd:		file descriptor
 * @buf:	buffer
 * @count:	size of data
 *
 * write()ing to a socket may return less bytes than requested, for example
 * due to fragmentation into TCP windows. Loop as long until all data has
 * been pushed out.
 * Retains all semantics of write().
 */
EXPORT_SYMBOL ssize_t circum_net_send(int fd, const void *buf, size_t count)
{
	size_t processed = 0;

	while (processed < count) {
		ssize_t ret = write(fd, buf + processed, count);
		if (ret < 0) {
			if (processed > 0)
				break;
			else
				return ret;
		}
		if (ret == 0)
			break;
		processed += ret;
	}
	return processed;
}

/**
 * a3_conn_send - serialize, send (and destroy)
 * @conn:	connection descriptor
 * @pkt:	packet to process
 *
 * Serializes the packet and sends it out.
 * On error, a negative code indicating errno is returned.
 * On success, the packet is freed.
 */
EXPORT_SYMBOL ssize_t a3_conn_send(struct a3_conn *conn, struct a3_packet *pkt)
{
	unsigned int wire_size;
	struct a3_avp *avp;
	void *wire_pkt;
	ssize_t ret = 0;

	/* Attention to the order - we are doing unshifts here */
	avp = a3_pkt_add_text(pkt, "Origin-Realm", conn->local_host ? : "");
	HXlist_del(&avp->list);
	HXlist_add(&pkt->avp_list.list, &avp->list);
	avp = a3_pkt_add_text(pkt, "Origin-Host", conn->local_host ? : "");
	HXlist_del(&avp->list);
	HXlist_add(&pkt->avp_list.list, &avp->list);

	if (pkt->cmdflags & A3_CMDFLAG_REQUEST)
		a3_pkt_add_text(pkt, "Destination-Host", conn->remote_host);
	if (conn->session_id != NULL)
		a3_pkt_add_text(pkt, "Session-Id", conn->session_id);

	if (circum_debug_flags & (CIRCUM_DUMP_TREE | CIRCUM_DUMP_WPKT))
		printf(">> SEND\n");
	if (circum_debug_flags & CIRCUM_DUMP_TREE)
		circum_treedump_pkt(pkt, 0);

	wire_pkt = a3_pkt_serialize(pkt, &wire_size);
	if (wire_pkt == NULL) {
		ret = -errno;
		a3_pkt_destroy(pkt);
		return ret;
	}

	if (circum_debug_flags & CIRCUM_DUMP_WPKT)
		circum_hexdump(wire_pkt, wire_size);

	ret = conn->tls_active ?
	      circum_tls_send(conn, wire_pkt, wire_size) :
	      circum_net_send(conn->fd, wire_pkt, wire_size);
	if (ret < 0)
		ret = -errno;
	free(wire_pkt);
	a3_pkt_destroy(pkt);
	return ret;
}

static bool circum_cap_exchange(struct a3_conn *conn)
{
	struct a3_packet *pkt;
	struct a3_avp *avp;
	bool success = false;

	pkt = a3_pkt_new();
	pkt->command  = a3_idmap_cmd("Capabilities-Exchange");
	pkt->cmdflags = A3_CMDFLAG_REQUEST;
	a3_pkt_add_text(pkt, "Host-IP-Address", conn->local_addr);
	a3_pkt_add_text(pkt, "Vendor-Id", "0");
	a3_pkt_add_text(pkt, "Product-Name", "Codename Circumference" OS_SUFFIX);
	a3_pkt_add_int(pkt, "Auth-Application-Id", conn->appl_id);

	a3_conn_send(conn, pkt);
	pkt = a3_conn_recv(conn);
	if (pkt == NULL)
		return false;
	if (pkt->command != a3_idmap_cmd("Capabilities-Exchange"))
		goto out;

	if ((avp = a3_pkt_find(pkt, "Result-Code")) != NULL &&
	    avp->d.v_uint32 == a3_idmap_enum("Result-Code", "Success"))
		success = true;

 out:
	a3_pkt_destroy(pkt);
	if (!success)
		return false;
	return true;
}

static inline uint64_t rand64(void)
{
	uint64_t x = 0;
	unsigned int i;
	for (i = 0; i < 4; ++i) {
		x <<= 16;
		x  ^= HX_rand();
	}
	return x;
}

/**
 * a3_session_genid - generate a session ID
 */
static char *a3_session_genid(const struct a3_conn *conn)
{
	unsigned int incr = HX_irand(1, 65);
	uint64_t seq;
	char buf[128+sizeof(";4294967296;4294967296;18446744073709551616")];

	/*
	 * Grab sequence counter, then increase by some amount,
	 * and quickly release the lock again.
	 */
	pthread_mutex_lock(&session_mctr_lock);
	seq = session_mctr;
	session_mctr += incr;
	pthread_mutex_unlock(&session_mctr_lock);

	snprintf(buf, sizeof(buf), "%s;%u;%u;%llu",
	         conn->local_host, static_cast(unsigned int, seq >> 32),
	         static_cast(unsigned int, seq & 0xFFFFFFFF),
	         static_cast(unsigned long long, rand64()));
	return HX_strdup(buf);
}

/**
 * circum_conn_open2 - construct metadata
 * @conn:	connection descriptor
 *
 * Build the metadata from the socket and the connected peer, like local
 * hostname and address.
 */
static bool circum_conn_open2(struct a3_conn *conn)
{
	/*
	 * After (automatic) bind(), we can retrieve our own outgoing
	 * address, and hence external hostname.
	 */
	union sockaddr_big skaddr;
	char node_name[128];
	socklen_t sksize = (conn->l3proto == PF_INET6) ?
	                   sizeof(skaddr.inet6) : sizeof(skaddr.inet4);

	memset(&skaddr, 0, sizeof(skaddr));
	if (getsockname(conn->fd, &skaddr.inet0, &sksize) < 0)
		return false;
	if (conn->l3proto == PF_INET6) {
		if (inet_ntop(conn->l3proto, &skaddr.inet6.sin6_addr,
		    node_name, sizeof(node_name)) < 0)
			return false;
	} else {
		if (inet_ntop(conn->l3proto, &skaddr.inet4.sin_addr,
		    node_name, sizeof(node_name)) < 0)
			return false;
	}
	if ((conn->local_addr = HX_strdup(node_name)) == NULL)
		return false;
	if (getnameinfo(static_cast(const void *, &skaddr.inet0), sksize,
	      node_name, sizeof(node_name), NULL, 0, 0) < 0)
		conn->local_host = conn->local_addr;
	else if ((conn->local_host = HX_strdup(node_name)) == NULL)
		return false;
	if (conn->session_id == NULL) {
		conn->session_id = a3_session_genid(conn);
		if (conn->session_id == NULL)
			return false;
	}

	return true;
}

/**
 * a3_conn_new - allocate data and wait
 *
 * Splitting the allocation from connecting allows for the user to set a
 * specific Session-Id to be reused.
 */
EXPORT_SYMBOL struct a3_conn *a3_conn_new(void)
{
	struct a3_conn *conn = calloc(1, sizeof(struct a3_conn));

	if (conn == NULL)
		return NULL;

	conn->fd = -1;
	return conn;
}

/**
 * circum_uri_breakdown - extract fields from URI
 */
static void circum_uri_breakdown(const char *uri, hxmc_t **host, hxmc_t **port,
    uint8_t *l4proto, bool *tls)
{
	const char *end;

	if (strncmp(uri, "aaa://", strlen("aaa://")) == 0) {
		uri += strlen("aaa://");
		*tls = false;
	} else if (strncmp(uri, "aaas://", strlen("aaas://")) == 0) {
		uri += strlen("aaas://");
		*tls = true;
	}

	if (*uri == '[' && (end = strchr(uri, ']')) != NULL) {
		++uri;
		*host = HXmc_meminit(uri, end - uri);
		uri = end + 1;
	} else {
		if ((end = strpbrk(uri, ":;")) == NULL)
			end = uri + strlen(uri);
		*host = HXmc_meminit(uri, end - uri);
		uri = end;
	}

	if (*uri == ':') {
		strtoul(++uri, const_cast(char **, &end), 10);
		*port = HXmc_meminit(uri, end - uri);
		uri = end;
	}

	while (*uri == ';') {
		const char *key = ++uri;
		const char *value = strchr(uri, '=');

		if (value == NULL) {
			if ((end = strchr(uri, ';')) == NULL) {
				end = uri + strlen(uri);
				continue;
			}
			uri = end;
			continue;
		}
		if ((end = strchr(value, ';')) == NULL)
			end = value + strlen(value);
		if (strncmp(key, "transport", value - key) == 0) {
			++value;
			if (strncmp(value, "tcp", end - value) == 0)
				*l4proto = IPPROTO_TCP;
			else if (strncmp(value, "sctp", end - value) == 0)
				*l4proto = IPPROTO_SCTP;
		}
		uri = end;
	}
}

/**
 * a3_conn_open - connect with remote host and do initial exchange
 */
EXPORT_SYMBOL bool a3_conn_open(struct a3_conn *conn, const char *uri)
{
	struct addrinfo remote_hint, *remote_addr = NULL;
	const struct addrinfo *remote_ptr;
	int saved_errno, ret;
	hxmc_t *host = NULL, *port = NULL;
	uint8_t l4proto = 0;
	bool tls_request = true; /* default to secure by default */

	circum_uri_breakdown(uri, &host, &port, &l4proto, &tls_request);

	memset(&remote_hint, 0, sizeof(remote_hint));
	remote_hint.ai_socktype = SOCK_STREAM;

	ret = getaddrinfo(host, (port != NULL) ? port : "3868",
	      &remote_hint, &remote_addr);
	if (ret != 0 || remote_addr == NULL) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(ret));
		goto out;
	}

	for (remote_ptr = remote_addr; remote_ptr != NULL;
	     remote_ptr = remote_ptr->ai_next)
	{
		conn->fd = socket(remote_ptr->ai_family,
		           remote_ptr->ai_socktype,
		           (l4proto != 0) ? l4proto : remote_ptr->ai_protocol);

		if (conn->fd < 0) {
			/* Ignore unavailable protocols */
			if (errno != EAFNOSUPPORT || errno != EINVAL ||
			    errno != EPROTONOSUPPORT)
				continue;
			goto out;
		}
		if (connect(conn->fd, static_cast(const void *,
		    remote_ptr->ai_addr), remote_ptr->ai_addrlen) == 0)
			/* Good connection */
			break;
		saved_errno = errno;
		close(conn->fd);
		conn->fd = -1;
		errno = saved_errno;
		if (errno != EAFNOSUPPORT && errno != ECONNREFUSED &&
		    errno != ENETUNREACH)
			/* Bail out if it was not due to unav. protocols */
			goto out;

		/* try next l3proto/l4proto */
	}

	if (conn->fd < 0)
		/* No more l3proto/l4proto to try */
		goto out;

	conn->l3proto = remote_ptr->ai_family;
	conn->remote_host = HX_strdup(host);
	if (tls_request)
		if (circum_conn_starttls(conn, true) <= 0) {
			errno = 0;
			goto out;
		}
	if (!circum_conn_open2(conn))
		goto out;
	HXmc_free(host);
	if (!circum_cap_exchange(conn)) {
		a3_conn_close(conn);
		errno = 0;
		freeaddrinfo(remote_addr);
		return false;
	}

	freeaddrinfo(remote_addr);
	return true;

 out:
	saved_errno = errno;
	close(conn->fd);
	if (remote_addr != NULL)
		freeaddrinfo(remote_addr);
	HXmc_free(host);
	HXmc_free(port);
	errno = saved_errno;
	return false;
}

/**
 * a3_conn_close - close connection and free data
 *
 * Sends the Diameter termination message over the channel and waits for the
 * response, then tears down the low-level socket.
 */
EXPORT_SYMBOL void a3_conn_close(struct a3_conn *conn)
{
	struct a3_packet *pkt;
	struct a3_avp *avp;

	pkt = a3_pkt_new();
	pkt->command  = a3_idmap_cmd("Session-Termination");
	pkt->cmdflags = A3_CMDFLAG_REQUEST;
	a3_pkt_add_text(pkt, "Destination-Realm", "");
	a3_pkt_add_int(pkt, "Auth-Application-Id", conn->appl_id);
	a3_pkt_add_enum(pkt, "Termination-Cause", "Logout");

	a3_conn_send(conn, pkt);

	pkt = a3_conn_recv(conn);
	if (pkt != NULL && (circum_debug_flags & CIRCUM_DEBUG_OTHER))
		if ((avp = a3_pkt_find(pkt, "Result-Code")) != NULL)
			fprintf(stderr, "%s: Termination code: %u\n",
			        __func__, avp->d.v_uint32);
	if (pkt != NULL)
		a3_pkt_destroy(pkt);

	circum_conn_leavetls(conn);
	circum_conn_free(conn);
}

EXPORT_SYMBOL void circum_conn_free(struct a3_conn *conn)
{
	free(conn->session_id);
	if (conn->local_addr != conn->local_host)
		free(conn->local_addr);
	free(conn->local_host);
	free(conn->remote_host);
	close(conn->fd);
	free(conn);
}

/**
 * circum_net_init - initialize network code
 *
 * This is called with the initializer lock held.
 */
void circum_net_init(void)
{
	unsigned int i;

	/*
	 * Kill highest bit to not wrap too early. RFC requires monotonically
	 * increasing counter, but does not mention wrap-around at all.
	 */
	session_mctr  = rand64();
	session_mctr &= ~(1ULL << 63);

	/* Used to generate a random invalid IPv6 address as identity */
	for (i = 0; i < 6; ++i)
		fakeident[i] = HX_rand();
}
