/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <arpa/inet.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/clist.h>
#include <libHX/defs.h>
#include <libHX/list.h>

#include <local/circum.h>
#include <local/config.h>
#include <libcircum/idmap.h>
#include <libcircum/internal.h>
#include <libcircum/packet.h>
#include <libcircum/wirepacket.h>

/**
 * circum_hton24 - convert a 24-bit entity from host to network order
 */
static inline unsigned int circum_hton24(unsigned int x)
{
#ifdef WORDS_BIGENDIAN
	return htonl(x);
#else
	return htonl(x) >> 8;
#endif
}

/**
 * circum_ntoh24 - convert a 24-bit entity from network to host order
 */
unsigned int circum_ntoh24(unsigned int x)
{
#ifdef WORDS_BIGENDIAN
	return ntohl(x);
#else
	return ntohl(x << 8);
#endif
}

static inline uint64_t circum_ntoh64(uint64_t x)
{
#ifdef WORDS_BIGENDIAN
	return x;
#else
	uint32_t lo = x, hi = x >> 32;
	return (static_cast(uint64_t, htonl(lo)) << 32) | htonl(hi);
#endif
}

#define circum_hton64(x) circum_ntoh64(x)

EXPORT_SYMBOL struct a3_packet *a3_pkt_new(void)
{
	struct a3_packet *pkt;

	pkt = calloc(1, sizeof(struct a3_packet));
	if (pkt == NULL)
		return NULL;

	HXclist_init(&pkt->avp_list);
	return pkt;
}

static void circum_avp_destroy(struct a3_avp *avp)
{
	struct a3_avp *subavp, *next_subavp;

	HXlist_for_each_entry_safe(subavp, next_subavp,
	    &avp->d.avp_list, list)
	{
		if (subavp->type == A3_AVPTYPE_GROUP)
			circum_avp_destroy(subavp);
		free(subavp);
	}
}

EXPORT_SYMBOL void a3_pkt_destroy(struct a3_packet *pkt)
{
	struct a3_avp *avp, *next_avp;

	HXlist_for_each_entry_safe(avp, next_avp, &pkt->avp_list, list) {
		if (avp->type == A3_AVPTYPE_GROUP)
			circum_avp_destroy(avp);
		free(avp);
	}
	free(pkt);
}

static const unsigned int circum_avp_typesize[] = {
	[A3_AVPTYPE_GROUP]    = sizeof(struct HXclist_head),
	[A3_AVPTYPE_INT32]    = sizeof(int32_t),
	[A3_AVPTYPE_UINT32]   = sizeof(uint32_t),
	[A3_AVPTYPE_INT64]    = sizeof(int64_t),
	[A3_AVPTYPE_UINT64]   = sizeof(uint64_t),
	[A3_AVPTYPE_FLOAT32]  = sizeof(float),
	[A3_AVPTYPE_FLOAT64]  = sizeof(double),
};

/**
 * a3_avp_new - create new AVP
 * @avp_id:	avp mapping structure (type and flags)
 * @data:	AVP contents
 * @length:	length of @data
 *
 * The AVP plus its data section is allocated as one block(!) whose size
 * depends on @avp_id->type. Adds the AVP to the packet's AVP list, and
 * returns a pointer to the AVP on success, otherwise %NULL.
 * AVP constraints, for example flags that must be raised according to
 * RFC, are set.
 * No padding is done.
 */
static struct a3_avp *a3_avp_new(const struct a3_idmap_avp *avp_id,
    const void *value, unsigned int length)
{
	unsigned int size = sizeof(struct a3_avp);
	struct a3_avp *avp;

	if (avp_id->type < ARRAY_SIZE(circum_avp_typesize))
		size += circum_avp_typesize[avp_id->type];
	else
		size += length + sizeof(unsigned int) + 1;

	avp = calloc(1, size);
	if (avp == NULL)
		return NULL;

	avp->code  = avp_id->code;
	avp->flags = avp_id->must;
	avp->type  = avp_id->type;

	switch (avp->type) {
	case A3_AVPTYPE_GROUP:
		HXclist_init(&avp->d.avp_list);
		break;
	case A3_AVPTYPE_INT32:
	case A3_AVPTYPE_UINT32:
		avp->d.v_int32 = strtol(value, NULL, 0);
		break;
	case A3_AVPTYPE_INT64:
	case A3_AVPTYPE_UINT64:
		avp->d.v_int64 = strtoll(value, NULL, 0);
		break;
	case A3_AVPTYPE_FLOAT32:
		avp->d.v_float32 = strtod(value, NULL);
		break;
	case A3_AVPTYPE_FLOAT64:
		avp->d.v_float64 = strtod(value, NULL);
		break;
	case A3_AVPTYPE_BLOB:
	case A3_AVPTYPE_STRING:
		avp->blob_length = length;
		memcpy(avp->blob, value, length);
		avp->blob[length] = '\0';
		break;
	default:
		fprintf(stderr, "%s: developer error\n", __func__);
		abort();
	}

	HXlist_init(&avp->list);
	return avp;
}

static const struct a3_idmap_avp a3_blob_avp = {
	.type = A3_AVPTYPE_BLOB,
};

/**
 * a3_pkt_add_avp - create AVP and add to packet (by name)
 * @pkt:	parent packet to assign with
 * @name:	the AVP's name
 * @data:	AVP contents
 * @length:	length of @data
 */
EXPORT_SYMBOL struct a3_avp *a3_pkt_add_avp(struct a3_packet *pkt,
    const char *name, const void *value, unsigned int length)
{
	const struct a3_idmap_avp *avp_id;
	struct a3_avp *avp;

	avp_id = circum_idmap_avp_name(name);
	if (avp_id == NULL)
		avp_id = &a3_blob_avp;

	avp = a3_avp_new(avp_id, value, length);
	if (avp == NULL)
		return NULL;
	HXclist_push(&pkt->avp_list, &avp->list);
	return avp;
}

EXPORT_SYMBOL struct a3_avp *a3_pkt_add_int(struct a3_packet *pkt,
    const char *name, long long value)
{
	char buf[sizeof("18446744073709551616")];
	snprintf(buf, sizeof(buf), "%lld", value);
	return a3_pkt_add_avp(pkt, name, buf, strlen(buf));
}

EXPORT_SYMBOL struct a3_avp *a3_pkt_add_enum(struct a3_packet *pkt,
    const char *name, const char *ename)
{
	return a3_pkt_add_int(pkt, name, a3_idmap_enum(name, ename));
}

/**
 * a3_pkt_add_text - short form of a3_pkt_add_avp
 * @pkt:	parent packet to assign with
 * @name:	the AVP's name
 * @value:	text string
 *
 * Use this function when you know that @value is a valid C string. For
 * binary data or when you already know the length, use a3_pkt_add_avp().
 */
EXPORT_SYMBOL struct a3_avp *a3_pkt_add_text(struct a3_packet *pkt,
    const char *name, const char *value)
{
	return a3_pkt_add_avp(pkt, name, value, strlen(value));
}

static void circum_assert_groupavp(const struct a3_avp *avp)
{
	if (avp->type != A3_AVPTYPE_GROUP) {
		fprintf(stderr, "%s: tried to add a subavp to a "
		        "non-group avp of type %u\n",
		        __func__, avp->type);
		abort();
	}
}

/**
 * a3_avp_add_avp - add a sub-AVP to a group AVP (by name)
 * @avp:	parent AVP to assign with
 * @name:	the AVP's symbolic name
 * @data:	AVP contents
 * @length:	length of @data
 */
EXPORT_SYMBOL struct a3_avp *a3_avp_add_avp(struct a3_avp *avp,
    const char *name, const void *value, unsigned int length)
{
	const struct a3_idmap_avp *avp_id;
	struct a3_avp *subavp;

	circum_assert_groupavp(avp);

	avp_id = circum_idmap_avp_name(name);
	if (avp_id == NULL)
		avp_id = &a3_blob_avp;

	subavp = a3_avp_new(avp_id, value, length);
	if (subavp == NULL)
		return NULL;
	HXclist_push(&avp->d.avp_list, &subavp->list);
	return subavp;
}

/**
 * a3_avp_add_text - short form of a3_avp_add_avp
 * @avp:	parent AVP to assign with
 * @name:	the AVP's name
 * @value:	text string
 *
 * Use this function when you know that @value is a valid C string. For
 * binary data or when you already know the length, use a3_pkt_add_avp().
 */
EXPORT_SYMBOL struct a3_avp *a3_avp_add_text(struct a3_avp *avp,
    const char *name, const char *value)
{
	return a3_avp_add_avp(avp, name, value, strlen(value));
}

EXPORT_SYMBOL struct a3_avp *a3_avp_add_int(struct a3_avp *avp,
    const char *name, long long value)
{
	char buf[sizeof("18446744073709551616")];
	snprintf(buf, sizeof(buf), "%lld", value);
	return a3_avp_add_avp(avp, name, buf, strlen(buf));
}

EXPORT_SYMBOL struct a3_avp *a3_avp_add_enum(struct a3_avp *avp,
    const char *name, const char *ename)
{
	return a3_avp_add_int(avp, name, a3_idmap_enum(name, ename));
}

/**
 * circum_avp_search - search for a specific AVP in a packet/AVP group
 * @list:	Linked list of AVPs
 * @name:	Name of AVP
 *
 * Returns a pointer to the first AVP found, or NULL on error/not found.
 */
static struct a3_avp *circum_avp_search(const struct HXclist_head *list,
    const char *name)
{
	const struct a3_idmap_avp *id = circum_idmap_avp_name(name);
	struct a3_avp *avp;

	if (id == NULL) {
		errno = ENOENT;
		return NULL;
	}

	HXlist_for_each_entry(avp, list, list)
		if (avp->code == id->code)
			return avp;
	errno = 0;
	return NULL;
}

/**
 * a3_pkt_find - find specific AVP in packet
 * @pkt:	pointer to packet
 * @name:	AVP name
 *
 * Returns a pointer to the first AVP found, or NULL on error/not found.
 */
EXPORT_SYMBOL struct a3_avp *a3_pkt_find(const struct a3_packet *pkt,
    const char *name)
{
	return circum_avp_search(&pkt->avp_list, name);
}

/**
 * a3_avp_find - find specific sub-AVP in AVP group
 * @pkt:	pointer to packet
 * @name:	AVP name
 *
 * Returns a pointer to the first AVP found, or NULL on error/not found.
 */
EXPORT_SYMBOL struct a3_avp *a3_avp_find(const struct a3_avp *avp,
    const char *name)
{
	if (avp->type != A3_AVPTYPE_GROUP) {
		errno = EINVAL;
		return NULL;
	}
	return circum_avp_search(&avp->d.avp_list, name);
}

static inline unsigned int round_up(unsigned int x, unsigned int n)
{
	return ((x + n - 1) & ~(n - 1));
}

/**
 * circum_avp_size - calculate size of the AVP including all sub-AVPs
 * @avp:	pointer to the AVP
 *
 * Returns the number of bytes the wire version of the AVP would occupy,
 * including appropriate padding.
 */
static unsigned int circum_avp_size(const struct a3_avp *avp)
{
	const struct a3_avp *subavp;
	unsigned int z;

	if (avp->flags & A3_AVPFLAG_VENDOR)
		z = sizeof(struct a3_wire_avp_vendor);
	else
		z = sizeof(struct a3_wire_avp);

	if (avp->type == A3_AVPTYPE_GROUP) {
		HXlist_for_each_entry(subavp, &avp->d.avp_list, list)
			z += circum_avp_size(subavp);
	} else if (avp->type == A3_AVPTYPE_BLOB ||
	    avp->type == A3_AVPTYPE_STRING) {
		z += round_up(avp->blob_length, sizeof(int32_t));
	} else if (avp->type < ARRAY_SIZE(circum_avp_typesize)) {
		z += circum_avp_typesize[avp->type];
	} else {
		fprintf(stderr, "%s: illegal type %u\n", __func__, avp->type);
		abort();
	}

	return z;
}

/**
 * circum_pkt_size - calculate size of packet including all AVPs
 * @pkt:	packet to process
 *
 * Returns the number of bytes the wire version of the packet would occupy,
 * including appropriate padding.
 */
static unsigned int circum_pkt_size(const struct a3_packet *pkt)
{
	const struct a3_avp *avp;
	unsigned int z = 0;

	z += sizeof(struct a3_wire_packet);
	HXlist_for_each_entry(avp, &pkt->avp_list, list)
		z += circum_avp_size(avp);

	return z;
}

/**
 * circum_avp_serialize_into - transform a3_avp into byte stream
 * @here:	pointer to memory block to write to
 * @avp:	pointer to the AVP
 *
 * Returns the new @here pointer for the caller.
 * It is assumed that the memory block given by @here is big enough,
 * i.e. calculate_size() was used appropriately before.
 */
static void *circum_avp_serialize_into(void *here, const struct a3_avp *avp)
{
	/*
	 * Note the cheating... always using the +vendor mask,
	 * but only writing to it when there is a vendor flag.
	 */
	struct a3_wire_avp_vendor *wire_avp = here;
	struct a3_avp *subavp;
	unsigned int length = 0;

	if (avp->flags & A3_AVPFLAG_VENDOR)
		here += sizeof(struct a3_wire_avp_vendor);
	else
		here += sizeof(struct a3_wire_avp);

	wire_avp->code  = htonl(avp->code);
	wire_avp->flags = avp->flags;

	if (avp->type == A3_AVPTYPE_GROUP) {
		void *const orig_here = here;
		HXlist_for_each_entry(subavp, &avp->d.avp_list, list)
			here = circum_avp_serialize_into(here, subavp);
		length = here - orig_here;
	} else if (avp->type == A3_AVPTYPE_BLOB ||
	    avp->type == A3_AVPTYPE_STRING) {
		length = round_up(avp->blob_length, sizeof(int32_t));
		memset(here, 0, length);
		memcpy(here, avp->blob, avp->blob_length);
		here  += length;
		length = avp->blob_length;
	} else if (avp->type == A3_AVPTYPE_INT32 ||
	    avp->type == A3_AVPTYPE_UINT32) {
		/*
		 * Type-punning will work here, since all parts are always
		 * aligned at least a 32-bit boundary.
		 */
		*(uint32_t *)here = htonl(avp->d.v_uint32);
		length = circum_avp_typesize[avp->type]; /* need for later too */
		here += length;
	} else if (avp->type == A3_AVPTYPE_INT64 ||
	    avp->type == A3_AVPTYPE_UINT64) {
		/*
		 * Type-punning will not work here, since it may not be
		 * on a 64-bit boundary. Need memcpy().
		 */
		uint64_t tmp = circum_hton64(avp->d.v_uint64);
		length = circum_avp_typesize[avp->type];
		memcpy(here, &tmp, length);
		here += length;
	} else if (avp->type < ARRAY_SIZE(circum_avp_typesize)) {
		length = circum_avp_typesize[avp->type];
		memcpy(here, &avp->d, length);
		here += length;
	} else {
		fprintf(stderr, "%s: illegal avp type %u\n",
		        __func__, avp->type);
		abort();
	}

	if (avp->flags & A3_AVPFLAG_VENDOR) {
		length += sizeof(struct a3_wire_avp_vendor);
		wire_avp->vendor_id = avp->vendor_id;
	} else {
		length += sizeof(struct a3_wire_avp);
	}

	wire_avp->length = circum_hton24(length);
	return here;
}

/**
 * a3_pkt_serialize - transform an a3_packet into byte stream
 * @pkt:	packet to serialize
 * @wblk_size:	pointer where the resulting size of the block is stored
 *
 * Allocates a buffer big enough to hold the byte stream and returns it.
 * The size is returned in @wblk_size. On failure, returns %NULL.
 */
EXPORT_SYMBOL void *a3_pkt_serialize(struct a3_packet *pkt,
    unsigned int *wblk_size)
{
	struct a3_wire_packet *wire_pkt;
	void *wire_block, *wire_ptr;
	struct a3_avp *avp;

	*wblk_size = circum_pkt_size(pkt);
	wire_block = malloc(*wblk_size);
	if (wire_block == NULL)
		return NULL;

	/* For debugging purposes */
	memset(wire_block, 0xDB, *wblk_size);

	wire_ptr           = wire_block;
	wire_pkt           = wire_ptr;
	wire_pkt->version  = DIAMETER_VERSION;
	wire_pkt->cmdflags = pkt->cmdflags;
	wire_pkt->length   = circum_hton24(*wblk_size);
	wire_pkt->command  = circum_hton24(pkt->command);
	wire_pkt->appl_id  = htonl(pkt->appl_id);
	wire_pkt->hbh_id   = htonl(pkt->hbh_id);
	wire_pkt->ete_id   = htonl(pkt->ete_id);

	wire_ptr += sizeof(struct a3_wire_packet);
	HXlist_for_each_entry(avp, &pkt->avp_list, list)
		wire_ptr = circum_avp_serialize_into(wire_ptr, avp);

	if (wire_ptr != wire_block + *wblk_size) {
		fprintf(stderr, "%s: failed wblk sanity check: %d bytes "
		        "written, expected %u\n", __func__,
		        (int)(wire_ptr - wire_block), *wblk_size);
		abort();
	}

	return wire_block;
}

/**
 * circum_avp_unserialize - transform an AVP byte stream into an a3_avp
 * @pkt:	decomposed packet
 * @wire_vpkt:	AVP in wire representation
 *
 * Returns a struct a3_avp, or %NULL on failure. @adv_ptr will contain the
 * value the pointer into the wire packet byte stream needs to be advanced.
 */
static struct a3_avp *circum_avp_unserialize(const void *wire_vpkt,
    unsigned int *adv_ptr)
{
	const struct a3_wire_avp *wire_avp = wire_vpkt;
	const struct a3_idmap_avp *avp_id;
	unsigned int splice_size, seg;
	const void *splice_ptr;
	struct a3_avp *avp;

	seg      = circum_ntoh24(wire_avp->length);
	*adv_ptr = round_up(seg, sizeof(int32_t));

	if (wire_avp->flags & A3_AVPFLAG_VENDOR) {
		splice_ptr  = wire_vpkt + sizeof(struct a3_wire_avp_vendor);
		splice_size = seg - sizeof(struct a3_wire_avp_vendor);
		wire_vpkt  += sizeof(struct a3_wire_avp_vendor);
	} else {
		splice_ptr  = wire_vpkt + sizeof(struct a3_wire_avp);
		splice_size = seg - sizeof(struct a3_wire_avp);
		wire_vpkt  += sizeof(struct a3_wire_avp);
	}

	avp = a3_avp_new(&a3_blob_avp, splice_ptr, splice_size);
	if (avp == NULL)
		return NULL;

	avp->code  = ntohl(wire_avp->code);
	avp->flags = wire_avp->flags;
	avp_id = circum_idmap_avp_code(avp->code);
	if (avp_id != NULL) {
		avp->type = avp_id->type;

		switch (avp->type) {
		case A3_AVPTYPE_INT32:
		case A3_AVPTYPE_UINT32:
			if (splice_size < sizeof(int32_t))
				break;
			avp->d.v_uint32 = ntohl(*(const int32_t *)avp->blob);
			break;
		case A3_AVPTYPE_INT64:
		case A3_AVPTYPE_UINT64:
			if (splice_size < sizeof(int64_t))
				break;
			memcpy(&avp->d.v_uint64, avp->blob, sizeof(uint64_t));
			avp->d.v_uint64 = circum_ntoh64(avp->d.v_uint64);
			break;
		case A3_AVPTYPE_FLOAT32:
			if (splice_size >= sizeof(float))
				avp->d.v_float32 = *(const float *)avp->blob;
			break;
		case A3_AVPTYPE_FLOAT64:
			if (splice_size >= sizeof(double))
				avp->d.v_float64 = *(const double *)avp->blob;
			break;
		case A3_AVPTYPE_GROUP: {
			int len = splice_size;
			struct a3_avp *subavp;

			HXclist_init(&avp->d.avp_list);

			while (len > 0) {
				/* reuse @seg */
				subavp = circum_avp_unserialize(wire_vpkt, &seg);
				if (subavp == NULL)
					goto out;
				HXclist_push(&avp->d.avp_list, &subavp->list);
				len -= seg;
				wire_vpkt += seg;
			}
			if (len < 0)
				/* Broken packet */
				goto out;
			break;
		}
		default:
			break;
		}
	}

	return avp;
 out:
	circum_avp_destroy(avp);
	return NULL;
}

/**
 * a3_pkt_unserialize - transform a byte stream into an a3_packet
 * @wire_vpkt:	Diameter message in wire representation
 * @wire_size:	size of memory block
 *
 * Decompose the network byte stream into the well-known struct a3_packet
 * and convert endianess, etc.
 */
EXPORT_SYMBOL struct a3_packet *a3_pkt_unserialize(const void *wire_vpkt,
    unsigned int wire_size)
{
	const struct a3_wire_packet *wire_pkt = wire_vpkt;
	struct a3_packet *pkt;
	unsigned int len;

	len = circum_ntoh24(wire_pkt->length);
	if (len != wire_size)
		fprintf(stderr, "%s: WARNING: length %u != wire_size %u\n",
		        __func__, len, wire_size);

	pkt = a3_pkt_new();
	if (pkt == NULL)
		return NULL;
	pkt->cmdflags = wire_pkt->cmdflags;
	pkt->command  = circum_ntoh24(wire_pkt->command);
	pkt->appl_id  = ntohl(wire_pkt->appl_id);
	pkt->hbh_id   = ntohl(wire_pkt->hbh_id);
	pkt->ete_id   = ntohl(wire_pkt->ete_id);

	len -= sizeof(struct a3_wire_packet);
	wire_vpkt += sizeof(struct a3_wire_packet);

	while (len > 0) {
		struct a3_avp *avp;
		unsigned int seg;

		avp = circum_avp_unserialize(wire_vpkt, &seg);
		if (avp == NULL)
			goto out;
		HXclist_push(&pkt->avp_list, &avp->list);
		len -= seg;
		wire_vpkt += seg;
	}

	return pkt;

 out:
	a3_pkt_destroy(pkt);
	return NULL;
}
