#ifndef _LIBCIRCUMFERENCE_PACKET_H
#define _LIBCIRCUMFERENCE_PACKET_H 1

#ifdef __cplusplus
#	include <cstdint>
#else
#	include <stdbool.h>
#	include <stdint.h>
#endif
#include <libHX/clist.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Command flags as per RFC 3588 Update, section 3.
 *
 * %A3_CMDFLAG_REQUEST:
 * 	"The message is a request."
 * %A3_CMDFLAG_PROXYABLE:
 * 	"The message may be proxied, relayed or redirected."
 * %A3_CMDFLAG_ERROR:
 * 	"The message contains a protocol error."
 * %A3_CMDFLAG_RETRANS:
 * 	"This flag is set after a link failover procedure, to aid the
 * 	removal of duplicate requests."
 * %A3_CMDFLAG_RESERVED:
 * 	Bitmask of reserved flags.
 */
enum {
	A3_CMDFLAG_REQUEST   = 1 << 7,
	A3_CMDFLAG_PROXYABLE = 1 << 6,
	A3_CMDFLAG_ERROR     = 1 << 5,
	A3_CMDFLAG_RETRANS   = 1 << 4,
	A3_CMDFLAG_RESERVED  = 0xF,
};

/**
 * Command flags as per RFC 3588 Update, section 4.1.
 *
 * %A3_AVPFLAG_VENDOR:
 * 	The optional Vendor-ID field is present.
 * %A3_AVPFLAG_MANDATORY:
 * 	Whether support for the AVP is required.
 * %A3_AVPFLAG_PROTECT:
 * 	Need end-to-end security.
 * %A3_AVPFLAG_RESERVED:
 * 	Bitmask of reserved flags.
 */
enum {
	A3_AVPFLAG_VENDOR    = 1 << 7,
	A3_AVPFLAG_MANDATORY = 1 << 6,
	A3_AVPFLAG_PROTECT   = 1 << 5,
	A3_AVPFLAG_RESERVED  = 0x1F,
};

/**
 * a3_packet - in-memory representation (i.e. with control structures)
 * of an AAA packet
 *
 * @cmdflags:	command flags (see %A3_CMDFLAG_*)
 * @command:	command number (see %A3_CMD_*)
 * @appl_id:	application number
 * @hbh_id:	hop-by-hop identification number
 * @ete_id:	end-to-end identification number
 * @avp_list:	list of AVPs for this packet
 */
struct a3_packet {
	uint8_t cmdflags;
	uint32_t command;
	uint32_t appl_id;
	uint32_t hbh_id;
	uint32_t ete_id;
	struct HXclist_head avp_list;
};

/**
 * a3_avp_type - denotes the internel representation of an AVP
 *
 * %A3_AVPTYPE_GROUP:		Grouped AVP
 * %A3_AVPTYPE_INT32:		32-bit signed integer
 * %A3_AVPTYPE_UINT32:		32-bit unsigned integer
 * %A3_AVPTYPE_INT64:		64-bit signed integer
 * %A3_AVPTYPE_UINT64:		64-bit unsigned integer
 * %A3_AVPTYPE_FLOAT32:		"single precision",
 * 				32-bit IEEE-754 floating point number
 * %A3_AVPTYPE_FLOAT64:		"double precision",
 * 				64-bit IEEE-754 floating point number
 * %A3_AVPTYPE_TIME:		Alias for A3_AVPTYPE_UINT32.
 * %A3_AVPTYPE_ENUM:		Alias for A3_AVPTYPE_UINT32.
 * %A3_AVPTYPE_STRING:		String.
 * %A3_AVPTYPE_STRING7B:	(7-bit) ASCII string.
 * 				Alias for A3_AVPTYPE_STRING.
 * %A3_AVPTYPE_BLOB:		Raw blob.
 * %A3_AVPTYPE_DIAMIDENT:	DiameterIdent (see RFC).
 * 				Alias for A3_AVPTYPE_STRING.
 * %A3_AVPTYPE_DIAMURI:		DiameterURI (see RFC).
 * 				Alias for A3_AVPTYPE_STRING.
 *
 * The a3_avp_type, within struct a3_avp, gives hint about the interpretation
 * of the blob, and what member in the union "d" to access.
 * Some of these are just aliases, because e.g. ENUM was defined to be an
 * INT32 in the RFC for some reason.
 */
enum a3_avp_type {
	/* Internal */
	A3_AVPTYPE_GROUP,

	/* Types of fixed size */
	A3_AVPTYPE_INT32,
	A3_AVPTYPE_UINT32,
	A3_AVPTYPE_INT64,
	A3_AVPTYPE_UINT64,
	A3_AVPTYPE_FLOAT32,
	A3_AVPTYPE_FLOAT64,

	A3_AVPTYPE_TIME     = A3_AVPTYPE_UINT32,
	A3_AVPTYPE_ENUM     = A3_AVPTYPE_UINT32,

	/* Dynamic-sized types */
	A3_AVPTYPE_STRING    = 0x8000,
	A3_AVPTYPE_STRING7B  = A3_AVPTYPE_STRING,
	A3_AVPTYPE_BLOB,
	A3_AVPTYPE_DIAMIDENT = A3_AVPTYPE_STRING,
	A3_AVPTYPE_DIAMURI   = A3_AVPTYPE_STRING,
	__A3_AVPTYPE_DYN     = 0x8000,
};

/**
 * a3_avp_morph - polymorphic structure for interpreted types
 */
union a3_avp_morph {
	struct HXclist_head avp_list;
	int32_t v_int32;
	uint32_t v_uint32;
	int64_t v_int64;
	uint64_t v_uint64;
	float v_float32;
	double v_float64;
};

/**
 * a3_avp - in-memory representation (i.e. with control structures) of an AVP
 * @list:		linked ring of AVPs belonging to the packet
 * @code:		AVP code
 * @flags:		AVP flags (see %A3_AVPFLAG_*)
 * @type:		AVP type (see %A3_AVPTYPE_*)
 * @vendor_id:
 * @d:			interpreted blob, depending @type
 * @blob_length:	length of @blob
 * @blob:		original blob representation
 */
struct a3_avp {
	struct HXlist_head list;
	uint32_t code;
	uint8_t flags;
	enum a3_avp_type type;
	uint32_t vendor_id;

	union a3_avp_morph d;
	unsigned int blob_length;
	char blob[0];
};

/* init.c */
extern bool a3_initialize(void);
extern void a3_deinitialize(void);

/* packet.c */
extern struct a3_packet *a3_pkt_new(void);
extern void a3_pkt_destroy(struct a3_packet *);

extern struct a3_avp *a3_pkt_add_avp(struct a3_packet *,
	const char *, const void *, unsigned int);
extern struct a3_avp *a3_pkt_add_int(struct a3_packet *,
	const char *, long long);
extern struct a3_avp *a3_pkt_add_enum(struct a3_packet *,
	const char *, const char *);
extern struct a3_avp *a3_pkt_add_text(struct a3_packet *,
	const char *, const char *);
extern struct a3_avp *a3_avp_add_avp(struct a3_avp *,
	const char *, const void *, unsigned int);
extern struct a3_avp *a3_avp_add_text(struct a3_avp *,
	const char *, const char *);
extern struct a3_avp *a3_avp_add_int(struct a3_avp *,
	const char *, long long);
extern struct a3_avp *a3_avp_add_enum(struct a3_avp *,
	const char *, const char *);

extern struct a3_avp *a3_pkt_find(const struct a3_packet *, const char *);
extern struct a3_avp *a3_avp_find(const struct a3_avp *, const char *);

extern void *a3_pkt_serialize(struct a3_packet *, unsigned int *);
extern struct a3_packet *a3_pkt_unserialize(const void *, unsigned int);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _LIBCIRCUMFERENCE_PACKET_H */
