/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <errno.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <local/config.h>
#include <local/circum.h>
#include <openssl/ssl.h>
#include "internal.h"
#include "net.h"

struct circum_tls_data {
	SSL *ssl;
	BIO *bio;
};

static pthread_mutex_t circum_tlsserver_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t circum_tlsclient_lock = PTHREAD_MUTEX_INITIALIZER;
static SSL_CTX *circum_tlsserver_ctx, *circum_tlsclient_ctx;
static bool circum_tls_inited, circum_tlsserver_inited, circum_tlsclient_inited;

/**
 * circum_tls_init - initialize OpenSSL's TLS
 *
 * Called while the initialization lock is held.
 */
static void circum_tls_init(void)
{
	pthread_mutex_lock(&circum_tlsserver_lock);
	pthread_mutex_lock(&circum_tlsclient_lock);
	if (!circum_tls_inited) {
		SSL_library_init();
		SSL_load_error_strings();
		circum_tls_inited = true;
	}
	pthread_mutex_unlock(&circum_tlsclient_lock);
	pthread_mutex_unlock(&circum_tlsserver_lock);
}

static bool circum_tlscommon_init(SSL_METHOD *meth, pthread_mutex_t *lock,
    bool *initp, SSL_CTX **ctx, const char *certfile, const char *keyfile)
{
	bool ret = false;

	if (!circum_tls_inited)
		circum_tls_init();

	pthread_mutex_lock(lock);
	if (*initp) {
		ret = true;
		goto out;
	}
	*ctx = SSL_CTX_new(meth);

	if (!SSL_CTX_use_certificate_file(*ctx, certfile, SSL_FILETYPE_PEM)) {
		fprintf(stderr, "Could not read certificate %s\n", certfile);
		goto out;
	}
	if (!SSL_CTX_use_PrivateKey_file(*ctx, keyfile, SSL_FILETYPE_PEM)) {
		fprintf(stderr, "Could not read key file %s\n", keyfile);
		goto out;
	}
	*initp = true;
	ret    = true;

 out:
	pthread_mutex_unlock(lock);
	return ret;
}

static bool circum_tlsserver_init(void)
{
	return circum_tlscommon_init(TLSv1_server_method(),
	       &circum_tlsserver_lock, &circum_tlsserver_inited,
	       &circum_tlsserver_ctx, "circumd.crt", "circumd.key");
}

static bool circum_tlsclient_init(void)
{
	return circum_tlscommon_init(TLSv1_client_method(),
	       &circum_tlsclient_lock, &circum_tlsclient_inited,
	       &circum_tlsclient_ctx, "libcircum.crt", "libcircum.key");
}

/**
 * circum_conn_verifycert - verify remote certificate
 * @ssl:		OpenSSL private data
 * @remote_host:	Remote host name we expect in the certificate
 * @client:		We are client
 */
static bool circum_conn_verifycert(SSL *ssl, const char *remote_host,
    bool client)
{
	char peer_cn[256];
	X509 *peer;
	int ret;

	ret = SSL_get_verify_result(ssl);
	if (ret == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) {
		fprintf(stderr, "TLS: warning: other side "
		        "uses a self-signed certificate\n");
	} else if (ret != X509_V_OK) {
		fprintf(stderr, "TLS: certificate does not verify\n");
		return false;
	}

	peer = SSL_get_peer_certificate(ssl);
	if (peer == NULL) {
		fprintf(stderr, "TLS: remote side provided no certificate%s\n",
		        client ? "" : " (ignored)");
		/* It is ok for the server that the client has no cert. */
		return !client;
	}

	X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName,
		peer_cn, sizeof(peer_cn));
	if (strcasecmp(peer_cn, remote_host) != 0) {
		fprintf(stderr, "TLS: peer's common name in the certificate "
		        "(%s) does not expected hostname (%s)\n",
		        peer_cn, remote_host);
		/* return false; */ /* ignore for now */
	}
	return true;
}

/**
 * circum_conn_starttls - start TLS on a connection
 * @conn:	connection
 * @client:	initialize for client or server
 *
 * This function is only exported so that circumd can reuse code easily.
 *
 * Returns positive on success, or a negative error code.
 */
EXPORT_SYMBOL bool circum_conn_starttls(struct a3_conn *conn, bool client)
{
	struct circum_tls_data *td;
	int saved_errno;

	if (client) {
		if (!circum_tlsclient_inited && !circum_tlsclient_init())
			return -EINVAL;
	} else {
		if (!circum_tlsserver_inited && !circum_tlsserver_init())
			return -EINVAL;
	}

	if ((td = malloc(sizeof(struct circum_tls_data))) == NULL) {
		fprintf(stderr, "%s: malloc: %s\n", __func__, strerror(errno));
		goto out;
	}

	conn->tls_data = td;
	td->ssl = SSL_new(client ? circum_tlsclient_ctx : circum_tlsserver_ctx);
	if (td->ssl == NULL)
		goto out;
	td->bio = BIO_new_socket(conn->fd, BIO_NOCLOSE);
	if (td->bio == NULL)
		goto out;

	SSL_set_bio(td->ssl, td->bio, td->bio);
	if (client) {
		if (SSL_connect(td->ssl) <= 0) {
			fprintf(stderr, "SSL connect error\n");
			return 0;
		}
	} else {
		if (SSL_accept(td->ssl) <= 0) {
			fprintf(stderr, "SSL accept error\n");
			return 0;
		}
	}
	conn->tls_active = true;
	return circum_conn_verifycert(td->ssl, conn->remote_host, client);

 out:
	saved_errno = errno;
	if (td != NULL) {
		/* need to close td->bio too? */
		if (td->ssl != NULL) {
			SSL_shutdown(td->ssl);
			SSL_free(td->ssl);
		}
		free(td);
	}
	errno = saved_errno;
	return -errno;
}

EXPORT_SYMBOL void circum_conn_leavetls(struct a3_conn *conn)
{
	struct circum_tls_data *td = conn->tls_data;

	conn->tls_active = false;
	conn->tls_data   = NULL;
	if (td != NULL) {
		SSL_shutdown(td->ssl);
		SSL_free(td->ssl);
		free(td);
	}
}

/**
 * circum_tls_recv - receive data through secured channel
 * @conn:	circum connection
 * @buffer:	buffer for incoming packet
 * @count:	buffer size
 */
ssize_t circum_tls_recv(const struct a3_conn *conn, void *buffer, size_t count)
{
	const struct circum_tls_data *td = conn->tls_data;
	const size_t orig_count = count;
	int err, ret;

	while (count > 0) {
		ret = SSL_read(td->ssl, buffer, count);
		err = SSL_get_error(td->ssl, ret);
		if (err == SSL_ERROR_ZERO_RETURN)
			return 0;
		if (err == SSL_ERROR_NONE) {
			count  -= ret;
			buffer += ret;
			continue;
		}
		fprintf(stderr, "TLS: read problem (%d)\n", ret);
		return -EPIPE;
	}

	return orig_count;
}

/**
 * circum_tls_send - send data through secured channel
 * @conn:	circum connection
 * @buffer:	buffer to send (usually serialized packet)
 * @count:	amount to send
 */
ssize_t circum_tls_send(const struct a3_conn *conn, const void *buffer,
    size_t count)
{
	const struct circum_tls_data *td = conn->tls_data;
	const size_t orig_count = count;
	int err, ret;

	while (count > 0) {
		ret = SSL_write(td->ssl, buffer, count);
		err = SSL_get_error(td->ssl, ret);
		if (err == SSL_ERROR_NONE) {
			count  -= ret;
			buffer += ret;
			continue;
		}
		fprintf(stderr, "TLS: write problem (%d)\n", ret);
		return -EPIPE;
	}

	return orig_count;
}

void circum_tls_exit(void)
{
	pthread_mutex_lock(&circum_tlsclient_lock);
	SSL_CTX_free(circum_tlsclient_ctx);
	circum_tlsclient_inited = false;
	pthread_mutex_unlock(&circum_tlsclient_lock);

	pthread_mutex_lock(&circum_tlsserver_lock);
	SSL_CTX_free(circum_tlsserver_ctx);
	circum_tlsserver_inited = false;
	pthread_mutex_unlock(&circum_tlsserver_lock);
}
