/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <stdbool.h>
#include <stdio.h>
#include <libHX/clist.h>
#include <libHX/list.h>
#include <libHX/libxml_helper.h>

#include <local/circum.h>
#include "idmap.h"
#include "internal.h"
#include "packet.h"

enum a3_rule_type {
	/* bits 0,1 */
	RULE_FIXED    = 0x01,
	RULE_REQUIRED = 0x02,
	RULE_OPTIONAL = 0x03,

	RULE_FRO_MASK = 0x03,

	/* bit 2 */
	RULE_REQUEST  = 1 << 2,
	RULE_ANSWER   = 0,
};

/* Manual traversers, split up, from <libHX/list.h> */
#define mt_first(pos, head, member) \
	(pos) = HXlist_entry((head)->next, typeof(*(pos)), member)
#define mt_eol(pos, head, member) \
	(&(pos)->member == (void *)(head))
#define mt_next(pos, head, member) \
	(pos) = HXlist_entry((pos)->member.next, typeof(*(pos)), member)

/* Forward decls */
static const struct circum_xmltag_map
	circum_dict_vfyavp_tags[], circum_dict_vfycmd_tags[];

/**
 * circum_pkt_avpcluster - cluster AVPs for easy verification
 * @pkt:	packet to modify
 *
 * "Sorts" the AVPs within a packet to make verification of contiguous
 * sequences easier.
 *
 * Worst-case runtime: (n^2 - n) / 2
 * When there is only one of each AVP type and nothing to cluster.
 */
static void circum_pkt_avpcluster(struct a3_packet *pkt)
{
	struct a3_avp *avp, *next_avp, *seek_avp, *seek_next_avp;

	for (mt_first(avp, &pkt->avp_list, list);
	     !mt_eol(avp, &pkt->avp_list, list);
	     mt_next(avp, &pkt->avp_list, list))
	{
		/* Consider e.g. "AAKWABKWA" -> "AAAAKKWWB" */

		/* Go to end of a contiguous sequence */
		for (next_avp = avp; ;) {
			mt_next(next_avp, &pkt->avp_list, list);
			if (mt_eol(next_avp, &pkt->avp_list, list))
				return;
			if (next_avp->code != avp->code)
				break;
			avp = next_avp;
		}

		/*
		 * @avp will be at the end of the contiguous sequence
		 * (e.g. "AA"). @next_avp will point to the next sequence
		 * (e.g. 'K'). All further 'A's will now be sorted so that
		 * they appear after @avp but before @next_avp.
		 *
		 * Since it is known that @next_avp cannot be an 'A',
		 * @seek_avp will start at the AVP following 'K'...
		 */
		for (seek_avp = next_avp,
		     mt_next(seek_avp, &pkt->avp_list, list);
		     !mt_eol(seek_avp, &pkt->avp_list, list);
		     mt_next(seek_avp, &pkt->avp_list, list))
		{
			/* ...and skip until we actually find a friendly AVP */
			if (seek_avp->code != avp->code)
				continue;

			/*
			 * The AVP at @seek_avp is going to be moved.
			 * Set up @seek_next_avp to repickup traversal later
			 * (at the 'B' in /WAB/).
			 */
			seek_next_avp = seek_avp;
			mt_next(seek_next_avp, &pkt->avp_list, list);

			HXlist_del(&seek_avp->list);
			HXlist_add(&avp->list, &seek_avp->list);
			avp = seek_avp;

			/* Now got something like "AAAKWBKWA" */
			if (mt_eol(seek_next_avp, &pkt->avp_list, list))
				break;
			seek_avp = seek_next_avp;
		}
		/* Now having "AAAAKWBKW" */
	}
}

/**
 * a3_pkt_verify - verify correct layout of packet
 * @pkt:	packet to inspect
 *
 * Verifies that all AVPs meet the specific layout rules set forth in the
 * dictionary. Returns >0 for success, 0 for failure and -1 for inability to
 * determine (i.e. when rules are missing).
 */
EXPORT_SYMBOL int a3_pkt_verify(struct a3_packet *pkt)
{
	const struct HXclist_head *rl_fixed, *rl_var;
	const struct a3_idmap_cmd *cmd_id;

	cmd_id = circum_idmap_cmd_code(pkt->command);
	if (cmd_id == NULL)
		return -1;

	circum_pkt_avpcluster(pkt);

	rl_fixed = (pkt->cmdflags & A3_CMDFLAG_REQUEST) ?
	           &cmd_id->reqrules_fixed : &cmd_id->rsprules_fixed;
	rl_var   = (pkt->cmdflags & A3_CMDFLAG_REQUEST) ?
	           &cmd_id->reqrules_var : &cmd_id->rsprules_var;

	if (rl_fixed->items == 0 && rl_var->items == 0)
		/*
		 * If both fixed and required/optional are empty, no rules were
		 * specified in the dictionary. Allowing all AVPs implies that
		 * this is explicitly said so in the dictionary!
		 * (Using <avprule name="AVP" />)
		 */
		return -1;

	return 1;
}

static const char *extract_attrs(xmlNode *node, unsigned int *type,
    struct circum_avprule **rule_ptr, bool avp)
{
	const char *structure_mismatch = avp ?
		"<avprule> in bad XML context -- must have a parent "
		"<fixed>, <required> or <optional> and a grandparent "
		"<grouped>" :
		"<avprule> in bad XML context -- must have a parent "
		"<fixed>, <required> or <optional> and a grandparent "
		"<requestrules> or <answerrules>";
	unsigned int min = 1, max = 1;
	struct circum_avprule *rule;
	xmlNode *p;
	char *tmp;

	*type = 0;
	p = node->parent;
	if (p == NULL || p->name == NULL) {
		return 0;
	} else if (xml_strcmp(p->name, "fixed") == 0) {
		min    = 1;
		max    = 1;
		*type |= RULE_FIXED;
	} else if (xml_strcmp(p->name, "required") == 0) {
		min    = 1;
		max    = ~0U;
		*type |= RULE_REQUIRED;
	} else if (xml_strcmp(p->name, "optional") == 0) {
		min    = 0;
		max    = ~0U;
		*type |= RULE_OPTIONAL;
	} else {
		return structure_mismatch;
	}

	p = p->parent;
	if (p == NULL || p->name == NULL)
		return 0;
	else if (avp && xml_strcmp(p->name, "grouped") == 0)
		/* nothing */;
	else if (!avp && xml_strcmp(p->name, "requestrules") == 0)
		*type |= RULE_REQUEST;
	else if (!avp && xml_strcmp(p->name, "answerrules") == 0)
		*type |= RULE_ANSWER;
	else
		return structure_mismatch;

	tmp = xml_getprop(node, "minimum");
	if (tmp != NULL) {
		min = strtoul(tmp, NULL, 0);
		free(tmp);
	}
	tmp = xml_getprop(node, "maximum");
	if (tmp != NULL) {
		max = strtoul(tmp, NULL, 0);
		free(tmp);
	}

	if ((rule = malloc(sizeof(struct circum_avprule))) == NULL)
		return "Memory allocation failure";

	tmp = xml_getprop(node, "name");
	if (tmp == NULL) {
		free(rule);
		return "\"name\" attribute missing in <avprule>";
	}

	rule->avp_id = circum_idmap_avp_name(tmp);
	if (rule->avp_id == NULL) {
		free(tmp);
		free(rule);
		return "Could not resolve the name of <avprule>";
	}

	free(tmp);
	rule->min = min;
	rule->max = max;
	rule->required = (*type & RULE_FRO_MASK) != RULE_OPTIONAL;
	HXlist_init(&rule->list);
	*rule_ptr = rule;
	return NULL;
}

/**
 * circum_dict_tag_avprule - handle <avp> subtags
 * @ptr:	XML node
 * @private:	struct a3_idmap_avp of the parent <avp>
 */
static const char *circum_dict_tag_avprule(xmlNode *node, void *private)
{
	struct a3_idmap_avp *avp = private;
	struct circum_avprule *rule;
	unsigned int type;
	const char *err;
	char *tmp;

	if (xml_strcmp(node->name, "avprule") != 0)
		return circum_dict_recursive(circum_dict_vfyavp_tags,
		       node, private);

	if (avp->type != A3_AVPTYPE_GROUP)
		return "AVP is not grouped";

	err = extract_attrs(node, &type, &rule, true);
	if (err != NULL)
		goto out;

	if ((type & RULE_FRO_MASK) == RULE_FIXED)
		HXclist_push(&avp->rules_fixed, &rule->list);
	else
		HXclist_push(&avp->rules_var, &rule->list);

	return NULL;

 out:
	tmp = xml_getprop(node, "name");
	fprintf(stderr, "%s: Unable to resolve AVP \"%s\" "
	        "(rule for grouped AVP %u)\n",
	        __func__, tmp, avp->code);
	free(tmp);
	return err;
}

/**
 * circum_dict_tag_cmdrule - handle <command> subtags
 * @ptr:	XML node
 * @private:	struct a3_idmap_cmd of the parent <command>
 */
static const char *circum_dict_tag_cmdrule(xmlNode *node, void *private)
{
	struct a3_idmap_cmd *cmd = private;
	struct circum_avprule *rule;
	unsigned int type;
	const char *err;
	char *tmp;

	/*
	 * We do it this way so that we do not need one function for each of
	 * the <requestrules><fixed>, <requestrules><required>,
	 * <requestrules><optional>, <answerrules><fixed>, etc.
	 */
	if (xml_strcmp(node->name, "avprule") != 0)
		return circum_dict_recursive(circum_dict_vfycmd_tags,
		       node, private);

	err = extract_attrs(node, &type, &rule, false);
	if (err != NULL)
		goto out;

	if (type & RULE_REQUEST) {
		if ((type & RULE_FRO_MASK) == RULE_FIXED)
			HXclist_push(&cmd->reqrules_fixed, &rule->list);
		else
			HXclist_push(&cmd->reqrules_var, &rule->list);
	} else {
		if ((type & RULE_FRO_MASK) == RULE_FIXED)
			HXclist_push(&cmd->rsprules_fixed, &rule->list);
		else
			HXclist_push(&cmd->rsprules_var, &rule->list);
	}

	return NULL;

 out:
	tmp = xml_getprop(node, "name");
	fprintf(stderr, "%s: Unable to resolve AVP \"%s\" "
	        "(rule for command %u)\n",
	        __func__, tmp, cmd->code);
	free(tmp);
	return err;
}

static const struct circum_xmltag_map circum_dict_vfyavp_tags[] = {
	{NULL, circum_dict_tag_avprule},
	{NULL},
};

static const struct circum_xmltag_map circum_dict_vfycmd_tags[] = {
	{NULL, circum_dict_tag_cmdrule},
	{NULL},
};

static const char *circum_dict_tag_avp2(xmlNode *node, void *private)
{
	struct a3_idmap_avp *avp_id;
	char *code;

	if ((code = xml_getprop(node, "code")) == NULL)
		return NULL;
	avp_id = circum_idmap_avp_code(strtoul(code, NULL, 0));
	free(code);
	if (avp_id == NULL)
		return NULL;

	return circum_dict_recursive(circum_dict_vfyavp_tags, node, avp_id);
}

static const char *circum_dict_tag_command2(xmlNode *node, void *private)
{
	struct a3_idmap_cmd *cmd_id;
	char *code;

	if ((code = xml_getprop(node, "code")) == NULL)
		return NULL;
	cmd_id = circum_idmap_cmd_code(strtoul(code, NULL, 0));
	free(code);
	if (cmd_id == NULL)
		return NULL;

	return circum_dict_recursive(circum_dict_vfycmd_tags, node, cmd_id);
}

static const char *circum_dict_vfy(xmlNode *node, void *private)
{
	return circum_dict_recursive(circum_dict_vfytop_tags, node, private);
}

/* Using a compressed layout to save on functions */
const struct circum_xmltag_map circum_dict_vfytop_tags[] = {
	{"avp",     circum_dict_tag_avp2},
	{"command", circum_dict_tag_command2},
	{NULL,      circum_dict_vfy},
	{NULL},
};
