/*
 *	Copyright © Jan Engelhardt, 2008
 *
 *	This file is part of Codename Circumference. It is free software;
 *	you can redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation, either version 3 of the License, or any later version.
 */
#include <libcircumplus/packet.hpp>

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/types.h>
#include <libcircum.h>
#include <libcircum/internal.h>

struct a3_packet *A3_packet::alloc_ll(void)
{
	struct a3_packet *pkt = a3_pkt_new();
	if (pkt == NULL)
		throw 0;
	return pkt;
}

A3_packet::A3_packet(void) :
	ll_pkt(alloc_ll()),
	cmdflags(ll_pkt->cmdflags),
	command(ll_pkt->command),
	appl_id(ll_pkt->appl_id),
	hbh_id(ll_pkt->hbh_id),
	ete_id(ll_pkt->ete_id)
{
}

/**
 * A3_packet -
 * @pkt:	lowlevel packet
 *
 * Private constructor to create an A3_packet class around an existing
 * a3_packet structure. Used during unserialize().
 */
A3_packet::A3_packet(struct a3_packet *pkt) :
	ll_pkt(pkt),
	cmdflags(ll_pkt->cmdflags),
	command(ll_pkt->command),
	appl_id(ll_pkt->appl_id),
	hbh_id(ll_pkt->hbh_id),
	ete_id(ll_pkt->ete_id)
{
	struct a3_avp *ll_avp;
	A3_avp *avp;

	HXlist_for_each_entry(ll_avp, &ll_pkt->avp_list, list) {
		avp = new A3_avp(ll_avp);
		avp_list.push_back(avp);
	}
}

A3_packet::~A3_packet(void)
{
	for (std::list<A3_avp *>::iterator iter = avp_list.begin();
	     iter != avp_list.end(); ++iter)
		delete *iter;

	if (ll_pkt != NULL)
		a3_pkt_destroy(ll_pkt);
}

/**
 * A3_packet::add_* -
 *
 * Add AVPs to packet, various functions.
 */
A3_avp *A3_packet::add_avp(const char *type, const void *value,
    unsigned int size)
{
 	A3_avp *avp = new A3_avp(ll_pkt, type, value, size);
	avp_list.push_back(avp);
	return avp;
}

A3_avp *A3_packet::add_text(const char *name, const char *value)
{
	return add_avp(name, value, strlen(value));
}

A3_avp *A3_packet::add_int(const char *name, long long value)
{
	char buf[sizeof("18446744073709551616")];
	snprintf(buf, sizeof(buf), "%lld", value);
	return add_avp(name, buf, strlen(buf));
}

A3_avp *A3_packet::add_enum(const char *name, const char *ename)
{
	return add_int(name, a3_idmap_enum(name, ename));
}

static A3_avp *Circum_avp_search(const std::list<A3_avp *> &list,
    const char *name)
{
	const struct a3_idmap_avp *id = circum_idmap_avp_name(name);

	if (id == NULL) {
		errno = ENOENT;
		return NULL;
	}

	for (std::list<A3_avp *>::const_iterator iter = list.begin();
	     iter != list.end(); ++iter)
		if ((*iter)->code == id->code)
			return *iter;
	errno = 0;
	return NULL;
}

A3_avp *A3_packet::find(const char *name) const
{
	return Circum_avp_search(avp_list, name);
}

void *A3_packet::serialize(unsigned int *size) const
{
	return a3_pkt_serialize(ll_pkt, size);
}

A3_packet *A3_packet::unserialize(const void *ptr, unsigned int size)
{
	struct a3_packet *ll_pkt = a3_pkt_unserialize(ptr, size);
	if (ll_pkt == NULL)
		return NULL;
	return new A3_packet(ll_pkt);
}

void A3_packet::treedump(void) const
{
	circum_treedump_pkt(ll_pkt, 0);
}

struct a3_avp *A3_avp::alloc_ll(struct a3_packet *pkt, const char *name,
    const void *value, unsigned int length)
{
	struct a3_avp *avp = a3_pkt_add_avp(pkt, name, value, length);
	if (avp == NULL)
		throw 0;
	return avp;
}

struct a3_avp *A3_avp::alloc_ll(struct a3_avp *avp, const char *name,
    const void *value, unsigned int length)
{
	struct a3_avp *subavp = a3_avp_add_avp(avp, name, value, length);
	if (subavp == NULL)
		throw 0;
	return subavp;
}

/**
 * A3_avp -
 * @obj:	may be one of struct a3_packet or struct a3_avp.
 * @name:	name of the AVP
 * @value:	and its value
 * @length:	blob length
 *
 * Private constructors for calling the lowlevel AVP create function and
 * constructing the A3_avp class around it.
 */
template<class ll_object>
A3_avp::A3_avp(ll_object *obj, const char *name, const void *value,
    unsigned int length) :
	ll_avp(alloc_ll(obj, name, value, length)),
	code(ll_avp->code),
	flags(ll_avp->flags),
	type(ll_avp->type),
	vendor_id(ll_avp->vendor_id),
	d(ll_avp->d),
	blob_length(ll_avp->blob_length),
	blob(ll_avp->blob)
{
}

/**
 * A3_avp -
 * @avp:	lowlevel AVP
 *
 * Private constructor to create an A3_avp class around an existing
 * a3_avp structure. Used during unserialize().
 */
A3_avp::A3_avp(struct a3_avp *avp) :
	ll_avp(avp),
	code(ll_avp->code),
	flags(ll_avp->flags),
	type(ll_avp->type),
	vendor_id(ll_avp->vendor_id),
	d(ll_avp->d),
	blob_length(ll_avp->blob_length),
	blob(ll_avp->blob)
{
	if (type != A3_AVPTYPE_GROUP)
		return;

	struct a3_avp *ll_subavp;
	A3_avp *subavp;

	HXlist_for_each_entry(ll_subavp, &ll_avp->d.avp_list, list) {
		subavp = new A3_avp(ll_subavp);
		avp_list.push_back(subavp);
	}
}

A3_avp::~A3_avp(void)
{
	if (type != A3_AVPTYPE_GROUP)
		return;

	for (std::list<A3_avp *>::iterator iter = avp_list.begin();
	     iter != avp_list.end(); ++iter)
		delete *iter;
}

/**
 * A3_avp::add_* -
 *
 * Add sub-AVPs to AVP, various functions.
 */
A3_avp *A3_avp::add_avp(const char *name, const void *value,
    unsigned int length)
{
	A3_avp *avp = new A3_avp(ll_avp, name, value, length);
	avp_list.push_back(avp);
	return avp;
}

A3_avp *A3_avp::add_text(const char *name, const char *value)
{
	return add_avp(name, value, strlen(value));
}

A3_avp *A3_avp::add_int(const char *name, long long value)
{
	char buf[sizeof("18446744073709551616")];
	snprintf(buf, sizeof(buf), "%lld", value);
	return add_avp(name, buf, strlen(buf));
}

A3_avp *A3_avp::add_enum(const char *name, const char *ename)
{
	return add_int(name, a3_idmap_enum(name, ename));
}

A3_avp *A3_avp::find(const char *name) const
{
	if (type != A3_AVPTYPE_GROUP)
		return NULL;
	return Circum_avp_search(avp_list, name);
}
